package com.gau.utils.net;

import java.net.URI;
import java.util.ArrayList;
import java.util.Collections;
import java.util.Comparator;
import java.util.List;
import java.util.Map;

import org.apache.http.HttpResponse;

import android.content.Context;
import android.os.Handler;
import android.os.Looper;

import com.gau.utils.net.connector.AbstractConnector;
import com.gau.utils.net.connector.ConnectorFactory;
import com.gau.utils.net.request.THttpRequest;
import com.gau.utils.net.response.IResponse;
import com.gau.utils.net.util.HeartSetting;
import com.gau.utils.net.util.NetLog;

/**
 * 网络连接调度器，通过调度器可以管理Http連接，實現多線程網絡連接
 * 
 * @author guoshengliang
 * @version 1.0
 */

/**
 * 修改内容 (1)将任务列表整合为只有2个，1个是等待任务列表，另一个是正在运行的任务列表，提高代码可读性以及可维护性
 * (2)将列表任务统一为AbstractConnector
 * ,当前AbstractConnector分为普通网络连接(HttpConnector)和网络长连接(HttpConnectorAlive)
 * 
 * @author liguoliang
 * @version 1.1
 */
public class HttpConnectScheduler implements IConnectListener {
	/**
	 * 最大并发网络连接数
	 */
	private int mMaxConnectThreadNum = 2;

	/**
	 * 等待任务列表
	 */
	private List<THttpRequest> mWaitRequestList = new ArrayList<THttpRequest>();

	/**
	 * 当前正在运行的任务列表
	 */
	private List<THttpRequest> mCurrentRequestlist = new ArrayList<THttpRequest>();

	private ConnectorFactory mConnectorFactory;

	private Context mContext;

	private byte[] mLock = new byte[0]; // 创建同步锁

	private Handler mHandler = new Handler(Looper.getMainLooper());

	/**
	 * 构造函数
	 * 
	 * @param context
	 * @throws IllegalArgumentException
	 */
	public HttpConnectScheduler(Context context)
			throws IllegalArgumentException {
		if (context == null) {
			throw new IllegalArgumentException("context must have value");
		}
		this.mContext = context;
		mConnectorFactory = new ConnectorFactory();
	}

	public void addRequest(THttpRequest request) {
		if (request == null) {
			// 抛异常
			throw new IllegalArgumentException("Addrequest must have request");
		}

		if (request.getIsAsync()) {
			synchronized (mLock) {
				// 异步连接加入到网络队列中
				mWaitRequestList.add(request);
				// 对任务队列根据优先级进行排序
				Collections.sort(mWaitRequestList, new Comparator<THttpRequest>() {

					@Override
					public int compare(THttpRequest object1, THttpRequest object2) {
						if (object1.getRequestPriority() < object2.getRequestPriority()) {
							return 1;
						} else if (object1.getRequestPriority() > object2.getRequestPriority()) {
							return -1;
						}
						return 0;
					}
				});
				connect();
			}
		} else {
			// 同步连接直接联网
			connect(request);
		}
	}

	public void cleanup() {
		if (mConnectorFactory != null) {
			mConnectorFactory.cleanup();
		}
	}

	public int getMaxConnectThreadNum() {
		return mMaxConnectThreadNum;
	}

	public void setMaxConnectThreadNum(int maxConnectThreadNum) {
		mMaxConnectThreadNum = maxConnectThreadNum;
	}

	/**
	 * 添加心跳包地址映射表,如果已经存在,则替换旧值
	 * 
	 * @param host
	 * @param heartUrl
	 * @return 被替换的心跳包地址值
	 */
	public String putCommonHeartUrl(String host, String heartUrl) {
		if (mConnectorFactory != null) {
			return mConnectorFactory.putCommonHeartUrl(host, heartUrl);
		}
		return null;
	}

	/**
	 * 移除心跳包地址键值对
	 * 
	 * @param host
	 * @return 被删除的心跳包地址值
	 */
	public String removeHeartUrl(String host) {
		if (mConnectorFactory != null) {
			return mConnectorFactory.removeHeartUrl(host);
		}
		return null;
	}

	public String getHeartUrl(String host) {
		if (mConnectorFactory != null) {
			return mConnectorFactory.getHeartUrl(host);
		}
		return null;
	}

	public Map<String, String> getAllCommonHeartUrl() {
		if (mConnectorFactory != null) {
			return mConnectorFactory.getAllHeartUrl();
		}
		return null;
	}

	public void setCommonHeartTime(long heartTime) {
		if (mConnectorFactory != null) {
			mConnectorFactory.setCommonHeartTime(heartTime);
		}
	}

	public long getCommonHeartTime() {
		if (mConnectorFactory != null) {
			return mConnectorFactory.getCommonHeartTime();
		}
		return HeartSetting.HEART_TIME_NO_SETTING;
	}

	/**
	 * 是否存在请求
	 * 
	 * @author huyong
	 * @param url
	 *            待检查的请求地址
	 * @param requestQuene
	 *            待检查的请求队列
	 * @return request which equals to the url，or null for not found。
	 */
	public THttpRequest isExistEqualRequest(final String url,
			final List<THttpRequest> requestQuene) {
		THttpRequest result = null;
		if (url == null || requestQuene == null || requestQuene.size() <= 0) {
			return result;
		}
		for (THttpRequest httpRequest : requestQuene) {
			URI uri = httpRequest.getUrl();
			if (uri != null && uri.toString() != null
					&& uri.toString().equals(url)) {
				result = httpRequest;
				break;
			}
		}

		return result;
	}

	private THttpRequest isExistRequests(final String url,
			final List<THttpRequest> requestList) {
		THttpRequest ret = null;
		if (url == null || requestList == null || requestList.isEmpty()) {
			return null;
		}
		for (THttpRequest request : requestList) {
			if (request == null) {
				continue;
			}
			URI uri = request.getUrl();
			if (uri != null && uri.toString() != null
					&& uri.toString().equals(url)) {
				ret = request;
				break;
			}
		}
		return ret;
	}

	public boolean cancelRequest(String url) {
		if (url == null || url.length() < 1) {
			return false;
		}
		boolean ret = false;
		synchronized (mLock) {
			if (mWaitRequestList != null && !mWaitRequestList.isEmpty()) {
				THttpRequest request = isExistRequests(url, mWaitRequestList);
				if (request != null) {
					// 如果任务处于等待状态，则直接从等待列表中移除任务即可
					ret = mWaitRequestList.remove(request);
				}
			}
			if (!ret && mCurrentRequestlist != null
					&& !mCurrentRequestlist.isEmpty()) {
				// 如果不在等待队列中,则遍历正在运行的队列
				THttpRequest request = isExistRequests(url, mCurrentRequestlist);
				if (request == null || mConnectorFactory == null) {
					return false;
				}
				request.setCanceled(true);
				AbstractConnector connector = mConnectorFactory
						.isExistConnector(request);
				if (connector != null) {
					connector.cancelCurrentConnect();
					mConnectorFactory.removeConnectorByRequest(request);
					ret = true;
				}
			}
		}
		return ret;
	}

	public boolean cancelRequest(THttpRequest request) {
		if (request == null) {
			return false;
		}
		// TODO:LIGUOLIANG 这句有什么用,为什么要放在此处,移走
		// request.setCanceled(true);
		boolean ret = false;
		synchronized (mLock) {
			if (mWaitRequestList != null && !mWaitRequestList.isEmpty()) {
				// 如果任务处于等待状态，则直接从等待列表中移除任务即可
				ret = mWaitRequestList.remove(request);
			}
			if (!ret && mCurrentRequestlist != null
					&& !mCurrentRequestlist.isEmpty()) {
				// 如果不在等待队列中,则遍历正在运行的队列
				request.setCanceled(true);
				AbstractConnector connector = mConnectorFactory
						.isExistConnector(request);
				if (connector != null) {
					connector.cancelCurrentConnect();
					mConnectorFactory.removeConnectorByRequest(request);
					ret = true;
				}
			}
		}
		return ret;
	}

	@Override
	public void onFinish(THttpRequest request, IResponse response) {
		NetLog.info("schedule onFinish", null);
		NetLog.info("testBattery, Begin HttpConnectScheduler onFinish", null);
		if (request == null) {
			return;
		}
		request.getReceiver().onFinish(request, response);
		synchronized (mLock) {
			if (mCurrentRequestlist != null && !mCurrentRequestlist.isEmpty()) {
				mCurrentRequestlist.remove(request);
				if (mConnectorFactory != null) {
					mConnectorFactory.removeConnectorByRequest(request);
				}
			}
		}
		tick();
		NetLog.info("testBattery, End HttpConnectScheduler onFinish", null);
	}

	@Override
	public void onStart(THttpRequest request) {
		IConnectListener receiver = request.getReceiver();
		receiver.onStart(request);
	}

	@Override
	public void onException(THttpRequest request, HttpResponse response, int reason) {
		NetLog.info("schedule onException", null);
		NetLog.info("testBattery, Begin HttpConnectScheduler onException", null);
		if (request == null) {
			return;
		}
		IConnectListener receiver = request.getReceiver();
		if (receiver != null) {
			receiver.onException(request, response, reason);
		}
		synchronized (mLock) {
			if (mCurrentRequestlist != null && !mCurrentRequestlist.isEmpty()) {
				mCurrentRequestlist.remove(request);
				if (mConnectorFactory != null) {
					mConnectorFactory.removeConnectorByRequest(request);
				}
			}
		}
		tick();
		NetLog.info("testBattery, End HttpConnectScheduler onException", null);

	}
	
	@Override
	public void onException(THttpRequest request, int reason) {
		onException(request, null, reason);
	}

	private void tick() {
		// TODO：LIGUOLIANG
		// 当一个联网任务完成或出现异常时会调用此接口,当从一个HttpConnector执行完时，此时是处在子线程的接口里，需要重新回到主线程，所以此处通过Handler发回主线程调用
		mHandler.post(new Runnable() {
			@Override
			public void run() {
				synchronized (mLock) {
					NetLog.info("testBattery, Begin HttpConnectScheduler tick",
							null);
					connect();
					NetLog.info("testBattery, end HttpConnectScheduler tick",
							null);
				}
			}
		});
	}

	/**
	 * 根据默认规则连接网络
	 */
	private void connect() {
		if (mWaitRequestList == null || mWaitRequestList.isEmpty()
				|| mCurrentRequestlist == null) {
			return;
		}
		int currentSize = mCurrentRequestlist.size();
		if (currentSize < mMaxConnectThreadNum) {
			for (int i = 0; i < mMaxConnectThreadNum - currentSize; ++i) {
				if (mWaitRequestList.isEmpty()) {
					// 如果任务列表中已经没有任务了，则退出
					break;
				}
			}
			THttpRequest request = mWaitRequestList.remove(0);
			if (request != null) {
				mCurrentRequestlist.add(request);
				connect(request);
			}
		}
	}

	private void connect(THttpRequest request) {
		AbstractConnector connector = mConnectorFactory.productConnector(
				request, this, mContext);
		if (connector == null) {
			return;
		}
		if (connector.getRequset().getIsAsync()) {
			connector.connectAsynchronous();
		} else {
			connector.connect();
		}
	}
}
