/**
 * 
 */
package com.gau.utils.net.connector;

import java.util.Iterator;
import java.util.Map;
import java.util.Map.Entry;
import java.util.concurrent.ConcurrentHashMap;

import android.content.Context;

import com.gau.utils.net.IConnectListener;
import com.gau.utils.net.connector.HttpConnectorAlive.IConnectorDestroyListener;
import com.gau.utils.net.request.THttpRequest;
import com.gau.utils.net.util.HeartSetting;

/**
 * 
 * @author liguoliang
 * 
 */
public class ConnectorFactory implements IConnectorDestroyListener {

	/**
	 * 最大同时可用的长连接数
	 */
	private int mMaxAliveConnectorNum = DEFAULT_MAX_ALIVE_CONNECTOR_NUM;

	/**
	 * 长连接域名映射表
	 */
	private Map<String, HttpConnectorAlive> mAliveConnectorMap;

	/**
	 * 普通连接列表
	 */
	private Map<THttpRequest, HttpConnector> mHttpConnectorMap;

	/**
	 * 心跳包地址映射
	 */
	private HeartSetting mHeartSetting = new HeartSetting();

	/**
	 * 默认的长连接最大连接数
	 */
	private static final int DEFAULT_MAX_ALIVE_CONNECTOR_NUM = 2;

	public AbstractConnector productConnector(THttpRequest request,
			IConnectListener connectReceiver, Context context) {
		AbstractConnector connector = null;
		if (request.getIsKeepAlive()) {
			// 长连接必须使用异步连接
			request.setIsAsync(true);
			synchronized (this) {
				connector = productAliveConnector(request, connectReceiver,
						context);
			}
		} else {
			connector = productHttpConnector(request, connectReceiver, context);
		}
		return connector;
	}

	/**
	 * <br>
	 * 功能简述:创建普通网络连接 <br>
	 * 功能详细描述: <br>
	 * 注意:
	 * 
	 * @param request
	 * @param connectReceiver
	 * @param context
	 * @return
	 */
	private AbstractConnector productHttpConnector(THttpRequest request,
			IConnectListener connectReceiver, Context context) {
		AbstractConnector connector = null;
		// 非长连接
		if (mHttpConnectorMap == null) {
			mHttpConnectorMap = new ConcurrentHashMap<THttpRequest, HttpConnector>();
		}
		connector = new HttpConnector(request, connectReceiver, context);
		mHttpConnectorMap.put(request, ( HttpConnector ) connector);
		return connector;
	}

	@SuppressWarnings("unchecked")
	private AbstractConnector productAliveConnector(THttpRequest request,
			IConnectListener connectReceiver, Context context) {
		AbstractConnector connector = null;
		if (mAliveConnectorMap == null) {
			mAliveConnectorMap = new ConcurrentHashMap<String, HttpConnectorAlive>();
		}
		// 以host作为键值,host相同则认为可以使用同一个连接
		String host = request.getUrl().getHost();
		if (mAliveConnectorMap.containsKey(host)) {
			// 如果已经存在该key值
			HttpConnectorAlive tempConnector = mAliveConnectorMap.get(host);
			if (request.getIsMustAlive()) {
				// 如果必须使用长连接
				request.setIsKeepAlive(true);
				tempConnector.addRequest(request, connectReceiver);
				connector = tempConnector;
			} else {
				// 非必须使用长连接,则为了保持并发,可以使用普通连接
				if (tempConnector.isIdle()) {
					// 并且该任务是空闲的则保持到任务列表中
					// Log.i("ABEN", "addAliveRequest url = " +
					// request.getUrl());
					tempConnector.addRequest(request, connectReceiver);
					connector = tempConnector;
				} else {
					// 如果该任务不是空闲的,则为了保持并发性,则创建普通的Connector
					request.setIsKeepAlive(false);
					// Log.i("ABEN", "not add AliveRequest url = " +
					// request.getUrl());
					connector = productHttpConnector(request, connectReceiver,
							context);
				}
			}

		} else {
			// 如果不存在该键值
			int size = getAliveConnectorSize();
			if (size < mMaxAliveConnectorNum) {
				// 如果还没达到最大长连接数,则创立新的连接
				connector = new HttpConnectorAlive(request, connectReceiver,
						context, this);
				(( HttpConnectorAlive ) connector)
						.setHeartSetting(mHeartSetting);
				(( HttpConnectorAlive ) connector).setTag(host);
				mAliveConnectorMap.put(host, ( HttpConnectorAlive ) connector);
			} else {
				// 如果已经达到最大连接数,则选择空闲时间最长的替换
				Iterator iterator = mAliveConnectorMap.entrySet().iterator();
				String prefKey = "";
				if (iterator == null) {
					return null;
				}
				long maxIdleTime = 0;
				while (iterator.hasNext()) {
					Map.Entry<String, HttpConnectorAlive> entry = ( Entry<String, HttpConnectorAlive> ) iterator
							.next();
					if (entry == null) {
						continue;
					}
					HttpConnectorAlive tempConnector = ( HttpConnectorAlive ) entry
							.getValue();
					if (tempConnector == null) {
						continue;
					}
					long idleTime = tempConnector.getIdleTime();
					if (idleTime > maxIdleTime) {
						maxIdleTime = idleTime;
						prefKey = ( String ) entry.getKey();
						connector = tempConnector;
					}
				}
				(( HttpConnectorAlive ) connector).addRequest(request,
						connectReceiver);
				(( HttpConnectorAlive ) connector).setTag(host);

				// 用新的键值替换旧的键值
				mAliveConnectorMap.remove(prefKey);
				mAliveConnectorMap.put(host, ( HttpConnectorAlive ) connector);
			}
		}
		return connector;
	}

	public void removeConnectorByRequest(THttpRequest request) {
		if (request == null) {
			return;
		}
		if (request.getIsKeepAlive()) {
			// 长连接不需要移除Connector
		} else {
			if (mHttpConnectorMap != null && !mHttpConnectorMap.isEmpty()) {
				mHttpConnectorMap.remove(request);
			}
		}
	}

	public AbstractConnector isExistConnector(THttpRequest request) {
		if (request == null) {
			return null;
		}
		AbstractConnector connector = null;
		if (request.getIsKeepAlive()) {

		} else {
			if (mHttpConnectorMap != null && !mHttpConnectorMap.isEmpty()) {
				connector = mHttpConnectorMap.get(request);
			}
		}
		return connector;
	}

	private int getAliveConnectorSize() {
		if (mAliveConnectorMap == null || mAliveConnectorMap.isEmpty()) {
			return 0;
		}
		return mAliveConnectorMap.size();
	}

	public void cleanup() {
		closeAllAliveTask();
	}

	@SuppressWarnings("unchecked")
	public void closeAllAliveTask() {
		// 清除所有长连接任务
		synchronized (this) {
			if (mAliveConnectorMap == null || mAliveConnectorMap.isEmpty()) {
				return;
			}
			Iterator iterator = mAliveConnectorMap.entrySet().iterator();
			if (iterator == null) {
				return;
			}
			while (iterator.hasNext()) {
				Map.Entry<String, AbstractConnector> entry = ( Entry<String, AbstractConnector> ) iterator
						.next();
				if (entry == null) {
					continue;
				}
				AbstractConnector connector = entry.getValue();
				connector.closeConnect();
			}
			mAliveConnectorMap.clear();
			mAliveConnectorMap = null;
		}

	}

	public void setMaxAliveNum(int num) {
		mMaxAliveConnectorNum = num;
	}

	public int getMaxAliveNum() {
		return mMaxAliveConnectorNum;
	}

	/**
	 * 添加心跳包地址映射表,如果已经存在,则替换旧值
	 * 
	 * @param host
	 * @param heartUrl
	 * @return 被替换的心跳包地址值
	 */
	public String putCommonHeartUrl(String host, String heartUrl) {
		if (mHeartSetting != null) {
			return mHeartSetting.putHeartUrl(host, heartUrl);
		}
		return null;
	}

	/**
	 * 移除心跳包地址键值对
	 * 
	 * @param host
	 * @return 被删除的心跳包地址值
	 */
	public String removeHeartUrl(String host) {
		if (mHeartSetting != null) {
			return mHeartSetting.removeHeartUrl(host);
		}
		return null;
	}

	public String getHeartUrl(String host) {
		if (mHeartSetting != null) {
			return mHeartSetting.getHeartUrl(host);
		}
		return null;
	}

	public Map<String, String> getAllHeartUrl() {
		if (mHeartSetting != null) {
			return mHeartSetting.getAllHeartUrl();
		}
		return null;
	}

	public void setCommonHeartTime(long heartTime) {
		if (mHeartSetting != null) {
			mHeartSetting.setCommonHeartTime(heartTime);
		}
	}

	public long getCommonHeartTime() {
		if (mHeartSetting != null) {
			return mHeartSetting.getCommonHeartTime();
		}
		return HeartSetting.HEART_TIME_NO_SETTING;
	}

	private void removeFromMap(AbstractConnector connector) {
		synchronized (this) {
			if (mAliveConnectorMap == null) {
				return;
			}
			String tag = ( String ) (( HttpConnectorAlive ) connector).getTag();
			mAliveConnectorMap.remove(tag);
		}
	}

	/*
	 * 连接销毁事件回调
	 */
	@Override
	public void onConnectorDestroy(AbstractConnector connector) {
		removeFromMap(connector);
	}
}
