/**
 * 
 */
package com.gau.utils.net.connector;

import java.io.IOException;
import java.net.SocketTimeoutException;
import java.net.URI;
import java.net.URISyntaxException;
import java.util.ArrayList;
import java.util.Collections;
import java.util.Comparator;
import java.util.List;

import org.apache.http.Header;
import org.apache.http.HttpHost;
import org.apache.http.HttpResponse;
import org.apache.http.HttpStatus;
import org.apache.http.client.ClientProtocolException;
import org.apache.http.client.HttpClient;
import org.apache.http.client.methods.HttpGet;
import org.apache.http.client.methods.HttpPost;
import org.apache.http.client.methods.HttpRequestBase;
import org.apache.http.conn.ConnectTimeoutException;
import org.apache.http.conn.params.ConnRoutePNames;
import org.apache.http.entity.ByteArrayEntity;
import org.apache.http.impl.client.DefaultHttpClient;
import org.apache.http.params.CoreConnectionPNames;

import android.content.Context;
import android.os.Handler;
import android.os.Looper;
import android.util.Log;

import com.gau.utils.net.IConnectListener;
import com.gau.utils.net.IHttpException;
import com.gau.utils.net.INetRecord;
import com.gau.utils.net.NetException;
import com.gau.utils.net.asrFiltier.IAsrFilter;
import com.gau.utils.net.operator.IHttpOperator;
import com.gau.utils.net.request.THttpRequest;
import com.gau.utils.net.response.IResponse;
import com.gau.utils.net.util.HeartSetting;
import com.gau.utils.net.util.NetLog;
import com.gau.utils.net.util.NetUtil;

/**
 * HttpConnectorAlive:长连接的实现方法
 * 由于长连接需要维持一个心跳包,心跳包简单而言其实也就是一个连接任务,所以其实应该实现一个长连接内部的任务调度器AliveTaskScheduler
 * ,通过这个调度器是实现长连接任务的调度
 * 
 * @author liguoliang
 * 
 */
public class HttpConnectorAlive extends AbstractConnector {

	private HttpClient mHttpClient = null;

	/**
	 * 任务列表
	 */
	private List<THttpRequest> mRequestList;

	/**
	 * 监听器列表
	 */
	private List<IConnectListener> mConnectorListenerList;

	private THttpRequest mHeartRequest;

	private IConnectListener mHeartListener;

	private Handler mHandler;

	/**
	 * 切换url地址次数
	 */
	private int mSwitchUriTimes = 0;

	private Thread mConnectThread;

	/**
	 * 是否维持连接线程
	 */
	private boolean mIsCloseConnect = false;

	/**
	 * 同步锁
	 */
	private Object mSyncObject = new Object();

	/**
	 * 长连接的生命时长,即如果长连接在此时间内都没有新任务,则销毁此长连接
	 */
	private static final long ALIVE_LIFT_TIME = 3 * 60 * 1000;

	/**
	 * 记录任务的时间戳
	 */
	private long mTimeStamp = 0;

	private HeartSetting mHeartSetting;

	/**
	 * 长连接销毁监听
	 */
	private IConnectorDestroyListener mDestroyListener;

	/**
	 * 
	 */
	private Object mTag;

	/**
	 * 长连接销毁监听
	 * 
	 * @author liguoliang
	 * 
	 */
	public interface IConnectorDestroyListener {
		void onConnectorDestroy(AbstractConnector connector);
	}

	public HttpConnectorAlive(THttpRequest httpRequest, Context context,
			IConnectorDestroyListener destroyListener) {
		super(httpRequest, context);
		mDestroyListener = destroyListener;
		init();
	}

	public HttpConnectorAlive(THttpRequest httpRequest,
			IConnectListener connectReceiver, Context context,
			IConnectorDestroyListener destroyListener)
			throws IllegalArgumentException {
		super(httpRequest, connectReceiver, context);
		mDestroyListener = destroyListener;
		init();
	}

	public void setHeartSetting(HeartSetting setting) {
		this.mHeartSetting = setting;
	}

	public void addRequest(THttpRequest httpRequest,
			IConnectListener connectReceiver) {
		synchronized (mSyncObject) {
			long currentTime = System.currentTimeMillis();
			if (!httpRequest.equals(mHeartRequest)) {
				// 如果不是心跳包,则更新时间戳
				mTimeStamp = currentTime;

				// 有请求时不需要发送此次心跳包
				if (mRequestList.contains(mHeartRequest)) {					
					mRequestList.remove(mHeartRequest);
				}
				if (mHandler != null) {
					mHandler.removeCallbacks(mHeartRunnable);
				}
			} else {
				// 如果是心跳包并且已经超出了长连接的生命周期,则销毁它
				if (currentTime - mTimeStamp > ALIVE_LIFT_TIME) {
					closeConnect();
					if (mDestroyListener != null) {
						mDestroyListener
								.onConnectorDestroy(HttpConnectorAlive.this);
					}
					return;
				}
			}
			// 根据优先级进行排序
			mRequestList.add(httpRequest);
			Collections.sort(mRequestList, new Comparator<THttpRequest>() {

				@Override
				public int compare(THttpRequest object1, THttpRequest object2) {
					if (object1.getRequestPriority() < object2
							.getRequestPriority()) {
						return 1;
					} else if (object1.getRequestPriority() > object2
							.getRequestPriority()) {
						return -1;
					}
					return 0;
				}
			});
			mConnectorListenerList.add(mRequestList.indexOf(httpRequest),
					connectReceiver);

			if (mIsCloseConnect) {
				mIsCloseConnect = false;
			}

			// 设置心跳包的时间
			if (mHeartRequest != null) {
				mHeartRequest.setHeartTime(httpRequest.getHeartTime());
			}
			mSyncObject.notifyAll();
		}
	}

	public List<THttpRequest> getRequestList() {
		return mRequestList;
	}

	public int getRequestSize() {
		if (mRequestList != null) {
			return mRequestList.size();
		}
		return 0;
	}

	public void setTag(Object object) {
		this.mTag = object;
	}

	public Object getTag() {
		return mTag;
	}

	/**
	 * <br>
	 * 功能简述:获取空闲时间 ,空闲时间越长的被销毁替换的个概率更高<br>
	 * 功能详细描述: <br>
	 * 注意:
	 * 
	 * @return
	 */
	public long getIdleTime() {
		long currentTime = System.currentTimeMillis();
		return currentTime - mTimeStamp;
	}

	public void removeRequest(THttpRequest httpRequest,
			IConnectListener connectReceiver) {
		synchronized (mSyncObject) {
			if (mRequestList != null && mRequestList.size() > 0) {
				mRequestList.remove(httpRequest);
			}
			if (mConnectorListenerList != null
					&& mConnectorListenerList.size() > 0) {
				mConnectorListenerList.remove(connectReceiver);
			}
			mSyncObject.notifyAll();
		}
	}

	private void init() {
		mTimeStamp = System.currentTimeMillis();
		mHttpClient = new DefaultHttpClient();
		mRequestList = new ArrayList<THttpRequest>();
		mConnectorListenerList = new ArrayList<IConnectListener>();
		addRequest(mRequest, mConnectObserver);
		initHeartListener();
		initHeartRequest(mRequest);
		configClient(mRequest, mHttpClient);
	}

	private void initHeartListener() {
		if (mHeartListener == null) {
			mHeartListener = new IConnectListener() {

				@Override
				public void onStart(THttpRequest request) {
				}

				@Override
				public void onFinish(THttpRequest request, IResponse response) {
				}

				@Override
				public void onException(THttpRequest request, HttpResponse response, int reason) {
					// 心跳包连接失败是否需要重连还是关闭连接,暂时使用关闭长连接
					Log.i("ABEN", "HttpConnectorAlive initHeartListener onException reason = "
							+ reason);
					closeConnect();
					if (mDestroyListener != null) {
						mDestroyListener.onConnectorDestroy(HttpConnectorAlive.this);
					}
				}

				@Override
				public void onException(THttpRequest request, int reason) {
					onException(request, null, reason);
				}
			};
		}
	}

	/**
	 * 选择心跳包地址
	 * 
	 * @param request
	 * @return
	 */
	private String selectHeartUrl(THttpRequest request) {
		String heartUrl = null;
		if (mRequest.getHeartUrl() != null) {
			// 首选THttpRequest的心跳包地址
			heartUrl = mRequest.getHeartUrl().toString();
		}
		if (heartUrl == null && mHeartSetting != null
				&& mRequest.getCurrentUrl() != null) {
			// 然后从默认心跳包地址中选择
			heartUrl = mHeartSetting.getHeartUrl(mRequest.getCurrentUrl()
					.getHost());
		}
		if (heartUrl == null) {
			// 如果地址仍然不匹配,则直接用任务的地址
			heartUrl = mRequest.getUrl().toString();
		}
		return heartUrl;
	}

	private void addHeartRequest() {
		if (!mRequestList.contains(mHeartRequest)) {
			// 避免重复添加
			// 配置当前的心跳包地址
			String heartUrl = selectHeartUrl(mRequest);
			if (heartUrl != null) {
				try {
					mHeartRequest.setUrl(heartUrl);
				} catch (URISyntaxException e) {
					e.printStackTrace();
				}
			}

			addRequest(mHeartRequest, mHeartListener);
		}
	}

	private void initHeartRequest(THttpRequest request) {
		if (mHeartRequest == null) {
			String heartUrl = selectHeartUrl(request);
			try {
				mHeartRequest = new THttpRequest(heartUrl, null, mHeartListener);
			} catch (Exception e) {
				return;
			}
			// 心跳包任务不需要读取数据
			mHeartRequest.setOperator(new IHttpOperator() {

				@Override
				public IResponse operateHttpResponse(THttpRequest request,
						HttpResponse response) throws IllegalStateException,
						IOException {
					return null;
				}
			});
			mHeartRequest.setNetRecord(new INetRecord() {

				@Override
				public void onTransFinish(THttpRequest request, Object arg1,
						Object arg2) {
				}

				@Override
				public void onStartConnect(THttpRequest request, Object arg1,
						Object arg2) {
				}

				@Override
				public void onException(Exception e, Object arg1, Object arg2) {
				}

				@Override
				public void onConnectSuccess(THttpRequest request, Object arg1,
						Object arg2) {
				}
			});

			mHeartRequest.setHeartTime(request.getHeartTime());

			// 心跳包的优先级运行比较低
			mHeartRequest.setRequestPriority(Thread.MIN_PRIORITY);
		}
	}

	private void configClient(THttpRequest request, HttpClient client) {
		if (request == null || client == null) {
			return;
		}
		int connectType = NetUtil.getNetWorkType(mContext);
		// 代理的设置
		if (NetUtil.NETWORKTYPE_CMWAP == connectType) {
			HttpHost proxy = NetUtil.getProxy(mContext);
			client.getParams().setParameter(ConnRoutePNames.DEFAULT_PROXY,
					proxy);
		}

		client.getParams().setParameter(CoreConnectionPNames.SO_TIMEOUT,
				request.getSocketTimeoutValue()); // 从socket读数据时发生阻塞的超时时间
		client.getParams().setParameter(
				CoreConnectionPNames.CONNECTION_TIMEOUT,
				request.getTimeoutValue()); // 连接的超时时间
	}

	private void setConnect() {
		NetLog.info("testBattery, Begin HttpConnector run", null);
		try {
			mRequest.setCurrentUrl(mRequest.getUrl());
			mConnectObserver.onStart(mRequest);
			IResponse response = connectSynchronous(mRequest);
			mConnectObserver.onFinish(mRequest, response);
		} catch (ClientProtocolException e) {
			NetLog.erro("ClientProtocolException", e);
			e.printStackTrace();
			mConnectObserver.onException(mRequest, null, IHttpException.ClientProtocolException);
		} catch (NetException e) {
			// 服务器异常
			NetLog.erro("IOException", e);
			e.printStackTrace();
			mConnectObserver.onException(mRequest, e.mResponse, e.mErrorCode);
		} catch (IOException e) {
			// 网络异常
			NetLog.erro("IOException", e);
			e.printStackTrace();
			mConnectObserver.onException(mRequest, null, IHttpException.IOException);
		} catch (IllegalAccessException e) {
			// 获取不到有效接入点
			NetLog.erro("can't find netWork", e);
			e.printStackTrace();
			mConnectObserver.onException(mRequest, null, IHttpException.IllegalAccessException);
		} catch (OutOfMemoryError e) {
			e.printStackTrace();
			NetLog.erro("unkown exception ", e);
			System.gc();
			mConnectObserver.onException(mRequest, null, IHttpException.OOMException);
		} catch (Throwable e) {
			// 特殊异常保护
			e.printStackTrace();
			NetLog.erro("unkown exception ", e);
			mConnectObserver.onException(mRequest, null, IHttpException.ThrowableException);
		}
		NetLog.info("testBattery, Begin HttpConnector run", null);
	}

	@Override
	public void connect() {
		// 长连接不能使用同步连接
	}

	private void handleConnect() {
		if (mRequestList.isEmpty() || mConnectorListenerList.isEmpty()) {
			return;
		}
		mRequest = mRequestList.remove(0);
		mConnectObserver = mConnectorListenerList.remove(0);
		if (mRequest == null || mConnectObserver == null) {
			return;
		}
		if (!mRequest.equals(mHeartRequest)) {
			// 如果不是心跳包则说明是在非空闲状态
			mIsIdle = false;
		} else {
			// 心跳包处理说明是在空闲状态下
			mIsIdle = true;
		}
		setConnect();
		mIsIdle = true;
	}

	private Runnable mHeartRunnable = new Runnable() {

		@Override
		public void run() {
			addHeartRequest();
			if (mHandler != null) {
				mHandler.removeCallbacks(this);
			}			
		}
	};

	@Override
	public void connectAsynchronous() {
		if (mHandler == null) {
			mHandler = new Handler(Looper.getMainLooper());
		}
		if (mConnectThread == null) {
			mConnectThread = new Thread(new Runnable() {
				@Override
				public void run() {
					while (!mIsCloseConnect) {
						synchronized (mSyncObject) {
							if (mRequestList.isEmpty()) {
								long heartTime = selectHeartTime();
								mHandler.postDelayed(mHeartRunnable, heartTime);
								try {
									mSyncObject.wait();
								} catch (InterruptedException e) {
								}
							}
						}
						handleConnect();
					}
					mConnectThread = null;
				}
			}, "AliveConnectorConnectAsynchronous");
			mConnectThread.start();
		}
	}

	/**
	 * 选择心跳包时间间隔
	 * 
	 * @return
	 */
	private long selectHeartTime() {
		long heartTime = HeartSetting.HEART_TIME_NO_SETTING;
		heartTime = mRequest.getHeartTime();
		if (heartTime == HeartSetting.HEART_TIME_NO_SETTING) {
			heartTime = mHeartSetting.getCommonHeartTime();
		}
		if (heartTime == HeartSetting.HEART_TIME_NO_SETTING) {
			heartTime = HeartSetting.DEFAULT_HEART_TIME_INTERVAL;
		}
		return heartTime;
	}

	/*
	 * 关闭长连接
	 */
	@Override
	public void closeConnect() {
		// TODO:LIGUOLIANG 调用此接口是否会有并发错误,需要关注
		if (mHttpClient != null) {
			try {
				mHttpClient.getConnectionManager().shutdown();
			} catch (Exception e) {
				e.printStackTrace();
			}
			mHttpClient = null;
		}
		if (mHandler != null) {
			mHandler.removeCallbacks(mHeartRunnable);
		}
		mIsCloseConnect = true;
	}

	@Override
	public void cancelCurrentConnect() {

	}

	private IResponse connectSynchronous(THttpRequest request)
			throws ClientProtocolException, IOException,
			IllegalAccessException, Exception {
		NetLog.info("StartConnect url= " + request.getUrl(), null);
		NetLog.info(
				"testBattery, Begin HttpConnector connectSynchronous url = "
						+ request.getUrl(), null);
		// 联网信息记录
		IResponse ret = null;
		INetRecord record = request.getNetRecord();
		try {
			URI uri = selectURI(request);
			if (uri == null) {
				throw new NetException(IHttpException.URL_ILLEGAL);
			}
			request.setCurrentUrl(uri);
			HttpHost httphost = new HttpHost(uri.getHost(), uri.getPort());
			HttpResponse response = null;
			if (mHttpClient == null) {
				mHttpClient = new DefaultHttpClient();
			}
			configClient(request, mHttpClient);
			if (request.getPostData() == null) {
				// use get
				HttpGet get = new HttpGet(uri);
				configHttpHeader(get);
				configAliveHeader(get);
				if (record != null) {
					// 连接开始回调
					record.onStartConnect(request, null, null);
				}
//				long t1 = System.currentTimeMillis();
				response = mHttpClient.execute(httphost, get);
//				long t = System.currentTimeMillis() - t1;
//				Log.i("ABEN",
//						"HttpConnectorAlive time = " + t + " "
//								+ request.getCurrentUrl());
				if (record != null) {
					// 连接成功回调
					record.onConnectSuccess(request, null, null);
				}
			} else {
				// use post method
				HttpPost httppost = new HttpPost(uri);
				configHttpHeader(httppost);
				configAliveHeader(httppost);
				ByteArrayEntity reqEntity = new ByteArrayEntity(
						request.getPostData());
				reqEntity.setChunked(false);
				httppost.setEntity(reqEntity);
				if (record != null) {
					// 连接开始回调
					record.onStartConnect(request, null, null);
				}
//				long t1 = System.currentTimeMillis();
				response = mHttpClient.execute(httphost, httppost);
//				long t = System.currentTimeMillis() - t1;
//				Log.i("ABEN",
//						"HttpConnectorAlive time = " + t + " "
//								+ request.getCurrentUrl());
				if (record != null) {
					// 连接成功回调
					record.onConnectSuccess(request, null, null);
				}
			}

			int responseCode = response.getStatusLine().getStatusCode();
			if (responseCode == HttpStatus.SC_OK) {
				IAsrFilter filter = request.getAsrFilter();
				if (filter != null && filter.isAsrResponse(response)) {
					NetLog.debug("find AsrResponse", null);
					// 使用相同地址重连
					int retryTime = request.getCurRetryTime();
					if (retryTime > 0) {
						request.setCurRetryTime(retryTime - 1);
						ret = connectSynchronous(request);
					} else {
						// 联网失败，抛出异常
						// TODO 抛个特殊的异常
						// throw new IOException();
						++mSwitchUriTimes;
						if (mSwitchUriTimes < request.getAllUrl().size()) {
							// 切换网络地址重连
							ret = connectSynchronous(request);
						} else {
							throw new NetException(
									IHttpException.FilterException);
						}
					}
				} else {
					IResponse reponseData = request.getOperator()
							.operateHttpResponse(request, response);

					// 长连接必须将数据消费掉
					response.getEntity().consumeContent();

					if (record != null) {
						// 数据传输完成回调
						record.onTransFinish(request, null, null);
					}

					return reponseData;
				}
			} else {
				// 未知原因网络连接异常，则重新连接。
				// 如果Http返回码为503，不要进行重试。Service Unavailable 服务器由于维护或者负载过重未能应答
				if (responseCode == HttpStatus.SC_SERVICE_UNAVAILABLE) {
					//503，把重试次数设置为0
					request.setCurRetryTime(0);
				}
				// 使用相同地址重连
				int retryTime = request.getCurRetryTime();
				if (retryTime > 0) {
					request.setCurRetryTime(retryTime - 1);
					ret = connectSynchronous(request);
				} else {
					// 抛个特殊的异常
					// throw new IOException();
					++mSwitchUriTimes;
					if (mSwitchUriTimes < request.getAllUrl().size()) {
						// 切换网络地址重连
						ret = connectSynchronous(request);
					} else {
						throw new NetException(responseCode);
					}
				}
			}
		} catch (Exception e) {
			// e.printStackTrace();
			// 统一在此处反馈异常码
			if (record != null) {
				record.onException(e, null, null);
			}

			// 使用相同地址重连
			int retryTime = request.getCurRetryTime();
			if (retryTime > 0) {
				request.setCurRetryTime(retryTime - 1);
				ret = connectSynchronous(request);
			} else {
				++mSwitchUriTimes;
				if (mSwitchUriTimes < request.getAllUrl().size()) {
					// 切换网络地址重连
					ret = connectSynchronous(request);
				} else {
					if (e instanceof SocketTimeoutException) {
						throw new NetException(
								IHttpException.SocketTimeoutException);
					} else if (e instanceof ConnectTimeoutException) {
						throw new NetException(
								IHttpException.ConnectTimeoutException);
					} else {
						throw e;
					}
				}
			}
		} finally {
			// 长连接不需要关闭
			// shutDown(client);
			// 不需要设置当前使用地址为空,因为心跳包地址需要使用最终使用的地址
			// request.setCurrentUrl(null);
			// 重置当前的切换Uri地址次数,由于端口是复用的
			mSwitchUriTimes = 0;
		}
		NetLog.info("testBattery, end HttpConnector connectSynchronous url = "
				+ request.getUrl(), null);
		return ret;
	}

	/**
	 * 配置http头信息
	 * 
	 * @param httpRequest
	 *            待配置的請求對象
	 */
	private void configHttpHeader(HttpRequestBase httpRequest) {
		List<Header> list = mRequest.getHeader();
		if (list != null) {
			int listSize = list.size();
			for (int i = 0; i < listSize; i++) {
				httpRequest.addHeader(list.get(i));
			}
		}
	}

	/**
	 * 配置长连接协议头
	 * 
	 * @param httpRequest
	 */
	private void configAliveHeader(HttpRequestBase httpRequest) {
		httpRequest.addHeader("Connection", "keep-alive"); // 设置长连接
	}

	/**
	 * 根据当前联网失败的次数选择网络地址
	 * 
	 * @param request
	 * @return
	 */
	private URI selectURI(THttpRequest request) {
		if (request == null) {
			return null;
		}
		if (mSwitchUriTimes == 0 && request.getAllUrl() != null
				&& mSwitchUriTimes < request.getAllUrl().size()) {
			return request.getAllUrl().get(mSwitchUriTimes);
		}
		return null;
	}
}
