package com.gau.utils.net.request;

import java.net.URI;
import java.net.URISyntaxException;
import java.util.ArrayList;
import java.util.HashMap;
import java.util.List;

import org.apache.http.Header;
import org.apache.http.message.BasicHeader;

import com.gau.utils.net.IConnectListener;
import com.gau.utils.net.INetRecord;
import com.gau.utils.net.asrFiltier.IAsrFilter;
import com.gau.utils.net.operator.ByteArrayHttpOperatror;
import com.gau.utils.net.operator.IHttpOperator;
import com.gau.utils.net.util.HeartSetting;

/**
 * Http请求对象
 * 
 * @author guoshengliang
 * @version 1.0
 */
public class THttpRequest {
	/**
	 * URL列表,用于跟URL区分,第一个为首选URL,其余为备用URL地址
	 */
	private List<URI> mUrlList;
	
	/**
	 * 当前正在连接的url地址
	 */
	private URI mCurrentUrl;
	
	/**
	 * 心跳包Url地址,仅在使用长连接时有效
	 */
	private URI mHeartUrl;

	/**
	 * Post数据
	 */
	private byte[] mPostData;
	
	/**
	 *post parma 
	 */
	private HashMap<String, String> mParamMap;
	
	/**
	 * 默认的重连次数
	 */
	private static final int DEFAULT_RETRY_TIME = 2;

	/**
	 * 重连次数
	 */
	private int mRetryTime = DEFAULT_RETRY_TIME;

	/**
	 * 当前重连次数
	 */
	private int mCurRetryTime = -1;

	/**
	 * 请求类型，方便子类扩充
	 */
	public int mRequestType = 0;

	/**
	 * 是否被取消
	 */
	private boolean mIsCanceled = false;

	/**
	 * 网络记录,用与记录网络连接等信息
	 */
	private INetRecord mNetRecord;

	/**
	 * http请求头信息
	 */
	List<Header> mHeaderStore;
	
	/**
	 * 该连接的观察者
	 */
	IConnectListener mReceiver;
	
	/**
	 * 默认的从socket读数据时发生阻塞的超时时间
	 */
	private static final int DEFAULT_SO_TIMEOUT_VALUE = 30 * 1000;
	
	/**
	 * 默认的连接的超时时间
	 */
	private static final int DEFAULT_CONNECTION_TIMEOUT = 15 * 1000;

	/**
	 * 超时时间
	 */
	int mTimeoutValue = DEFAULT_CONNECTION_TIMEOUT;

	int mSocketTimeoutValue = DEFAULT_SO_TIMEOUT_VALUE;

	/**
	 * httpResponse处理器，如果设置为null默认是byteArrayHttpOperator
	 */
	IHttpOperator mOperator;

	/**
	 * 运营商数据过滤器
	 */
	IAsrFilter mAsrFilter;
	private final int DEFAULT_THREAD_PRIORITY = Thread.NORM_PRIORITY + 3;
	private int mRequestPriority = Thread.NORM_PRIORITY;
	
	/**
	 * 是否使用异步连接,默认为true
	 */
	private boolean mIsAsync = true;
	
	/**
	 * 是否使用长连接,默认为false
	 */
	private boolean mIsAlive = false;
	
	/**
	 * 是否必须使用长连接
	 */
	private boolean mIsMustAlive = false;
	
	/**
	 * 心跳包的发送时间间隔,仅在支持长连接的时候有效
	 */
	private long mHeartTime = HeartSetting.HEART_TIME_NO_SETTING;

	public static final int PROTOCOL_NOT_DEFINE = -1;
	public static final int PROTOCOL_GET = 0;
	public static final int PROTOCOL_POST = 1;
	

	private int mProtocol = PROTOCOL_NOT_DEFINE;
	
	/**
	 * 构造函数
	 * 
	 * @param url
	 *            请求Url地址
	 * @param postData
	 *            post数据
	 * @param receiver
	 *            请求监控者
	 * @throws IllegalArgumentException
	 *             如果url为空或receiver为空抛出
	 * @throws URISyntaxException
	 *             如果傳入Url不符合URI格式抛出
	 */
	public THttpRequest(String url, byte[] postData, IConnectListener receiver)
			throws IllegalArgumentException, URISyntaxException {
		if (url == null) {
			throw new IllegalArgumentException("url==null");
		}
		if (receiver == null) {
			throw new IllegalArgumentException("receiver==null");
		}
		setUrl(url);
		this.mPostData = postData;
		this.mReceiver = receiver;
	}
	
	/**
	 * 构造函数
	 * 
	 * @param url
	 *            请求Url地址
	 * @param receiver
	 *            请求监控者
	 * @throws IllegalArgumentException
	 *             如果url为空或receiver为空抛出
	 * @throws URISyntaxException
	 *             如果傳入Url不符合URI格式抛出
	 */
	public THttpRequest(String url,IConnectListener receiver)
			throws IllegalArgumentException, URISyntaxException {
		this(url, null, receiver);
	}
	
	/**
	 * 添加备用网络连接地址 如果用户添加了备用地址,在首选地址连接失败时,会自动切换到备用地址
	 * 注意如果添加备用地址过多并且当前网络不通畅的时候,可能会导致联网时间过长,建议不要添加过多的备用地址
	 * 
	 * @param url
	 * @throws URISyntaxException
	 */
	public void addAlternateUrl(String url) throws URISyntaxException {
		if (mUrlList == null) {
			mUrlList = new ArrayList<URI>();
		}
		URI alternateUri = new URI(url);
		mUrlList.add(alternateUri);
	}

	public List<URI> getAlternateUrl() {
		if (mUrlList != null && mUrlList.size() > 1) {
			List<URI> alternateUrl = mUrlList.subList(1, mUrlList.size());
			return alternateUrl;
		}
		return null;
	}
	
	/**
	 * 设置心跳包的Url地址,仅在使用长连接时有效
	 * @param heartUrl
	 * @throws URISyntaxException 
	 */
	public void setHeartUrl(String heartUrl) throws URISyntaxException {
		mHeartUrl = new URI(heartUrl);		
	}
	
	/**
	 * 获取心跳包的Url地址
	 * @return
	 */
	public URI getHeartUrl() {
		return mHeartUrl;
	}

	/**
	 * 获取全部联网地址,包括首选url和备用url,不包括心跳包url
	 * @return
	 */
	public List<URI> getAllUrl() {
		return mUrlList;
	}
	
	/**
	 * 获取当前正在联网的url地址
	 * @return
	 */
	public URI getCurrentUrl() {
		return mCurrentUrl;
	}
	
	/**
	 * 设置正在联网的url地址
	 * @param url
	 */
	public void setCurrentUrl(URI url) {
		mCurrentUrl = url;
	}

	public IAsrFilter getAsrFilter() {
		return mAsrFilter;
	}

	public void setAsrFilter(IAsrFilter asrFilter) {
		this.mAsrFilter = asrFilter;
	}

	public IHttpOperator getOperator() {
		if (mOperator == null) {
			mOperator = new ByteArrayHttpOperatror();
		}
		return mOperator;
	}

	public void setOperator(IHttpOperator operator) {
		this.mOperator = operator;
	}

	public int getCurRetryTime() {
		if (mCurRetryTime == -1) {
			mCurRetryTime = mRetryTime;
		}
		return mCurRetryTime;
	}

	public void setCurRetryTime(int curRetryTime) {
		this.mCurRetryTime = curRetryTime;
	}

	public int getRetryTime() {
		return mRetryTime;
	}

	public void setRetryTime(int retryTime) {
		this.mRetryTime = retryTime;
	}

	public int getTimeoutValue() {
		return mTimeoutValue;
	}

	public void setTimeoutValue(int timeoutValue) {
		this.mTimeoutValue = timeoutValue;
	}

	public int getSocketTimeoutValue() {
		return mSocketTimeoutValue;
	}

	public void setSocketTimeoutValue(int socketTimeoutValue) {
		this.mSocketTimeoutValue = socketTimeoutValue;
	}

	public IConnectListener getReceiver() {
		return mReceiver;
	}

	public void setReceiver(IConnectListener receiver) {
		this.mReceiver = receiver;
	}

	public URI getUrl() {
		if (mUrlList != null && mUrlList.size() > 0) {
			return mUrlList.get(0);
		}
		return null;
	}

	/**
	 * 设置首选Url地址
	 * 
	 * @param url
	 * @throws URISyntaxException
	 */
	public void setUrl(String url) throws URISyntaxException {
		if (url == null) {
			return;
		}
		URI uri = new URI(url);
		if (mUrlList == null) {
			mUrlList = new ArrayList<URI>();
		}
		if (mUrlList.isEmpty()) {
			mUrlList.add(uri);
		} else {
			mUrlList.set(0, uri);
		}
	}

	public byte[] getPostData() {
		return mPostData;
	}

	public void setPostData(byte[] postData) {
		this.mPostData = postData;
	}
	
	/**
	 * 设置是否使用异步连接,默认为true
	 * 注意如果使用为长连接,则设置同步连接无效,强制使用异步连接,因为长连接需要使用心跳包维持连接
	 * @param isAsync
	 */
	public void setIsAsync(boolean isAsync) {
		mIsAsync = isAsync;
	}
	
	/**
	 * 获取是否使用异步连接,默认为true
	 * @return
	 */
	public boolean getIsAsync() {
		return mIsAsync;
	}
	
	/** <br>功能简述:设置是否使用长连接,默认为false
	 * <br>功能详细描述:
	 * <br>注意:如果你使用长连接,则必须使用异步接口,同步接口无法使用;
	 * 另外,为了保持并发,就算你设置该属性为true,也并不一定使用的是长连接
	 * 如果要确保使用的一定是长连接,使用
	 * @param isAlive
	 */
	public void setIsKeepAlive(boolean isAlive) {
		mIsAlive = isAlive;
	}
	
	/** <br>功能简述:是否肯定使用长连接
	 * <br>功能详细描述:如果该设置为true,则setIsAlive的结果也必定为true
	 * <br>注意:
	 * @param isAlwaysAlive
	 */
	public void setIsMustAlive(boolean isAlwaysAlive) {
		mIsMustAlive = isAlwaysAlive;
	}
	
	public boolean getIsMustAlive() {
		return mIsMustAlive;
	}
	
	/**
	 * 获取是否使用长连接,默认为false
	 * @return
	 */
	public boolean getIsKeepAlive() {
		return mIsAlive;
	}

	public void addHeader(String name, String value) {
		if (name == null || value == null) {
			throw new IllegalArgumentException(
					"AddHeader must have name and value");
		}
		BasicHeader header = new BasicHeader(name, value);
		if (mHeaderStore == null) {
			mHeaderStore = new ArrayList<Header>();
		}
		mHeaderStore.add(header);
	}

	public List<Header> getHeader() {
		return mHeaderStore;
	}

	public synchronized boolean isCanceled() {
		return mIsCanceled;
	}

	public synchronized void setCanceled(boolean canceled) {
		mIsCanceled = canceled;
	}

	public THttpRequest clone() {
		return null;
	}

	public int getRequestPriority() {
		return mRequestPriority;
	}

	public void setRequestPriority(int requestPriority) {
		if (requestPriority >= Thread.MIN_PRIORITY
				&& requestPriority <= Thread.MAX_PRIORITY) {
			mRequestPriority = requestPriority;
		}			
		else {
			mRequestPriority = Thread.NORM_PRIORITY;
		}			
	}

	/**
	 * 设置网络进程记录回调
	 * 
	 * @author liguoliang
	 * @param record
	 */
	public void setNetRecord(INetRecord record) {
		mNetRecord = record;
	}

	public INetRecord getNetRecord() {
		return mNetRecord;
	}
	
	/**
	 * 设置心跳包的发送时间间隔,仅在长连接的时候有效
	 * @param heartTime
	 */
	public void setHeartTime(long heartTime) {
		mHeartTime = heartTime;
	}
	
	/**
	 * 获取长连接的发送时间间隔
	 * @return
	 */
	public long getHeartTime() {
		return mHeartTime;
	}
	
	public void cleanup() {
		if (mUrlList != null) {
			mUrlList.clear();
			mUrlList = null;
		}
		if (mPostData != null) {
			mPostData = null;
		}
		if (mHeaderStore != null) {
			mHeaderStore.clear();
			mHeaderStore = null;
		}
		if (mNetRecord != null) {
			mNetRecord = null;
		}
		if (mReceiver != null) {
			mReceiver = null;
		}
		if (mAsrFilter != null) {
			mAsrFilter = null;
		}
		if (mOperator != null) {
			mOperator = null;
		}
	}	
	
	public void setParamMap(HashMap<String, String> mParamMap) {
		this.mParamMap = mParamMap;
	}

	public HashMap<String, String> getParamMap() {
		return mParamMap;
	}

	public void setProtocol(int type) {
		mProtocol = type;
	}

	public int getProtocol() {
		return mProtocol;
	}
}
