package com.jb.ga0.commerce.util;

import android.app.AlarmManager;
import android.app.PendingIntent;
import android.content.BroadcastReceiver;
import android.content.Context;
import android.content.Intent;
import android.content.IntentFilter;
import android.content.SharedPreferences;
import android.util.SparseArray;

/**
 * 自定义闹钟，支持同时设置多个闹钟
 *
 * @author matt
 * @date: 2015年1月30日
 */
public class CustomAlarm {
	/**无效闹钟ID*/
	public static final int ALARMID_INVALID = -1;
	/**
	 * 闹钟ID的key
	 */
	private static final String KEY_ALARMID = "alarmId";
	
	private Context mContext;
	private AlarmManager mAlarmManager = null;
	private TaskReceiver mReceiver = null;
	/**
	 * 闹钟Action
	 */
	private String mAlarmAction;
	
	private SparseArray<AlarmInfo> mPendingArray = new SparseArray<AlarmInfo>();

	public CustomAlarm(Context context, String alarmAction) {
		context = context.getApplicationContext();
		mContext = context;
		mAlarmManager = (AlarmManager) context.getSystemService(Context.ALARM_SERVICE);
		mReceiver = new TaskReceiver();
		mAlarmAction = alarmAction;
		
		registerReceiver();
	}
	
    /**
     * 注册广播监听器
     */
    private void registerReceiver() {
        //开启广播接收服务
        IntentFilter filter = new IntentFilter();
        filter.addAction(mAlarmAction);
		try {
			mContext.registerReceiver(mReceiver, filter);
		} catch (Exception e) {
			e.printStackTrace();
		}
	}
    
    /**
     * 清除所有已设置闹钟。
     */
    public void clear() {
        //取消所有闹钟
        for (int i = 0; i < mPendingArray.size(); ++i) {
        	AlarmInfo alarmInfo = mPendingArray.valueAt(i);
			mAlarmManager.cancel(alarmInfo.mPendingIntent);
        }
    }
    
    /**
     * 延时triggerInterval后启动，一次性闹钟
     * @param alarmId 闹钟ID， 外部自己定义，要保证不重复且>0
     * @param triggerDelay 延时
     * @param isWakeup 手机睡眠时是否仍唤醒执行闹钟
     * @param listener 闹钟监听
     */
    public void alarmOneTime(int alarmId, long triggerDelay, boolean isWakeup, OnAlarmListener listener) {
    	LogUtils.i("matt", String.format("[CustomAlarm::alarm] alarmId:%d, triggerInterval:%d", alarmId, triggerDelay));
    	if (null == listener) {
    		throw new IllegalArgumentException("参数错误！listener不能为null");
    	}
    	
        try {
        	Intent intent = new Intent(mAlarmAction);
        	intent.putExtra(KEY_ALARMID, alarmId);
        	PendingIntent pendingIntent = PendingIntent.getBroadcast(mContext, alarmId, intent, PendingIntent.FLAG_UPDATE_CURRENT);
        	
            long triggerAtTime = System.currentTimeMillis() + triggerDelay;
            int alarmType = isWakeup ? AlarmManager.RTC_WAKEUP : AlarmManager.RTC;
            mAlarmManager.set(alarmType, triggerAtTime, pendingIntent);
            
            mPendingArray.put(alarmId, new AlarmInfo(pendingIntent, listener, false, 0, 0));
        } catch (Exception e) {
            e.printStackTrace();
        }
    }
    
    /**
     * 延时triggerInterval后启动，周期性闹钟
     * @param alarmId 闹钟ID， 外部自己定义，要保证不重复且>0
     * @param triggerDelay 第一次启动延时
     * @param interval 周期时长
     * @param isWakeup 手机睡眠时是否仍唤醒执行闹钟
     * @param listener 闹钟监听
     */
    public void alarmRepeat(int alarmId, long triggerDelay, long interval, boolean isWakeup, OnAlarmListener listener) {
    	LogUtils.i("matt", String.format("[CustomAlarm::alarmRepeat] alarmId:%d, triggerInterval:%d<>interval:%d", alarmId, triggerDelay, interval));
    	if (null == listener) {
    		throw new IllegalArgumentException("参数错误！listener不能为null");
    	}

        try {
        	Intent intent = new Intent(mAlarmAction);
        	intent.putExtra(KEY_ALARMID, alarmId);
        	PendingIntent pendingIntent = PendingIntent.getBroadcast(mContext, alarmId, intent, PendingIntent.FLAG_UPDATE_CURRENT);
        	
            long triggerAtTime = System.currentTimeMillis() + triggerDelay;
            int alarmType = isWakeup ? AlarmManager.RTC_WAKEUP : AlarmManager.RTC;
//            mAlarmManager.setRepeating(alarmType, triggerAtTime, interval, pendingIntent);
            mAlarmManager.set(alarmType, triggerAtTime, pendingIntent);
            
            mPendingArray.put(alarmId, new AlarmInfo(pendingIntent, listener, true, alarmType, interval));
        } catch (Exception e) {
            e.printStackTrace();
        }
    }

	/**
	 * 延时triggerInterval后启动，周期性闹钟.
	 * <br/>此方法会缓存每次闹钟执行的时间到sharepreference，以sp的时间作为计时的起点。第n>1次闹钟，忽视原triggerDelay。
	 * <br/>如一个闹钟，interval为8小时，1点闹钟触发后，进程被杀死，6点应用重启又再设置同一个(id)闹钟，则9点闹钟即被触发。
	 * <br/>要与 {@link #saveTriggerTime}方法配套使用
	 * @param alarmId 闹钟ID， 外部自己定义，要保证不重复且>0
	 * @param triggerDelay 第一次启动延时
	 * @param interval 周期时长
	 * @param isWakeup 手机睡眠时是否仍唤醒执行闹钟
	 * @param listener 闹钟监听
	 */
	public void alarmRepeatInRealTime(SharedPreferences sp, int alarmId, long triggerDelay, long interval, boolean isWakeup, OnAlarmListener listener) {
		LogUtils.i("matt", String.format("[CustomAlarm::alarmRepeatInRealTime] alarmId:%d, triggerInterval:%d<>interval:%d", alarmId, triggerDelay, interval));

		// 获取上次触发闹钟的时间
		long lastTime = sp.getLong(getLastTimeKey(alarmId), -1);
		if (lastTime != -1) {
			//已经执行过闹钟，计算triggerDelay
			long passTime = System.currentTimeMillis() - lastTime;
			triggerDelay = interval - passTime;
			//已经超时，需要即刻触发闹钟
			if (triggerDelay < 0) {
				triggerDelay = 0;
			}
		}

		alarmRepeat(alarmId, triggerDelay, interval, isWakeup, listener);
	}

	/**
	 * 保存闹钟触发时间
	 * <br/>要与 {@link #alarmRepeatInRealTime}方法配套使用
	 */
	public void saveTriggerTime(SharedPreferences sp, int alarmId) {
		sp.edit().putLong(getLastTimeKey(alarmId), System.currentTimeMillis()).commit();
	}

	private String getLastTimeKey(int alarmId) {
		return mAlarmAction.hashCode() + "_" + alarmId;
	}
    
    /**
     * 取消闹钟
     * @param alarmId 闹钟ID， 外部自己定义，要保证不重复且>0
     */
    public void cancelAarm(int alarmId) {
    	AlarmInfo alarmInfo = mPendingArray.get(alarmId);
    	if (null == alarmInfo) {
    		return;
    	}
    	mAlarmManager.cancel(alarmInfo.mPendingIntent);
    	mPendingArray.remove(alarmId);
    }
    
    public boolean isAlarmActive(int alarmId) {
    	return mPendingArray.get(alarmId) != null;
    }
    
	/**
	 * 任务广播接受者
	 * @author matt
	 *
	 */
	private class TaskReceiver extends BroadcastReceiver {
		@Override
		public void onReceive(Context context, Intent intent) {
			final String action = intent.getAction();
			if (!mAlarmAction.equals(action)) {
				return;
			}
			int alarmId = intent.getIntExtra(KEY_ALARMID, ALARMID_INVALID);
//			if (LogUtils.sIsLog) {
//				LogUtils.i("matt", "[CustomAlarm.TaskReceiver::onReceive] --->alarmId:" + alarmId);
//			}
			AlarmInfo alarmInfo = mPendingArray.get(alarmId);
			if (null == alarmInfo) {
				return;
			}
			if (!alarmInfo.mIsRepeat) { //非重复闹钟，只使用一次
				mPendingArray.remove(alarmId);
			} else {
				try {
					mAlarmManager.set(alarmInfo.mAlarmType, System.currentTimeMillis() + alarmInfo.mRepeatInterval, alarmInfo.mPendingIntent);
				} catch (Exception e) {
				}
			}
			alarmInfo.mOnAlarmListener.onAlarm(alarmId);
		}
	}
	
	/**
	 * 闹钟信息
	 *
	 * @author matt
	 * @date: 2015年1月30日
	 *
	 */
	private class AlarmInfo {
		PendingIntent mPendingIntent;
		OnAlarmListener mOnAlarmListener;
		boolean mIsRepeat;
		int mAlarmType;
		long mRepeatInterval;
		
		AlarmInfo(PendingIntent pendingIntent, OnAlarmListener listener, boolean isRepeat, int alarmType, long repeatInterval) {
			mPendingIntent = pendingIntent;
			mOnAlarmListener = listener;
			mIsRepeat = isRepeat;
			mAlarmType = alarmType;
			mRepeatInterval = repeatInterval;
		}
	}
	
	/**
	 * 闹钟监听
	 *
	 * @author matt
	 * @date: 2015年1月30日
	 *
	 */
	public static interface OnAlarmListener {
		/**
		 * 此方法在主线程回调
		 * @param alarmId
		 */
		void onAlarm(int alarmId);
	}
}
