package com.jb.ga0.commerce.util.encrypt;

import java.util.Map;

import com.jb.ga0.commerce.util.LogUtils;
import com.jb.ga0.commerce.util.io.MultiprocessSharedPreferences;

import android.annotation.SuppressLint;
import android.content.Context;
import android.content.SharedPreferences;
import android.content.SharedPreferences.Editor;
import android.content.SharedPreferences.OnSharedPreferenceChangeListener;

/**
 * 带加密功能的PreferencesManager, 支持默认秘钥或自定义秘钥
 *
 * @author matt
 * @date: 2015年2月26日
 *
 */
public final class CryptPreferencesManager {
	//默认公钥，必须是常量
	private static final String CRYPT_KEY = "NaubrwWEGiJEQqRxx7aXntbGOf4YiRmW0WY9043rcqRhJreE4sReMC1OFRaeI7TXWBJUiJQGpwA1UdSsR65vvNieo70IUqvUnj1mn1mLUTKEMqeM9l5g90WJJo4gBN3n";
	private SharedPreferences mPreferences;
	private Editor mEditor;

	/**
	 * 构造函数
	 * @param context
	 * @param name
	 * @param mode 详见 {@link Context#MODE_PRIVATE}等
	 */
	@SuppressLint("CommitPrefEdits")
	public CryptPreferencesManager(Context context, String name, int mode) {
		mPreferences = context.getSharedPreferences(name, mode);
		mEditor = mPreferences.edit();
	}
	
	/**
	 * 构造函数，可配置SharedPreferences，如使用{@link MultiprocessSharedPreferences}
	 * @param sp
	 */
	public CryptPreferencesManager(SharedPreferences sp) {
		mPreferences = sp;
		mEditor = sp.edit();
	}
	
	public SharedPreferences getPreferences() {
		return mPreferences;
	}

	@SuppressLint("CommitPrefEdits")
	public void setPreferences(Context context, int mode, String name) {
		mPreferences = context.getSharedPreferences(name, mode);
		mEditor = mPreferences.edit();
	}

	/**
	 * 清除数据
	 */
	public void clear() {
		if (mEditor != null) {
			mEditor.clear().commit();
		} else if (mPreferences != null) {
			mEditor = mPreferences.edit();
			mEditor.clear().commit();
		}
	}

	public boolean commit() {
		boolean bRet = false;
		if (mEditor != null) {
			bRet = mEditor.commit();
		}
		return bRet;
	}

	public void remove(String key) {
		mPreferences.edit().remove(key).commit();
	}

	public Map<String, ?> getAll() {
		return mPreferences.getAll();
	}

	public boolean contains(String key) {
		return mPreferences.contains(key);
	}
	
	public void putString(String key, String s) {
		putString(key, s, CRYPT_KEY);
	}
	
	/**
	 * 
	 * @param key
	 * @param s
	 * @param cryptKey 自定义秘钥
	 */
	public void putString(String key, String s, String cryptKey) {
		s = encrypt(cryptKey, s);
		if (mEditor != null) {
			mEditor.putString(key, s);
		}
	}
	
	public String getString(String key, String defValue) {
		return getString(key, defValue, CRYPT_KEY);
	}
	
	/**
	 * 
	 * @param key
	 * @param defValue
	 * @param cryptKey 自定义秘钥
	 * @return
	 */
	public String getString(String key, String defValue, String cryptKey) {
		String encrypted = null;
		if (mPreferences != null) {
			encrypted =  mPreferences.getString(key, null);
		}
		if (encrypted != null) {
			return decrypt(cryptKey, encrypted);
		}
		return defValue;
	}
	
	public void putLong(String key, long l) {
		putLong(key, l, CRYPT_KEY);
	}
	
	/**
	 * 
	 * @param key
	 * @param l
	 * @param cryptKey 自定义秘钥
	 */
	public void putLong(String key, long l, String cryptKey) {
		putString(key, String.valueOf(l), cryptKey);
	}
	
	public long getLong(String key, long defValue) {
		return getLong(key, defValue, CRYPT_KEY);
	}
	
	/**
	 * 
	 * @param key
	 * @param defValue
	 * @param cryptKey 自定义秘钥
	 * @return
	 */
	public long getLong(String key, long defValue, String cryptKey) {
		String valueString = null;
		if (mPreferences != null) {
			valueString = getString(key, null, cryptKey);
		}
		if (valueString != null) {
			try {
				return Long.parseLong(valueString);
			} catch (Exception e) {
			}
		}
		return defValue;
	}
	
	public void putInt(String key, int i) {
		putInt(key, i, CRYPT_KEY);
	}
	
	/**
	 * 
	 * @param key
	 * @param i
	 * @param cryptKey 自定义秘钥
	 */
	public void putInt(String key, int i, String cryptKey) {
		putString(key, String.valueOf(i), cryptKey);
	}
	
	public int getInt(String key, int defValue) {
		return getInt(key, defValue, CRYPT_KEY);
	}
	
	/**
	 * 
	 * @param key
	 * @param defValue
	 * @param cryptKey 自定义秘钥
	 * @return
	 */
	public int getInt(String key, int defValue, String cryptKey) {
		String valueString = null;
		if (mPreferences != null) {
			valueString = getString(key, null, cryptKey);
		}
		if (valueString != null) {
			try {
				return Integer.parseInt(valueString);
			} catch (Exception e) {
			}
		}
		return defValue;
	}
	
	public void putBoolean(String key, boolean b) {
		putBoolean(key, b, CRYPT_KEY);
	}
	
	/**
	 * 
	 * @param key
	 * @param b
	 * @param cryptKey 自定义秘钥
	 */
	public void putBoolean(String key, boolean b, String cryptKey) {
		putString(key, String.valueOf(b), cryptKey);
	}

	public boolean getBoolean(String key, boolean defValue) {
		return getBoolean(key, defValue, CRYPT_KEY);
	}
	
	/**
	 * 
	 * @param key
	 * @param defValue
	 * @param cryptKey 自定义秘钥
	 * @return
	 */
	public boolean getBoolean(String key, boolean defValue, String cryptKey) {
		String valueString = null;
		if (mPreferences != null) {
			valueString = getString(key, null, cryptKey);
		}
		if (valueString != null) {
			try {
				return Boolean.parseBoolean(valueString);
			} catch (Exception e) {
			}
		}
		return defValue;
	}
	
	/**
	 * 
	 * @param key
	 * @param f
	 * @备注 有精度丢失，(-)3.1415916写入再读出会变成(-)3.141592，系统接口也有同样问题  
	 */
	public void putFloat(String key, float f) {
		putFloat(key, f, CRYPT_KEY);
	}
	
	/**
	 * 
	 * @param key
	 * @param f
	 * @param cryptKey 自定义秘钥
	 * @备注 有精度丢失，(-)3.1415916写入再读出会变成(-)3.141592，系统接口也有同样问题  
	 */
	public void putFloat(String key, float f, String cryptKey) {
		putString(key, String.valueOf(f), cryptKey);
	}

	/**
	 * 
	 * @param key
	 * @param defValue
	 * @return
	 * @备注 有精度丢失，(-)3.1415916写入再读出会变成(-)3.141592，系统接口也有同样问题  
	 */
	public float getFloat(String key, float defValue) {
		return getFloat(key, defValue, CRYPT_KEY);
	}
	
	/**
	 * 
	 * @param key
	 * @param defValue
	 * @param cryptKey 自定义秘钥
	 * @return
	 * @备注 有精度丢失，(-)3.1415916写入再读出会变成(-)3.141592，系统接口也有同样问题  
	 */
	public float getFloat(String key, float defValue, String cryptKey) {
		String valueString = null;
		if (mPreferences != null) {
			valueString = getString(key, null, cryptKey);
		}
		if (valueString != null) {
			try {
				return Float.parseFloat(valueString);
			} catch (Exception e) {
			}
		}
		return defValue;
	}

	public void registerOnSharedPreferenceChangeListener(OnSharedPreferenceChangeListener listener) {
		mPreferences.registerOnSharedPreferenceChangeListener(listener);
	}

	public void unregisterOnSharedPreferenceChangeListener(OnSharedPreferenceChangeListener listener) {
		mPreferences.unregisterOnSharedPreferenceChangeListener(listener);
	}
	
	public String encrypt(String seed, String value) {
		if (null == value) {
			return null;
		}
		try {
			return AESCrypt.encrypt(seed, value);
		} catch (Exception e) {
			LogUtils.e("matt", LogUtils.getStackTraceString(e));
		}
		return value;
	}
	
	public String decrypt(String seed, String encryptedValue) {
		if (null == encryptedValue) {
			return null;
		}
		try {
			return AESCrypt.decrypt(seed, encryptedValue);
		} catch (Exception e) {
			LogUtils.e("matt", LogUtils.getStackTraceString(e));
		}
		return encryptedValue;
	}
}
