package com.jb.ga0.commerce.util.http;

import java.io.UnsupportedEncodingException;
import java.net.URLEncoder;
import java.util.HashMap;
import java.util.Map;

import org.apache.http.HttpResponse;
import org.json.JSONException;
import org.json.JSONObject;

import com.gau.go.gostaticsdk.StatisticsManager;

import android.content.Context;
import android.os.Build;

/**
 * go http-post协议处理器
 *
 * @author matt
 * @date: 2014年12月4日
 * @备注： go协议请求格式为 {"data":{"phead":?, ...业务协议键值对}, "handle":?, "shandle":?, "pkey":?, "sign":?}
 *           响应格式为  {"result":{"status":?, "errorcode":?, "msg":?}, ...业务协议键值对 }
 */
public class GoHttpPostHandlerForNet {
	/** ====go协议==请求参数key====BEGIN======= */
	/**
	 * 提交的json格式请求数据。
	 * @备注：
	 * =>默认：json.toString(); 
	 * =>gzip压缩: json.toString().getBytes("UTF-8")；然后对bytes进行压缩处理；最后对压缩结果bytes采用ISO-8859-1编码成字符串。
	 * =>BASE64加密: json.toString().getBytes("UTF-8")；然后对bytes进行BASE64加密处理；最后对加密结果bytes采用ISO-8859-1编码成字符串。
	 */
	protected static final String KEY_PARAM_DATA = "data";
	/**
	 * 对提交的表单中data字段进行处理， 可选值详见 {@link #VALUE_HANDLE_NONE}
	 * @备注：
	 * 默认0：无处理；
	 * 1：采用gzip压缩（请求上传的数据量比较大（大于10k）时使用）；
	 * 2：BASE64加密（涉及到采集用户本地数据的时候使用）；
	 */
	protected static final String KEY_PARAM_HANDLE = "handle";
	/**
	 * 
	 */
	protected static final String KEY_PARAM_SHANDLE = "shandle";
	/**
	 * 通信校验key
	 * @备注： 必须，由接口提供方提供
	 */
	protected static final String KEY_PARAM_PKEY = "pkey";
	/**
	 * 数据校验码
	 * @备注：
	 * 必须，String sign = md5((私钥+data数据+私钥).getBytes(“utf-8”));
	 * 私钥由接口提供方提供
	 */
	protected static final String KEY_PARAM_SIGN = "sign";
	/**
	 * 备用的funid
	 */
	protected static final String KEY_PARAM_PFUNID = "pfunid";
	
	
	/** ====go协议==请求参数key====END=======*/
	
	/** ====go协议==数据处理方式====BEGIN=======*/
	public static final int VALUE_HANDLE_NONE = 0;
	public static final int VALUE_HANDLE_GZIP = 1;
	@Deprecated //暂不支持
	public static final int VALUE_HANDLE_BASE64 = 2;
	/** ====go协议==数据处理方式====END=======*/
	
	/** ====go协议==返回数据格式====BEGIN======= */
	/**
	 * response响应头信息
	 */
	protected static final String TAG_RESULT = "result";		
	/**
	 * 响应头信息result中的字段：服务器处理结果， 可选值详见 {@link #RESULT_STATUS_OK} 等
	 * status=1:处理成功；
	 * status=-1:服务器处理出错； 
	 * status=-2:业务处理异常；  
	 * 备注：仅当status=1时客户端才解析数据。
	 */
	protected static final String TAG_STATUS = "status";		
	protected static final String TAG_ERRORCODE = "errorcode";	//响应头信息result中的字段：错误码
	protected static final String TAG_MSG = "msg";				//响应头信息result中的字段：错误信息
	/** ====go协议==返回数据格式====END=======*/
	
	/** ====go协议==服务器处理结果定义====BEGIN=======*/
	public static final int RESULT_STATUS_OK = 1; //服务器处理成功
	public static final int RESULT_STATUS_ERR_SERVICE = -1; //服务器处理出错
	public static final int RESULT_STATUS_ERR_BUSSINESS = -2; //业务处理异常
	/** ====go协议==服务器处理结果定义====END=======*/
	
	protected Context mContext;
	
	/**
	 * 
	 * @param context
	 * @param httpManager 这个参数要用全局的变量或单例，减少线程调用
	 * @param protocolVersion
	 */
	public GoHttpPostHandlerForNet(Context context) {
		mContext = context;
	}
	
	/**
	 * 构建请求url
	 * @param domainName 服务器域名, 形如"http://域名/common?funid="
	 * @param funid 业务协议约定的功能id
	 * @return 格式为 "http://域名/common?funid=1&rd=系统时间毫秒数"
	 */
	public String createUrl(String domainName, int funid) {
//		return domainName + "/common?funid=" + funid + "&rd=" + System.currentTimeMillis();
		return domainName + funid + "&rd=" + System.currentTimeMillis();
	}
	
	/**
	 * 构建post参数表, 不加密，上传数据不压缩，下发数据不压缩
	 * @param bussinessParams 业务参数(业务协议键值对), 
	 *        其中的每一个键值对<String, Object>，都是与go协议头"phead"并行的
	 * @param isZipPostData 是否压缩上传的数据
	 * @param isRequestReturnDataZiped 是否要求服务器压缩下发的数据
	 * @return
	 */
	public HashMap<String, String> createPostParams(Map<String, Object> bussinessParams, String funid) {
		return createPostParams(bussinessParams, VALUE_HANDLE_NONE, VALUE_HANDLE_NONE, " ", " ", funid);
	}
	
	/**
	 * 构建post数据, 不加密，上传数据不压缩，下发数据不压缩
	 * @param bussinessParams 业务参数(业务协议键值对), 
	 *        其中的每一个键值对<String, Object>，都是与go协议头"phead"并行的
	 * @param isZipPostData 是否压缩上传的数据
	 * @param isRequestReturnDataZiped 是否要求服务器压缩下发的数据
	 * @return
	 * @备注 使用这个接口，必须设置-->httpRequest.addHeader("Content-Type", "application/x-www-form-urlencoded");
	 */
	public String createPostData(Map<String, Object> bussinessParams, String funid) {
		return createPostData(bussinessParams, VALUE_HANDLE_NONE, VALUE_HANDLE_NONE, " ", " ", funid);
	}
	
	/**
	 * 构建post参数表, 不加密
	 * @param bussinessParams 业务参数(业务协议键值对), 
	 *        其中的每一个键值对<String, Object>，都是与go协议头"phead"并行的
	 * @param isZipPostData 是否压缩上传的数据
	 * @param isRequestReturnDataZiped 是否要求服务器压缩下发的数据
	 * @return
	 */
	public HashMap<String, String> createPostParams(
			Map<String, Object> bussinessParams, int handle, int shandle, String funid) {
		return createPostParams(bussinessParams, handle, shandle, " ", " ", funid);
	}
	
	/**
	 * 
	 * 构建post参数表,
	 * @param bussinessParams 业务参数, 可为null
	 *        其中的每一个键值对<String, Object>，都是与go协议头"phead"并行的
	 * @param handle 对上传数据("data"字段value)的处理方式，可选值详见
	 *        {@link #VALUE_HANDLE_NONE}, {@link #VALUE_HANDLE_GZIP}, {@link #VALUE_HANDLE_BASE64}
	 * @param shandle 要求服务器下发数据的处理方式，可选值同@param handle
	 * @param pkey
	 * @param sign
	 * @return
	 */
	public HashMap<String, String> createPostParams(
			Map<String, Object> bussinessParams, int handle, int shandle,
			String pkey, String sign, String pFunid) {
		HashMap<String, String> paramMap = new HashMap<String, String>();
		paramMap.put(KEY_PARAM_DATA, handleDataParam(createDataParam(bussinessParams), handle));
		paramMap.put(KEY_PARAM_HANDLE, "" + handle);
		paramMap.put(KEY_PARAM_SHANDLE, "" + shandle);
		paramMap.put(KEY_PARAM_PKEY, pkey);
		paramMap.put(KEY_PARAM_SIGN, sign);
		paramMap.put(KEY_PARAM_PFUNID, pFunid);
		
		return paramMap;
	}
	
	/**
	 * 
	 * 构建post数据
	 * @param bussinessParams 业务参数, 可为null
	 *        其中的每一个键值对<String, Object>，都是与go协议头"phead"并行的
	 * @param handle 对上传数据("data"字段value)的处理方式，可选值详见
	 *        {@link #VALUE_HANDLE_NONE}, {@link #VALUE_HANDLE_GZIP}, {@link #VALUE_HANDLE_BASE64}
	 * @param shandle 要求服务器下发数据的处理方式，可选值同@param handle
	 * @param pkey
	 * @param sign
	 * @return
	 * @备注 使用这个接口，必须设置-->httpRequest.addHeader("Content-Type", "application/x-www-form-urlencoded");
	 */
	public String createPostData(
			Map<String, Object> bussinessParams, int handle, int shandle,
			String pkey, String sign, String pFunid) {
		String dataParam = handleDataParam(createDataParam(bussinessParams), handle);
		try {
			dataParam = URLEncoder.encode(dataParam, "UTF-8");
			pkey = URLEncoder.encode(pkey, "UTF-8");
			sign = URLEncoder.encode(sign, "UTF-8");
		} catch (UnsupportedEncodingException e) {
			e.printStackTrace();
		}
		
		StringBuffer sb = new StringBuffer();
		sb.append(KEY_PARAM_DATA).append("=").append(dataParam)
		.append("&").append(KEY_PARAM_HANDLE).append("=").append(handle)
		.append("&").append(KEY_PARAM_SHANDLE).append("=").append(shandle)
		.append("&").append(KEY_PARAM_PKEY).append("=").append(pkey)
		.append("&").append(KEY_PARAM_SIGN).append("=").append(sign)
		.append("&").append(KEY_PARAM_PFUNID).append("=").append(pFunid);
		return sb.toString();
	}
	
	/**
	 * 处理上传的"data"字段数据
	 * @param dataJson 
	 * @param handle 处理方式
	 * @return
	 */
	private static String handleDataParam(JSONObject dataJson, int handle) {
		if (null == dataJson) {
			return null;
		}
		switch (handle) {
		case VALUE_HANDLE_GZIP:
			return HttpUtil.gzip(dataJson);
			
		case  VALUE_HANDLE_BASE64:
			//TODO DEBUG
			// return ...
		case VALUE_HANDLE_NONE:
		default:
			return dataJson.toString();
		}
	}
	
	/**
	 * 创建go-post协议数据--"data"数据单元
	 * @param bussinessParams 业务参数, 
	 *        其中的每一个键值对<String, Object>，都是与go协议头"phead"并行的
	 * @return
	 */
	protected JSONObject createDataParam(Map<String, Object> bussinessParams) {
		JSONObject dataJson = new JSONObject();
		try {
			dataJson.put("phead", createHead());
			HttpUtil.putInJSONObject(bussinessParams, dataJson);
		} catch (JSONException e) {
			e.printStackTrace();
		}
		return dataJson;
	}
	
	/**
	 * 创建go协议头--"phead"数据单元
	 * @return
	 */
	protected JSONObject createHead() {
		Context context = mContext;
		JSONObject pHeadJson = new JSONObject();
		if (context != null) {
			try {
				//int  协议版本号  必填 
				pHeadJson.put("pversion", 20); 
				//String 手机androidid 必填 
				pHeadJson.put("aid", GoHttpHeadUtil.getAndroidId(context));
				//string google广告id 必填
//				pHeadJson.put("gadid", ProductConfigManager.getInstance().getGoogleAdId()); 
				//String 手机imei 选填 
				pHeadJson.put("imei", GoHttpHeadUtil.getVirtualIMEI(context)); 
				//String go系列软件用户id go系列软件必填，其他选填
				pHeadJson.put("goid", StatisticsManager.getGOID(context)); 
				//String 用户id 如果项目有特定的用户id则必填，没有则不填 
				pHeadJson.put("uid", GoHttpHeadUtil.getVirtualIMEI(context)); 
				//int 产品id 必填，找服务端分配： 具体见页面：应用分发-基础信息--->数据请求协议中对应关系为：1:极桌面，2:GO桌面，3:客户端，4：输入法，8：KP
				pHeadJson.put("cid", 0); 
				//int 客户端软件版本号  必填，versioncode值例如：1
				pHeadJson.put("cversion", GoHttpHeadUtil.getVersionCode(context)); 
				//String 客户端软件版名称 必填，versionname值例如：2.3 
				pHeadJson.put("cversionname", GoHttpHeadUtil.getVersionName(context)); 
				//int 真实渠道号  必填  
				pHeadJson.put("channel", GoHttpHeadUtil.getUid(context)); 
				
				//String 国家(大写) 必填，例如：CN 
				pHeadJson.put("local", GoHttpHeadUtil.getLocal(context));
				//String 语言（小写）  必填，例如:en 
				pHeadJson.put("lang", GoHttpHeadUtil.getLanguage(context)); 
				//String 运营商编码   必填，例如：460030912121001
				pHeadJson.put("imsi", GoHttpHeadUtil.getCnUser(context)); 
				//String 手机分辨率   必填，宽*高；例如：320*480 
				pHeadJson.put("dpi", GoHttpHeadUtil.getDeviceDIP(context)); 
				//sdk int 系统sdklevel版本   必填，例如：10
				pHeadJson.put("sdk", Build.VERSION.SDK_INT); 
				//String 系统版本  必填：2.3.6
				pHeadJson.put("sys", Build.VERSION.RELEASE); 
				//String 机型   必填
				pHeadJson.put("model", Build.MODEL); 
				//String 请求时间戳   必填，客户端请求服务器的手机时间戳 
				pHeadJson.put("requesttime", System.currentTimeMillis()); 
				//int 入口id 必填，默认1
				pHeadJson.put("entranceId", 0); 
				//int 是否为官方包  选填，默认0：无法辨认（服务端根据渠道+国家判断）; 1：官方；2：非官方； 
				pHeadJson.put("official", 0); 
				
				//int 是否安装了电子市场   必填，（0：未安装，1：已安装），例如：1
				pHeadJson.put("hasmarket", GoHttpHeadUtil.isExistGoogleMarket(context) ? 1 : 0); 
				//String 网络类型  必填（模块会根据网络类型下发），unknown,wifi,gprs,3g,4g 
				pHeadJson.put("net", GoHttpHeadUtil.getNetType(context)); 
				//String 经纬度信息 选填，经度#纬度；
				pHeadJson.put("coordinates", "0#0"); 
				//String 定位到的位置信息  选填，国家#省份#城市
				pHeadJson.put("positions", "0#0#0"); 
			} catch (JSONException e) {
				e.printStackTrace();
			}
		}
		return pHeadJson;
	}
	
	/**
	 * 根据go协议解析服务器响应信息
	 * @param response
	 * @param shandle
	 * @return
	 */
	public static JSONObject parseResponse(HttpResponse response, int shandle) {
		if (null == response) {
			return null;
		}
		JSONObject json = HttpUtil.getJsonObject(response, shandle == VALUE_HANDLE_GZIP);
		if (null == json) {
			return null;
		}
		
		try {
			JSONObject resultJson = json.getJSONObject(TAG_RESULT);
			int status = resultJson.getInt(TAG_STATUS);
			if (status == RESULT_STATUS_OK || status == RESULT_STATUS_ERR_BUSSINESS) { 
				//服务器处理成功 （业务异常, 可认为是特殊的协议， 不是error）
				return json;
			}
		} catch (Exception e) {
			e.printStackTrace();
		}
		
		return null;
	}
	
	/**
	 * 解析go协议响应结果信息， 主要用于处理业务异常，即status为{@link #RESULT_STATUS_ERR_BUSSINESS}的情形
	 * @param responseJsonObject 响应数据，是调用{@link #parseResponse}解析出来的
	 * @return
	 */
	public static GoHttpPostResult parseResult(JSONObject responseJsonObject) {
		if (null == responseJsonObject) {
			return null;
		}
		
		try {
			GoHttpPostResult result = new GoHttpPostResult();
			JSONObject resultJson = responseJsonObject.getJSONObject(TAG_RESULT);
			result.mStatus = resultJson.getInt(TAG_STATUS);
			try {
				result.mErrorCode = resultJson.getInt(TAG_ERRORCODE);
				result.mErrorMsg = resultJson.getString(TAG_MSG);
			} catch (JSONException e) {
				//可以没有error信息
			}
			return result;
		} catch (JSONException e) {
			e.printStackTrace();
		}
		
		return null;
	}
	
	/**
	 * go协议响应结果
	 *
	 * @author matt
	 * @date: 2014年12月6日
	 *
	 */
	public static class GoHttpPostResult {
		/** 状态码， 详见 {@link #RESULT_STATUS_OK} 等 */
		public int mStatus = Integer.MIN_VALUE;
		/** 错误码， 由业务协议定义， 仅在mStatus不为{@link #RESULT_STATUS_OK}时有效  */
		public int mErrorCode;
		/** 错误信息， 由业务协议定义 */
		public String mErrorMsg;
	}
	
//	/**
//	 * go协议解析器
//	 *
//	 * @author matt
//	 * @date: 2014年12月5日
//	 *
//	 */
//	public static class GoResponseParser implements IResponseParser {
//
//		@Override
//		public MHttpResponse parse(MHttpRequest request, HttpResponse response) {
//			JSONObject jsonObject = GoHttpPostHandler.parseResponse(response, GoHttpPostHandler.VALUE_HANDLE_GZIP);
//			if (null == jsonObject) {
//				return null;
//			}
//			return new MHttpResponse(response, jsonObject);
//		}
//
//	}
}
