package com.jb.ga0.commerce.util.observer;

import java.util.ArrayList;
import java.util.List;

import android.content.BroadcastReceiver;
import android.content.Context;
import android.content.Intent;
import android.content.IntentFilter;

/**
 * 应用（安装卸载事件）监听器
 *
 * @author matt
 * @date: 2015年1月28日
 *
 */
public class AppChangeObserver {
	/**
	 * 是否动态注册广播监听，默认为true
	 * 如该值为false(静态注册)，需要在AndroidManifest.xml里配置 {@link AppChangedReceiver}
	 */
	private static final boolean IS_DYNAMIC_REGISTER = true;
	
	private static AppChangeObserver sInstance = null;
	private Context mContext;
	private AppChangedReceiver mReceiver;
	
    private ArrayList<OnAppChangedListener> mListeners = new ArrayList<OnAppChangedListener>();
    private byte[] mLock = new byte[0];
    
    private AppChangeObserver(Context context) {
    	mContext = context.getApplicationContext();
    }
    
    public static AppChangeObserver getInstance(Context context) {
    	if (null == sInstance) {
    		sInstance = new AppChangeObserver(context);
    	}
    	return sInstance;
    }
    
    /**
     * 在Activity:onDestroy回调此方法，注销系统广播监听
     */
    public static void onDestroy() {
    	if (sInstance != null) {
    		sInstance.unregisterReceiver();
    		sInstance.clearListeners();
    		sInstance = null;
    	}
    }
    
    private void registerReceiver() {
    	if (mReceiver != null) {
    		return;
    	}
    	mReceiver = new AppChangedReceiver();
    	IntentFilter appChangedFilter = new IntentFilter();
    	appChangedFilter.addAction(Intent.ACTION_PACKAGE_ADDED);  //状态监听：程序安装/卸载
    	appChangedFilter.addAction(Intent.ACTION_PACKAGE_REMOVED);       
    	appChangedFilter.addAction(Intent.ACTION_PACKAGE_REPLACED);      
    	appChangedFilter.addDataScheme("package");
    	mContext.registerReceiver(mReceiver, appChangedFilter);
    }
    
    private void unregisterReceiver() {
    	if (mReceiver == null) {
    		return;
    	}
    	mContext.unregisterReceiver(mReceiver);
    	mReceiver = null;
    }
    
	 /**
	  * 清空监听
	  */
	 public void clearListeners() {
		 synchronized (mLock) {
			 mListeners.clear();
		 }
	 }
    
    /**
     * 注册监听
     * @param listener
     */
    public void registerListener(OnAppChangedListener listener) {
    	if (null == listener) {
    		return;
    	}
    	//检测并注册系统广播监听
    	if (IS_DYNAMIC_REGISTER) {
    		registerReceiver();
    	}
    	
    	synchronized (mLock) {
    		for (OnAppChangedListener item : mListeners) {
    			if (item == listener) {
    				return;
    			}
    		}
    		mListeners.add(listener);
		}
    }
    
    /**
     * 注销监听
     * @param listener
     */
    public void unregisterListener(OnAppChangedListener listener) {
        if (listener != null) {
        	synchronized (mLock) {
        		mListeners.remove(listener);
			}
        }
    }
    
    private List<OnAppChangedListener> getListenersCopy() {
    	synchronized (mLock) {
    		return (ArrayList<OnAppChangedListener>) mListeners.clone();
    	}
    }
    
    private void notifyAppUninstalled(String pkgName) {
    	List<OnAppChangedListener> listeners = getListenersCopy();
    	for (OnAppChangedListener listener : listeners) {
			if (listener != null) {
				listener.onAppUninstalled(pkgName);
			}
		}
    }
    private void notifyAppReplaced(String pkgName) {
    	List<OnAppChangedListener> listeners = getListenersCopy();
    	for (OnAppChangedListener listener : listeners) {
    		if (listener != null) {
    			listener.onAppReplaced(pkgName);
    		}
    	}
    }
    private void notifyAppInstalled(String pkgName) {
    	List<OnAppChangedListener> listeners = getListenersCopy();
    	for (OnAppChangedListener listener : listeners) {
    		if (listener != null) {
    			listener.onAppInstalled(pkgName);
    		}
    	}
    }
    
    /**
     * 
     * @param action
     * @param pkgName
     * @param isStatic
     */
    private void onReceive(String action, String pkgName) {
    	if (Intent.ACTION_PACKAGE_REMOVED.equals(action)) {
    		notifyAppUninstalled(pkgName);
    	} else if (Intent.ACTION_PACKAGE_REPLACED.equals(action)) {
    		notifyAppReplaced(pkgName);
    	} else if (Intent.ACTION_PACKAGE_ADDED.equals(action)) {
    		notifyAppInstalled(pkgName);
    	}
    }
    
    /**
     * 程序安装卸载事件的监听
     *
     * @author matt
     * @date: 2015年1月28日
     *
     */
    public static interface OnAppChangedListener {

        /**
         * 应用安装
         * @param pkgName
         */
        public void onAppInstalled(String pkgName);

        /**
         * 应用卸载
         * @param pkgName
         */
        public void onAppUninstalled(String pkgName);

        /**
         * 应用覆盖安装
         * @param pkgName
         */
        public void onAppReplaced(String pkgName);
    }
    
    /**
     * 系统事件监听
     *
     * @author matt
     * @date: 2015年1月29日
     *
     */
    public static class AppChangedReceiver extends BroadcastReceiver {

        @Override
        public void onReceive(Context context, Intent intent) {
        	if (null == context || null == intent || null == sInstance) {
        		//只有监听器被初始化过，才回调
        		return;
        	}
        	
        	String action = intent.getAction();
        	String packageName = intent.getData().getSchemeSpecificPart();
        	if (Intent.ACTION_PACKAGE_REMOVED.equals(action) 
        			|| Intent.ACTION_PACKAGE_REPLACED.equals(action)
        			|| Intent.ACTION_PACKAGE_ADDED.equals(action)) {
        		sInstance.onReceive(action, packageName);
        	}
        }
    }
}
