package com.jb.ga0.commerce.util.observer;

import android.content.BroadcastReceiver;
import android.content.Context;
import android.content.Intent;
import android.content.IntentFilter;
import android.net.ConnectivityManager;
import android.net.wifi.WifiManager;

import com.jb.ga0.commerce.util.NetUtil;
import com.jb.ga0.commerce.util.observer.AppChangeObserver.AppChangedReceiver;

import java.util.ArrayList;
import java.util.List;

/**
 * 网络状态监听器
 *
 * @author matt
 * @date: 2015年1月29日
 *
 */
public class NetStateObserver {
	/**
	 * 是否动态注册广播监听，默认为true
	 * 如该值为false(静态注册)，需要在AndroidManifest.xml里配置 {@link AppChangedReceiver}
	 */
	private static final boolean IS_DYNAMIC_REGISTER = true;
	/**
	 * 是否监听普通网络状态变化（false只监听wifi状态变化)
	 */
	private static final boolean IS_OBSERVER_GENERAL_NETSTATE = true;

	private static NetStateObserver sInstance = null;
	private Context mContext;
	private NetStateReceiver mReceiver;
	private boolean mIsNetAvailable;
	private boolean mIsWifiConnected;

	private ArrayList<OnNetStateChangeListener> mListeners = new ArrayList<OnNetStateChangeListener>();
	private byte[] mLock = new byte[0];

	private NetStateObserver(Context context) {
		mContext = context.getApplicationContext();
		mIsNetAvailable = NetUtil.isNetWorkAvailable(context);
		mIsWifiConnected = NetUtil.isWifiEnable(context);
	}

	public static NetStateObserver getInstance(Context context) {
		if (null == sInstance) {
			sInstance = new NetStateObserver(context);
		}
		return sInstance;
	}

	/**
	 * 在进程退出时回调此方法，注销系统广播监听
	 */
	public static void onDestroy() {
		if (sInstance != null) {
    		sInstance.unregisterReceiver();
    		sInstance.clearListeners();
    		sInstance = null;
    	}
	}

	private void registerReceiver() {
		if (mReceiver != null) {
			return;
		}

		mReceiver = new NetStateReceiver();
		IntentFilter filter = new IntentFilter();
		filter.addAction(WifiManager.NETWORK_STATE_CHANGED_ACTION);
		if (IS_OBSERVER_GENERAL_NETSTATE) {
			filter.addAction(ConnectivityManager.CONNECTIVITY_ACTION);
		}
		mContext.registerReceiver(mReceiver, filter);
	}

	private void unregisterReceiver() {
		if (mReceiver == null) {
			return;
		}
		mContext.unregisterReceiver(mReceiver);
		mReceiver = null;
	}

	/**
	 * 注册监听
	 * @param listener
	 */
	 public void registerListener(OnNetStateChangeListener listener) {
		 if (null == listener) {
			 return;
		 }
		 //检测并注册系统广播监听
		 if (IS_DYNAMIC_REGISTER) {
			 registerReceiver();
		 }

		 synchronized (mLock) {
			 for (OnNetStateChangeListener item : mListeners) {
				 if (item == listener) {
					 return;
				 }
			 }
			 mListeners.add(listener);
		 }
	 }

	 /**
	  * 注销监听
	  * @param listener
	  */
	 public void unregisterListener(OnNetStateChangeListener listener) {
		 if (null == listener) {
			 return;
		 }
		 synchronized (mLock) {
			 mListeners.remove(listener);
		 }
	 }
	 
	 /**
	  * 清空监听
	  */
	 public void clearListeners() {
		 synchronized (mLock) {
			 mListeners.clear();
		 }
	 }

	 private List<OnNetStateChangeListener> getListenersCopy() {
		 synchronized (mLock) {
			 return (ArrayList<OnNetStateChangeListener>) mListeners.clone();
		 }
	 }
	 
	 /**
	  * 网络状态变化
	  * @param isAvailable
	  */
	 private void notifyNetStateChanged(boolean isAvailable) {
		 List<OnNetStateChangeListener> listeners = getListenersCopy();
		 for (OnNetStateChangeListener listener : listeners) {
			 if (listener != null) {
				 listener.onNetStateChanged(isAvailable);
			 }
		 }
	 }

	 /**
	  * Wifi状态变化
	  * @param isConnected
	  */
	 private void notifyWifiStateChanged(boolean isConnected) {
		 List<OnNetStateChangeListener> listeners = getListenersCopy();
		 for (OnNetStateChangeListener listener : listeners) {
			 if (listener != null) {
				 listener.onWifiStateChanged(isConnected);
			 }
		 }
	 }

	 private void setNetState(boolean isAvailable) {
		 if (mIsNetAvailable == isAvailable) {
			 return;
		 }
		 mIsNetAvailable = isAvailable;
		 notifyNetStateChanged(isAvailable);
	 }

	 private void setWifiState(boolean isConnected) {
		 if (mIsWifiConnected == isConnected) {
			 return;
		 }
		 mIsWifiConnected = isConnected;
		 notifyWifiStateChanged(isConnected);
	 }

	 /**
	  * 网络状态监听
	  *
	  * @author matt
	  * @date: 2015年1月29日
	  *
	  */
	 public static interface OnNetStateChangeListener {
		 /**
		  * 网络状态变化
		  * @param isAvailable
		  */
		 void onNetStateChanged(boolean isAvailable);

		 /**
		  * Wifi状态变化
		  * @param isConnected
		  */
		 void onWifiStateChanged(boolean isConnected);
	 }

	 /**
	  * 系统事件监听
	  *
	  * @author matt
	  * @date: 2015年1月29日
	  *
	  */
	 public static class NetStateReceiver extends BroadcastReceiver {

		 @Override
		 public void onReceive(Context context, Intent intent) {
			 if (null == context || null == intent || null == sInstance) {
				 return;
			 }
			 sInstance.setNetState(NetUtil.isNetWorkAvailable(context));
			 sInstance.setWifiState(NetUtil.isWifiEnable(context));

//			 Log.i("NetStateReceiver", "[NetStateReceiver#onReceive] net:" + NetUtil.isNetWorkAvailable(context) + ", wifi:" + NetUtil.isWifiEnable(context) + ", action:" + intent.getAction());
//
//			 String action = intent.getAction();
//			 if (WifiManager.NETWORK_STATE_CHANGED_ACTION.equals(action)) {
//				 /**
//				  * WifiManager.WIFI_STATE_CHANGED_ACTION， 这个监听wifi的打开与关闭，与wifi的连接无关
//				  * WifiManager.NETWORK_STATE_CHANGED_ACTION， 这个监听wifi的连接状态
//				  */
//				 NetworkInfo networkInfo = intent.getParcelableExtra(WifiManager.EXTRA_NETWORK_INFO);
//				 if (networkInfo != null) {
//					 sInstance.setWifiState(networkInfo.isConnected());
//				 }
//			 } else if (IS_OBSERVER_GENERAL_NETSTATE && ConnectivityManager.CONNECTIVITY_ACTION.equals(action)) {
//				 //这个监听网络连接的设置，包括wifi和移动数据的打开和关闭。.
//				 //最好用的还是这个监听。wifi如果打开，关闭，以及连接上可用的连接都会接到监听。见log
//				 // 这个广播的最大弊端是比上边两个广播的反应要慢，如果只是要监听wifi，我觉得还是用上边两个配合比较合适
//				 sInstance.setNetState(NetUtil.isNetWorkAvailable(context));
//			 }
		 }
	 }	
}
