package com.jb.ga0.commerce.util.observer;

import java.util.ArrayList;
import java.util.List;

import com.jb.ga0.commerce.util.Machine;

import android.content.BroadcastReceiver;
import android.content.Context;
import android.content.Intent;
import android.content.IntentFilter;

/**
 * 应用（安装卸载事件）监听器
 *
 * @author matt
 * @date: 2015年1月28日
 *
 */
public class SDCardMountObserver {
	/**
	 * 是否动态注册广播监听，默认为true
	 * 如该值为false(静态注册)，需要在AndroidManifest.xml里配置 {@link SDCardMountReceiver}
	 */
	private static final boolean IS_DYNAMIC_REGISTER = true;
	
	private static SDCardMountObserver sInstance = null;
	private Context mContext;
	private SDCardMountReceiver mReceiver;
	
    private ArrayList<OnSdCardStatusChangedListener> mListeners = new ArrayList<OnSdCardStatusChangedListener>();
    private byte[] mLock = new byte[0];
    
    private SDCardMountObserver(Context context) {
    	mContext = context.getApplicationContext();
    }
    
    /**
     * 获取sd卡是否可用
     * @return
     */
    public static boolean isSDCardAvailable() {
    	return Machine.isSDCardExist();
    }
    
    public static SDCardMountObserver getInstance(Context context) {
    	if (null == sInstance) {
    		sInstance = new SDCardMountObserver(context);
    	}
    	return sInstance;
    }
    
    /**
     * 在Activity:onDestroy回调此方法，注销系统广播监听
     */
    public static void onDestroy() {
    	if (sInstance != null) {
    		sInstance.unregisterReceiver();
    		sInstance.clearListeners();
    		sInstance = null;
    	}
    }
    
    private void registerReceiver() {
    	if (mReceiver != null) {
    		return;
    	}
    	mReceiver = new SDCardMountReceiver();
    	IntentFilter intentFilter = new IntentFilter();
        intentFilter.addAction(Intent.ACTION_MEDIA_MOUNTED);
        intentFilter.addAction(Intent.ACTION_MEDIA_UNMOUNTED);
        intentFilter.addDataScheme("file");
    	mContext.registerReceiver(mReceiver, intentFilter);
    }
    
    private void unregisterReceiver() {
    	if (mReceiver == null) {
    		return;
    	}
    	mContext.unregisterReceiver(mReceiver);
    	mReceiver = null;
    }
    
	 /**
	  * 清空监听
	  */
	 public void clearListeners() {
		 synchronized (mLock) {
			 mListeners.clear();
		 }
	 }
	 
	 private List<OnSdCardStatusChangedListener> getListenersCopy() {
		 synchronized (mLock) {
			 return (ArrayList<OnSdCardStatusChangedListener>) mListeners.clone();
		 }
	 }
	 
	 /**
	  * SD卡状态变化
	  * @param isAvailable
	  */
	 private void notifySdCardAvailableChanged(boolean isAvailable) {
		 List<OnSdCardStatusChangedListener> listeners = getListenersCopy();
		 for (OnSdCardStatusChangedListener listener : listeners) {
			 if (listener != null) {
				 listener.onSdCardAvailableChanged(isAvailable);
			 }
		 }
	 }
    
    /**
     * 注册监听
     * @param listener
     */
    public void registerListener(OnSdCardStatusChangedListener listener) {
    	if (null == listener) {
    		return;
    	}
    	//检测并注册系统广播监听
    	if (IS_DYNAMIC_REGISTER) {
    		registerReceiver();
    	}
    	
    	synchronized (mLock) {
    		for (OnSdCardStatusChangedListener item : mListeners) {
    			if (item == listener) {
    				return;
    			}
    		}
    		mListeners.add(listener);
		}
    }
    
    /**
     * 注销监听
     * @param listener
     */
    public void unregisterListener(OnSdCardStatusChangedListener listener) {
        if (listener != null) {
        	synchronized (mLock) {
        		mListeners.remove(listener);
			}
        }
    }
    
    /**
     * 
     * @param action
     * @param pkgName
     * @param isStatic
     */
    private void onReceive(String action) {
    	
    	synchronized (mLock) {
    		if (Intent.ACTION_MEDIA_MOUNTED.equals(action)) {
    			notifySdCardAvailableChanged(true);
    		} else if (Intent.ACTION_MEDIA_UNMOUNTED.equals(action)) {
    			notifySdCardAvailableChanged(false);
    		}
		}
    }
    
    /**
     * 程序安装卸载事件的监听
     *
     * @author matt
     * @date: 2015年1月28日
     *
     */
    public static interface OnSdCardStatusChangedListener {

        /**
         * SD卡可用变化
         * @param isAvailable SD卡是否可用
         */
        public void onSdCardAvailableChanged(boolean isAvailable);
    }
    
    /**
     * 系统事件监听
     *
     * @author matt
     * @date: 2015年1月29日
     *
     */
    public static class SDCardMountReceiver extends BroadcastReceiver {

        @Override
        public void onReceive(Context context, Intent intent) {
        	if (null == context || null == intent || null == sInstance) {
        		//只有监听器被初始化过，才回调
        		return;
        	}
        	
        	String action = intent.getAction();
        	sInstance.onReceive(action);
        }
    }
}
