package com.jb.ga0.commerce.util.observer;

import java.util.ArrayList;
import java.util.List;

import android.content.Context;
import android.content.res.Configuration;

/**
 * 系统配置通知器, 如横竖屏切换，语言切换等
 *
 * @author matt
 * @date: 2014年12月10日
 *
 */
public class SysConfigChangedObserver {
	private static SysConfigChangedObserver sInstance = null;
	private ArrayList<ISysConfigChanged> mListeners = new ArrayList<ISysConfigChanged>();
	private byte[] mLock = new byte[0];
	private int mOrientation;
	
	private SysConfigChangedObserver(Context context) {
		Configuration config = context.getResources().getConfiguration();
		mOrientation = config.orientation;
	}
	
	/**
	 * 应用启动时调用此初始化方法， 否则会丢失事件
	 * @param context
	 */
	public static SysConfigChangedObserver getInstance(Context context) {
		if (null == sInstance) {
			sInstance = new SysConfigChangedObserver(context);
		}
		return sInstance;
	}
	
	/**
	 * Activity或Service的onConfigurationChanged事件里调用此方法处理事件
	 * @param newConfig
	 */
	public void onConfigurationChanged(Configuration newConfig) {
		if (null == newConfig) {
			return;
		}
		
		if (mOrientation != newConfig.orientation) {
			mOrientation = newConfig.orientation;
			notifyOrientationChanged(mOrientation);
		}
		
	}
	
	/**
	 * 注册监听
	 * @param listener
	 */
	public void registerListener(ISysConfigChanged listener) {
		if (null == listener) {
			return;
		}
		synchronized (mLock) {
			if (mListeners.contains(listener)) {
				return;
			}
			mListeners.add(listener);
		}
	}
	
	/**
	 * 注销监听
	 * @param listener
	 */
	public void unregisterListener(ISysConfigChanged listener) {
		if (null == listener) {
			return;
		}
		synchronized (mLock) {
			mListeners.remove(listener);
		}
	}
	
	 /**
	  * 清空监听
	  */
	 public void clearListeners() {
		 synchronized (mLock) {
			 mListeners.clear();
		 }
	 }
	 
	 private List<ISysConfigChanged> getListenersCopy() {
		 synchronized (mLock) {
			 return (ArrayList<ISysConfigChanged>) mListeners.clone();
		 }
	 }
	 
	 /**
	  * 网络状态变化
	  * @param isAvailable
	  */
	 private void notifyOrientationChanged(int newOrientation) {
		 List<ISysConfigChanged> listeners = getListenersCopy();
		 for (ISysConfigChanged listener : listeners) {
			 if (listener != null) {
				 listener.onOrientationChanged(newOrientation);
			 }
		 }
	 }
	
	/**
	 * 系统配置变化监听器
	 *
	 * @author matt
	 * @date: 2014年12月10日
	 *
	 */
	public static interface ISysConfigChanged {
		/**
		 * 屏幕方向变化
		 * @param newOrientation 
		 * @see Configuration#ORIENTATION_PORTRAIT
		 * @see Configuration#ORIENTATION_LANDSCAPE
		 */
		void onOrientationChanged(int newOrientation);
		
//		void onLaunguageChanged();
	}
}
