package com.jb.ga0.commerce.util.retrofit.Interceptor;

import android.content.Context;
import android.util.Log;

import com.jb.ga0.commerce.util.LogUtils;
import com.jb.ga0.commerce.util.observer.NetStateObserver;
import com.jb.ga0.commerce.util.observer.NetStateObserver.OnNetStateChangeListener;

import java.util.ArrayList;
import java.util.HashMap;
import java.util.List;

import static com.jb.ga0.commerce.util.retrofit.HttpConstants.LogTag.TAG;

/**
 * Created by matt on 2017/10/19.
 */
public class RetryAfterNetOkCtrl implements OnNetStateChangeListener {
    public static final String RETRY_AFTER_NET_OK_KEY = "retry_after_net_ok_key";

    private static final int DEF_MAX_RETRY_COUNT = 3;

    private static RetryAfterNetOkCtrl sInstance;
    private Context mContext;
    /**
     * 需要失败重试的请求
     */
    private HashMap<String, CallRetryParam> mOrders = new HashMap<>();
    /**
     * 已经失败，在等待重试的请求
     */
    private List<CallRetryParam> mRetryCalls = new ArrayList<>();

    private RetryAfterNetOkCtrl(Context context) {
        mContext = context.getApplicationContext();
        NetStateObserver.getInstance(context).registerListener(this);
    }

    public static RetryAfterNetOkCtrl getInstance(Context context) {
        if (sInstance == null) {
            synchronized (RetryAfterNetOkCtrl.class) {
                if (sInstance == null) {
                    sInstance = new RetryAfterNetOkCtrl(context);
                }
            }
        }
        return sInstance;
    }

    /**
     * 发起一个http请求，该请求失败后可以在网络正常时重试，默认重试次数 {@link #DEF_MAX_RETRY_COUNT}
     * <br/>在调用此方法的同时，需要在header里添加retry_after_net_ok_key, 值同@param requestKey
     * <br/>Retrofit设置形如 @Headers("retry_after_net_ok_key: 你的requestKey")，示例如下：
     * <br/>@Headers("retry_after_net_ok_key: https://api.douban.com/v2/movie/retry")
     * <br/> 假定同一个requestKey不会重复call，重复call只认最后1次操作
     *
     * @param requestKey
     * @param r          执行http请求的runnable
     */
    public void call(String requestKey, final Runnable r) {
        call(requestKey, DEF_MAX_RETRY_COUNT, 0, r);
    }

    /**
     * 发起一个http请求，该请求失败后可以在网络正常时重试
     * <br/>在调用此方法的同时，需要在header里添加retry_after_net_ok_key, 值同@param requestKey
     * <br/>Retrofit设置形如 @Headers("retry_after_net_ok_key: 你的requestKey")，示例如下：
     * <br/>@Headers("retry_after_net_ok_key: https://api.douban.com/v2/movie/retry")
     * <br/> 假定同一个requestKey不会重复call，重复call只认最后1次操作
     *
     * @param requestKey
     * @param maxRetryCount 最大重试次数
     * @param r             执行http请求的runnable
     */
    public void call(String requestKey, int maxRetryCount, final Runnable r) {
        call(requestKey, maxRetryCount, 0, r);
    }

    /**
     * 发起一个http请求，该请求失败后可以在网络正常时重试
     * <br/> 假定同一个requestKey不会重复call，重复call只认最后1次操作
     *
     * @param requestKey
     * @param maxRetryCount 最大重试次数
     * @param r             执行http请求的runnable
     */
    void call(String requestKey, int maxRetryCount, int retryCount, final Runnable r) {
        if (mOrders.containsKey(requestKey)) {
            Log.w(TAG, "[RetryAfterNetOkCtrl#call] requestKey已存在，请检查是否equestKey定义重复！", new Throwable());
        }
        if (maxRetryCount > retryCount) {
            mOrders.put(requestKey, new CallRetryParam(requestKey, r)
                    .setMaxRetryCount(maxRetryCount).setRetryCount(retryCount));
        }
        r.run();
    }

    void recordResult(String requestKey, boolean success) {
        CallRetryParam callRetryParam = mOrders.remove(requestKey);
        if (!success) {
            addRetryCall(callRetryParam);
        }
    }

    void addRetryCall(CallRetryParam retryCall) {
        synchronized (mRetryCalls) {
            mRetryCalls.add(retryCall);
        }
    }

    void removeRetryCall(CallRetryParam retryCall) {
        synchronized (mRetryCalls) {
            mRetryCalls.remove(retryCall);
        }
    }

    @Override
    public void onNetStateChanged(boolean b) {
//        LogUtils.i(TAG, "RetryAfterNetOkCtrl#onNetStateChanged() called with: b = [" + b + "]");
        if (b) {
            List<CallRetryParam> calls = clearRetryCalls();
            for (CallRetryParam call : calls) {
                if (call != null) {
                    call.retry();
                }
            }
        }
    }

    private List<CallRetryParam> clearRetryCalls() {
        List<CallRetryParam> calls = new ArrayList<>();
        synchronized (mRetryCalls) {
            calls.addAll(mRetryCalls);
            mRetryCalls.clear();
        }
        return calls;
    }

    @Override
    public void onWifiStateChanged(boolean b) {
    }

    class CallRetryParam {
        String mRequestKey;
        Runnable mRunnable;
        int mMaxRetryCount = 3;
        int mRetryCount = 0;

        CallRetryParam(String requestKey, Runnable runnable) {
            mRequestKey = requestKey;
            mRunnable = runnable;
        }

        CallRetryParam setMaxRetryCount(int maxRetryCount) {
            mMaxRetryCount = maxRetryCount;
            return this;
        }

        public CallRetryParam setRetryCount(int retryCount) {
            mRetryCount = retryCount;
            return this;
        }

        void retry() {
            if (mRetryCount < mMaxRetryCount) {
                mRetryCount++;

                LogUtils.i(TAG, "CallRetryParam#retry() " + "mMaxRetryCount = " + mMaxRetryCount
                        + ", mRetryCount = " + mRetryCount + ", mRequestKey = " + mRequestKey);
                call(mRequestKey, mMaxRetryCount, mRetryCount, mRunnable);
            }
        }
    }
}