package com.jb.ga0.commerce.util.retrofit;

import android.content.Context;
import android.content.SharedPreferences;

import com.jb.ga0.commerce.util.CustomAlarm;
import com.jb.ga0.commerce.util.retrofit.Interceptor.RepeatRequestCtrl;
import com.jb.ga0.commerce.util.retrofit.Interceptor.RetryAfterNetOkCtrl;
import com.jb.ga0.commerce.util.retrofit.cache.Cache;
import com.jb.ga0.commerce.util.retrofit.cache.CustomCacheInterceptor;
import com.jb.ga0.commerce.util.retrofit.cache.CustomCacheInterceptor.CacheType;

import java.io.IOException;
import java.util.HashMap;
import java.util.Map;

import okhttp3.CacheControl;
import okhttp3.OkHttpClient;
import okhttp3.RequestBody;
import okhttp3.ResponseBody;
import retrofit2.Call;
import retrofit2.Callback;
import retrofit2.Response;
import retrofit2.Retrofit;

import static com.jb.ga0.commerce.util.retrofit.RetrofitRequest.Method.get;

/**
 * Created by matt on 2017/11/2.
 */
public class RetrofitRequest {
    public static enum Method {
        get,
        post,
    }

    String mUrl;
    Method mMethod;
    Map<String, String> mHeaderMap;
    /**
     * get请求参数
     */
    Map<String, String> mQueryMap;
    /**
     * post数据键值对
     */
    Map<String, String> mFieldMap;
    /**
     * post数据，优先级高于 {@link #mFieldMap}
     */
    RequestBody mRequestBody;
    /**
     * retrofit callback，优先级高于 {@link #mHttpCallback}
     */
    Callback mCallback;
    /**
     * 自封装的callback
     */
    RetrofitProxy.HttpCallback mHttpCallback;
    /**
     * 缓存设置
     */
    RequestCache mRequestCache;
    /**
     * 失败等网络好后重连设置
     */
    RequestRetryAfterNetOk mRequestRetryAfterNetOk;
    /**
     * 定时请求设置
     */
    RequestRepeat mRequestRepeat;

    public RetrofitRequest(String url, Method method) {
        mUrl = url;
        mMethod = method;
    }

    /**
     * 设置Retrofit原生callback
     * @param callback
     * @return
     */
    public RetrofitRequest setCallback(Callback<ResponseBody> callback) {
        mCallback = callback;
        return this;
    }

    /**
     * 设置封装callback
     * @param httpCallback
     * @return
     */
    public RetrofitRequest setCallback(RetrofitProxy.HttpCallback<ResponseBody> httpCallback) {
        mHttpCallback = httpCallback;
        return this;
    }

    /**
     * 添加请求头
     * @param key
     * @param value
     * @return
     */
    public RetrofitRequest addHeader(String key, String value) {
        if (mHeaderMap == null) {
            mHeaderMap = new HashMap<>();
        }
        mHeaderMap.put(key, value);
        return this;
    }

    /**
     * 设置请求头map
     * @param headerMap
     * @return
     */
    public RetrofitRequest setHeaders(Map<String, String> headerMap) {
        mHeaderMap = headerMap;
        return this;
    }

    /**
     * 添加get请求参数
     * @param key
     * @param value
     * @return
     */
    public RetrofitRequest addQuery(String key, String value) {
        if (mQueryMap == null) {
            mQueryMap = new HashMap<>();
        }
        mQueryMap.put(key, value);
        return this;
    }

    /**
     * 设置get请求参数map
     * @param queryMap
     * @return
     */
    public RetrofitRequest setQueryMap(Map<String, String> queryMap) {
        mQueryMap = queryMap;
        return this;
    }

    /**
     * 添加post键值对
     * @param key
     * @param value
     * @return
     */
    public RetrofitRequest addField(String key, String value) {
        if (mFieldMap == null) {
            mFieldMap = new HashMap<>();
        }
        mFieldMap.put(key, value);
        return this;
    }

    /**
     * 设置post键值对
     * @param fieldMap
     * @return
     */
    public RetrofitRequest setFieldMap(Map<String, String> fieldMap) {
        mFieldMap = fieldMap;
        return this;
    }

    /**
     * 设置post body
     * @param body
     * @return
     */
    public RetrofitRequest setRequestBody(RequestBody body) {
        mRequestBody = body;
        return this;
    }

    /**
     * 设置缓存
     * @param requestCache
     * @return
     */
    public RetrofitRequest setRequestCache(RequestCache requestCache) {
        mRequestCache = requestCache;
        return this;
    }

    /**
     * 设置失败等网络好后自动重连
     * @param requestRetryAfterNetOk
     * @return
     */
    public RetrofitRequest setRequestRetryAfterNetOk(RequestRetryAfterNetOk requestRetryAfterNetOk) {
        mRequestRetryAfterNetOk = requestRetryAfterNetOk;
        return this;
    }

    /**
     * 设置定时请求
     * @param requestRepeat
     * @return
     */
    public RetrofitRequest setRequestRepeat(RequestRepeat requestRepeat) {
        mRequestRepeat = requestRepeat;
        return this;
    }

    /**
     * 异步执行
     *
     * @param context
     */
    public void enqueue(final Context context) {
        if (mRequestRepeat == null && mRequestRetryAfterNetOk == null) {
            //无定时请求，无网络好后失败重试
            enqueueRaw(context);
        } else if (mRequestRepeat == null && mRequestRetryAfterNetOk != null) {
            //无定时请求，有网络好后失败重试
            enqueueRetryAfterNetOk(context);
        } else if (mRequestRepeat != null && mRequestRetryAfterNetOk == null) {
            //有定时请求，无网络好后失败重试
            enqueueRepeat(context, new Runnable() {
                @Override
                public void run() {
                    enqueueRaw(context);
                }
            });
        } else { //mRequestRepeat != null && mRequestRetryAfterNetOk != null
            //有定时请求，也有网络好后失败重试
            enqueueRepeat(context, new Runnable() {
                @Override
                public void run() {
                    enqueueRetryAfterNetOk(context);
                }
            });
        }
    }

    void enqueueRaw(Context context) {
        Call call = buildCall(context);
        if (mCallback != null) {
            call.enqueue(mCallback);
        } else if (mHttpCallback != null) {
            call.enqueue(new RetrofitProxy.RetrofitCallback(mHttpCallback));
        } else {
            throw new IllegalArgumentException("执行enqueue方法，callback不能为空！");
        }
    }

    void enqueueRetryAfterNetOk(final Context context) {
        String requestKey = mRequestRetryAfterNetOk.mRequestKey;
        int maxRetryCount = mRequestRetryAfterNetOk.mMaxRetryCount;

        addHeader(RetryAfterNetOkCtrl.RETRY_AFTER_NET_OK_KEY, requestKey);
        //发起可失败重试的请求
        RetryAfterNetOkCtrl.getInstance(context).call(requestKey, maxRetryCount, new Runnable() {
            @Override
            public void run() {
                enqueueRaw(context);
            }
        });
    }

    void enqueueRepeat(Context context, Runnable r) {
        SharedPreferences sp = mRequestRepeat.mSp;
        CustomAlarm alarm = mRequestRepeat.mAlarm;
        int alarmId = mRequestRepeat.mAlarmId;
        long delay = mRequestRepeat.mDelay;
        long interval = mRequestRepeat.mInterval;
        String requestKey = mRequestRepeat.mRequestKey;

        addHeader(RepeatRequestCtrl.REPEAT_REQUEST_KEY, requestKey);
        //发起定时请求
        RepeatRequestCtrl.getInstance().callRepeat(sp, alarm, alarmId, delay, interval, requestKey, r);
    }

    /**
     * 同步执行
     *
     * @param context
     * @return
     * @throws IOException
     */
    public Response execute(Context context) throws IOException {
        if (mRequestRetryAfterNetOk != null) {
            throw new IllegalArgumentException("execute方法不支持RequestFailRetry");
        }
        if (mRequestRepeat != null) {
            throw new IllegalArgumentException("execute方法不支持RequestRepeat");
        }
        Call call = buildCall(context);
        return call.execute();
    }

    Call buildCall(Context context) {
        OkHttpClient client = ClientProvider.getInstance(context).get();
        if (mRequestCache != null) {
            client = client.newBuilder()
                    .addInterceptor(new CustomCacheInterceptor.Builder(
                            context, mRequestCache.mCacheType)
                            .setCacheControl(mRequestCache.mCacheControl)
                            .setCacheKey(mRequestCache.mCacheKey)
                            .setCache(mRequestCache.mCache)
                            .build()
                    )
                    .build();
        }
        Retrofit retrofit = new Retrofit.Builder()
                .client(client)
                .baseUrl(mUrl.substring(0, mUrl.lastIndexOf("/") + 1))
                .build();

        Call<ResponseBody> call = null;
        if (mHeaderMap == null) {
            mHeaderMap = new HashMap<>();
        }
        if (mMethod == get) {
            if (mQueryMap == null) {
                mQueryMap = new HashMap<>();
            }
            return retrofit.create(CommonService.class).get(mUrl, mHeaderMap, mQueryMap);
        } else {
            if (mRequestBody != null) {
                return retrofit.create(CommonService.class).post(mUrl, mHeaderMap, mRequestBody);
            } else {
                if (mFieldMap == null) {
                    mFieldMap = new HashMap<>();
                }
                return retrofit.create(CommonService.class).post(mUrl, mHeaderMap, mFieldMap);
            }
        }
    }

    /**
     * 缓存参数
     */
    public static class RequestCache {
        CacheType mCacheType;
        CacheControl mCacheControl;
        String mCacheKey;
        Cache mCache;

        public RequestCache(CacheType cacheType, CacheControl cacheControl, String cacheKey) {
            mCacheType = cacheType;
            mCacheControl = cacheControl;
            mCacheKey = cacheKey;
        }

        /**
         * 不设置的话，默认Cache使用 {@link Cache#getsInstance(Context)}
         * @param cache
         * @return
         */
        public RequestCache setCache(Cache cache) {
            mCache = cache;
            return this;
        }
    }

    /**
     * 网络好后重试参数
     */
    public static class RequestRetryAfterNetOk {
        String mRequestKey;
        int mMaxRetryCount;

        public RequestRetryAfterNetOk(String requestKey, int maxRetryCount) {
            mRequestKey = requestKey;
            mMaxRetryCount = maxRetryCount;
        }
    }

    /**
     * 定时请求参数
     */
    public static class RequestRepeat {
        SharedPreferences mSp;
        CustomAlarm mAlarm;
        int mAlarmId;
        long mDelay;
        long mInterval;
        String mRequestKey;

        public RequestRepeat(SharedPreferences sp, CustomAlarm alarm, int alarmId,
                             long delay, long interval, String requestKey) {
            mSp = sp;
            mAlarm = alarm;
            mAlarmId = alarmId;
            mDelay = delay;
            mInterval = interval;
            mRequestKey = requestKey;
        }
    }
}
