package com.jiubang.commerce.daemon;

import android.app.Service;
import android.content.Context;
import android.content.Intent;
import android.content.SharedPreferences;
import android.content.SharedPreferences.Editor;
import android.os.Build;
import android.text.TextUtils;

import com.jb.ga0.commerce.util.io.MultiprocessSharedPreferences;
import com.jiubang.commerce.daemon.util.CustomAlarm;
import com.jiubang.commerce.daemon.util.DaemonProcessAlarm;
import com.jiubang.commerce.daemon.util.LogUtils;
import com.jiubang.commerce.daemon.util.PackageUtils;

import java.io.BufferedReader;
import java.io.File;
import java.io.FileReader;
import java.io.IOException;
import java.util.List;

/**
 * 
 * @author
 *
 */
public class DaemonClient implements IDaemonClient {
	private static DaemonClient sInstance;

	private Context mContext;
	private DaemonConfigurations mConfigurations;
	private BufferedReader mBufferedReader; //release later to save time

	public DaemonClient() {
	}

	public static DaemonClient getInstance() {
		if (sInstance == null) {
			sInstance = new DaemonClient();
		}
		return sInstance;
	}

	/**
	 * 初始化配置信息
	 * @param configurations
	 */
	public void init(DaemonConfigurations configurations) {
		mConfigurations = configurations;
		if (LogUtils.sIsLog) {
			LogUtils.i("Daemon", String.format("DaemonClient::init-->{persistent:%s}, {daemon:%s}",
					configurations.mPersistentConfig.toString(), configurations.mDaemonAssistantConfig.toString()));
		}
	}

	/**
	 * 设置debug模式，打开log
	 */
	public void setDebugMode() {
		LogUtils.sIsLog = true;
	}

	/**
	 * 获取配置信息
	 * @return
	 */
	public DaemonConfigurations getConfigurations() {
		return mConfigurations;
	}

	/**
	 * 获取被守护的进程服务名称
	 * @return
	 */
	public String getPersistentServiceName() {
		if (mConfigurations != null && mConfigurations.mPersistentConfig != null) {
			return mConfigurations.mPersistentConfig.mServiceName;
		}
		return null;
	}

	/**
	 * 关联上下文
	 * @param base
	 */
	@Override
	public void onAttachBaseContext(Context base) {
		mContext = base;
		initDaemon(base);
		//防止开机广播被杀软禁止
		PackageUtils.setComponentEnable(base, BootCompleteReceiver.class.getName());
	}

	/**
	 * do some thing about daemon
	 * @param base
	 */
	private void initDaemon(Context base) {
		LogUtils.d("Daemon", "DaemonClient::initDaemon-->enter");
		if (mConfigurations == null) {
			LogUtils.d("Daemon", "DaemonClient::initDaemon-->exit, mConfigurations == null");
			return ;
		}
		if (!isDaemonPermitting(base)) {
			LogUtils.d("Daemon", "DaemonClient::initDaemon-->exit, isDaemonPermitting == false");
			return ;
		}
		if (DaemonConstants.isNotSupportDaemon()) {
			LogUtils.d("Daemon", "DaemonClient::initDaemon-->exit, isNotSupportDaemon == true, fingerprint:" + Build.FINGERPRINT);
			return ;
		}

		String processName = getProcessName();
		String packageName = base.getPackageName();
		if (LogUtils.sIsLog) {
			LogUtils.d("Daemon", "DaemonClient::initDaemon-->processName:" + processName + ", pkgName:" + packageName);
		}

		IDaemonStrategy strategy = IDaemonStrategy.Fetcher.fetchStrategy();
		strategy.setProcessName(processName);
		if (processName.equals(mConfigurations.mPersistentConfig.mProcessName)) {
			strategy.onPersistentCreate(base, mConfigurations);
			daemonOtherPersistentProcesses(base);

			//仅用于统计
			if (mConfigurations.isStatisticsDaemonEffect()) {
				LogUtils.i("Daemon", "[DaemonClient::initDaemon] invoke startService getPersistentServiceName");
				PackageUtils.startService(base, getPersistentServiceName(), false);
			}
		} else if (processName.equals(mConfigurations.mDaemonAssistantConfig.mProcessName)) {
			strategy.onDaemonAssistantCreate(base, mConfigurations);
		} else {
			strategy.onInitialization(base);
		}
		
		releaseIO();
	}

	/**
	 * 守护其它常驻进程
	 */
	private void daemonOtherPersistentProcesses(Context context) {
		List<String> otherPersistentServices = mConfigurations.getOtherPersistenServices();
		if (otherPersistentServices == null || otherPersistentServices.size() <= 0) {
			return;
		}

		int alarmId = DaemonProcessAlarm.ALARMID_ACTIVATE_OTHRE_PERSISTENT_PROCESS;
		long triggerDelay = mConfigurations.getDaemonOtherPersistentProcessTrigger() * 1000;
		long interval = mConfigurations.getDaemonOtherPersistentProcessInterval() * 1000;
		DaemonProcessAlarm.getAlarm(context).alarmRepeat(alarmId, triggerDelay, interval, true, new CustomAlarm.OnAlarmListener() {

			@Override
			public void onAlarm(int alarmId) {
				List<String> otherPersistentServices = mConfigurations.getOtherPersistenServices();
				for (String serviceName : otherPersistentServices) {
					PackageUtils.startServiceIfNotRunning(mContext, serviceName);
				}
			}
		});

	}
	
	/* spend too much time !! 60+ms
	private String getProcessName() {
		ActivityManager am = (ActivityManager) this.getSystemService(Context.ACTIVITY_SERVICE);
		int pid = android.os.Process.myPid();
		List<RunningAppProcessInfo> infos = am.getRunningAppProcesses();
		for (int i = 0; i < infos.size(); i++) {
			RunningAppProcessInfo info = infos.get(i);
			if(pid == info.pid) {
				return info.processName;
			}
		}
		return null;
	}
	*/

	private String mCurrProcessName;
	public String getProcessName() {
		if (TextUtils.isEmpty(mCurrProcessName)) {
			try {
				File file = new File("/proc/" + android.os.Process.myPid() + "/" + "cmdline");
				mBufferedReader = new BufferedReader(new FileReader(file));
				mCurrProcessName = mBufferedReader.readLine();
			} catch (Exception e) {
				e.printStackTrace();
			}
			if (mCurrProcessName != null) {
				mCurrProcessName = mCurrProcessName.trim();
			}
		}
		return mCurrProcessName;
	}
	
	/**
	 * release reader IO
	 */
	private void releaseIO() {
		if (mBufferedReader != null) {
			try {
				mBufferedReader.close();
			} catch (IOException e) {
				e.printStackTrace();
			}
			mBufferedReader = null;
		}
	}

    /**
     * 是否开启守护功能
     * @param context
     * @return
     */
	public boolean isDaemonPermitting(Context context) {
//		SharedPreferences sp = context.getSharedPreferences(DaemonConstants.SP_FILENAME, Context.MODE_PRIVATE);
		SharedPreferences sp = MultiprocessSharedPreferences.getSharedPreferences(context, DaemonConstants.SP_FILENAME, Context.MODE_PRIVATE);
		return sp.getBoolean(DaemonConstants.SP_KEY_PERMITTING, true);
	}

    /**
     * 设置是否开启守护功能
     *
     * @param context
     * @param isPermitting 是否开启，false时若守护已启动则关停。true时要下次{@link #onAttachBaseContext}被调用才生效
     */
	public void setDaemonPermiiting(Context context, boolean isPermitting) {
		LogUtils.i("daemon", "[DaemonClient#setDaemonPermiiting] isPermitting: " + isPermitting);

//		SharedPreferences sp = context.getSharedPreferences(DaemonConstants.SP_FILENAME, Context.MODE_PRIVATE);
		SharedPreferences sp = MultiprocessSharedPreferences.getSharedPreferences(context, DaemonConstants.SP_FILENAME, Context.MODE_PRIVATE);
		Editor editor = sp.edit();
		editor.putBoolean(DaemonConstants.SP_KEY_PERMITTING, isPermitting);
        editor.commit();

		if (!isPermitting) {
			IDaemonStrategy strategy = IDaemonStrategy.Fetcher.fetchStrategy();
			strategy.cancelDaemon(context);
		}
	}

	/**
	 * 设置forgroundService为前台服务，在{@link Service#onCreate()}里调用
	 * @param forgroundService
	 * @param innerServiceClass
	 * @备注 forgroundService和innerServiceClass必须配置在同一个进程内
	 * @备注 开了守护的应用，不建议再设置前台进程，会影响用户升级
     */
	@Deprecated
	public void setForgroundService(Service forgroundService, Class<? extends NotificationAssistService> innerServiceClass) {
		NotificationAssistService.setForgroundService(forgroundService, innerServiceClass);
	}

	/**
	 * 是否守护拉起的主常驻服务
	 */
	private Boolean mIsDaemonWork;
	/**
	 * 统计守护效果, 在{@link Service#onStartCommand(Intent, int, int)}和{@link android.content.BroadcastReceiver#onReceive(Context, Intent)}里调用
	 * @param associatedObj 传{@link Service}或{@link android.content.BroadcastReceiver}
	 * @param intent
     */
	public void statisticsDaemonEffect(Object associatedObj, Intent intent) {
		if (mIsDaemonWork != null || intent == null
				|| mConfigurations == null
				|| !mConfigurations.isStatisticsDaemonEffect()
				|| !mConfigurations.mPersistentConfig.mProcessName.equals(getProcessName())) {
			//初始化过mIsDaemonWork、参数异常、不开启统计、不是主常驻进程，即返回
			return;
		}
		mIsDaemonWork = PackageUtils.isFromDaemonSdk(intent);
		//TODO DEBUG
		if (mIsDaemonWork) {
			//守护SDK拉起的服务
			LogUtils.i("Daemon", "[DaemonClient::onServiceStartCommand] --->service start from DaemonSdk");
		} else {
			LogUtils.i("Daemon", "[DaemonClient::onServiceStartCommand] --->service start from outside of DaemonSdk");
		}
	}

	/**
	 * TODO DEBUG
	 */
	private void checkIfUnlock() {

	}

	/**
	 * TODO DEBUG
	 * 应用crash收集时,回调此接口,做保护处理。防止频繁crash
	 * @param e
     */
	public void onAppCrash(Throwable e) {
		if (e == null) {
			return;
		}

	}
}
