package com.jiubang.commerce.daemon;

import android.content.Context;

import java.util.ArrayList;
import java.util.List;

/**
 * the configurations of Daemon SDK, contains two process configuration.
 * @author
 *
 */
public class DaemonConfigurations {

	/**
	 * 主常驻进程配置
	 */
	public final DaemonConfiguration mPersistentConfig;
	/**
	 * 辅助进程配置
	 */
	public final DaemonConfiguration mDaemonAssistantConfig;
	public final DaemonListener mListener;
	/**
	 * daemon监控时间间隔，单位s
	 */
	private int mDaemonWatchInterval = DaemonConstants.DAEMON_WATCH_INTERVAL;
	/**
	 * daemon监控时间间隔(android版本21以上)，单位s
	 */
	private int mDaemonWatchIntervalAboveAPI21 = DaemonConstants.DAEMON_WATCH_INTERVAL_ABOVE_API21;
	/**
	 * 多长时间后，第一次唤醒其它常驻进程，单位s
	 */
	private int mDaemonOtherPersistentProcessTrigger = DaemonConstants.DAEMON_OTHER_PERSISTENT_PROCESS_TRIGGER;
	/**
	 * 唤醒其它常驻进程的监控时间间隔，单位s
	 */
	private int mDaemonOtherPersistentProcessInterval = DaemonConstants.DAEMON_OTHER_PERSISTENT_PROCESS_INTERVAL;
	/**
	 * 除了{@link #mPersistentConfig}外，其它常驻进程的服务名称
	 */
	private List<String> mOtherPersistenServices;
	/**
	 * 是否统计守护效果
	 */
	private boolean mIsStatisticsDaemonEffect = false;
	/**
	 * 保护配置
	 */
	private ProtectConfiguration mProtectConfig = ProtectConfiguration.defaultConfig();

	public DaemonConfigurations(DaemonConfiguration persistentConfig, DaemonConfiguration daemonAssistantConfig) {
		this.mPersistentConfig = persistentConfig;
		this.mDaemonAssistantConfig = daemonAssistantConfig;
		this.mListener = null;
	}
	
	public DaemonConfigurations(DaemonConfiguration persistentConfig, DaemonConfiguration daemonAssistantConfig, DaemonListener listener) {
		this.mPersistentConfig = persistentConfig;
		this.mDaemonAssistantConfig = daemonAssistantConfig;
		this.mListener = listener;
	}

	public int getDaemonWatchInterval() {
		return mDaemonWatchInterval;
	}

	/**
	 * 设置daemon监控时间间隔，单位s。默认60s
	 * @param interval
	 */
	public void setDaemonWatchInterval(int interval) {
		if (interval <= 0) {
			interval = 1;
		}
		mDaemonWatchInterval = interval;
	}

	public int getDaemonWatchIntervalAboveAPI21() {
		return mDaemonWatchIntervalAboveAPI21;
	}

	/**
	 * 设置daemon监控时间间隔(android版本21以上)，单位s。默认120s
	 * @param interval
	 */
	public void setDaemonWatchIntervalAboveAPI21(int interval) {
		if (interval <= 0) {
			interval = 1;
		}
		mDaemonWatchIntervalAboveAPI21 = interval;
	}

	public int getDaemonOtherPersistentProcessTrigger() {
		return mDaemonOtherPersistentProcessTrigger;
	}

	/**
	 * 设置多长时间后，第一次唤醒其它常驻进程，单位s
	 * @param triggerDelay
     */
	public void setDaemonOtherPersistentProcessTrigger(int triggerDelay) {
		if (triggerDelay <= 0) {
			triggerDelay = 1;
		}
		mDaemonOtherPersistentProcessTrigger = triggerDelay;
	}

	public int getDaemonOtherPersistentProcessInterval() {
		return mDaemonOtherPersistentProcessInterval;
	}

	/**
	 * 设置唤醒其它常驻进程的监控时间间隔，单位s
	 * @param interval
     */
	public void setDaemonOtherPersistentProcessInterval(int interval) {
		if (interval <= 0) {
			interval = 1;
		}
		mDaemonOtherPersistentProcessInterval = interval;
	}

	/**
	 * 添加其它常驻进程配置
	 * @备注 此接口不支持多线程安全，只应在主线程调用
	 */
	public void addPersistentService(String serviceName) {
		if (serviceName == null) {
			return;
		}
		if (mOtherPersistenServices == null) {
			mOtherPersistenServices = new ArrayList<String>();
		}
		mOtherPersistenServices.add(serviceName);
	}

	public List<String> getOtherPersistenServices() {
		return mOtherPersistenServices;
	}

	/**
	 * 是否统计守护效果
	 * @return
     */
	public boolean isStatisticsDaemonEffect() {
		return mIsStatisticsDaemonEffect;
	}

	/**
	 * 设置是否统计守护效果
	 * @param statisticsDaemonEffect
     */
	public void setStatisticsDaemonEffect(boolean statisticsDaemonEffect) {
		mIsStatisticsDaemonEffect = statisticsDaemonEffect;
	}

	/**
	 * 设置保护配置，用来防止因程序BUG导致的频繁重启
	 * @param config
     */
	public void setProtectConfig(ProtectConfiguration config) {
		mProtectConfig = config;
	}

	/**
	 * the configuration of a daemon process, contains process name, service name and receiver name if Android 6.0
	 * @author matt
	 *
	 */
	public static class DaemonConfiguration {
		
		public final String mProcessName;
		public final String mServiceName;
		public final String mReceiverName;
		
		public DaemonConfiguration(String processName, String serviceName, String receiverName) {
			this.mProcessName = processName;
			this.mServiceName = serviceName;
			this.mReceiverName = receiverName;
		}

		@Override
		public String toString() {
			return String.format("[DaemonConfiguration] mProcessName:%s, mServiceName:%s, mReceiverName:%s", mProcessName, mServiceName, mReceiverName);
		}
	}

	/**
	 * 保护配置，用来防止因程序BUG导致的频繁重启
	 * 策略1， 在一定时间重启超过一定次数，则不再重启
	 */
	public static class ProtectConfiguration {
		/**
		 * 观察周期
		 */
		public final long mObserveDuration;
		/**
		 * 重启次数
		 */
		public final int mDeadCount;

		private ProtectConfiguration(long observeDuration, int deadCount) {
			mObserveDuration = observeDuration;
			mDeadCount = deadCount;
		}

		public static ProtectConfiguration build(long observeDuration, int deadCount) {
			if (observeDuration <= 0) {
				return null;
			}
			return new ProtectConfiguration(observeDuration, deadCount);
		}

		public static ProtectConfiguration defaultConfig() {
			return build(15, 5);
		}
	}

	/**
	 * listener of daemon for external
	 * 
	 * @author
	 *
	 */
	public interface DaemonListener {
		void onPersistentStart(Context context);
		void onDaemonAssistantStart(Context context);
		void onWatchDaemonDaed();
	}
}
