package com.jiubang.commerce.dyload.manager;

import android.app.Activity;
import android.content.Context;
import android.content.pm.PackageInfo;
import android.support.annotation.NonNull;
import android.text.TextUtils;
import android.util.AndroidRuntimeException;
import android.util.Log;

import com.jb.ga0.commerce.util.AppUtils;
import com.jb.ga0.commerce.util.LogUtils;
import com.jb.ga0.commerce.util.io.FileUtil;
import com.jb.ga0.commerce.util.thread.CustomThreadExecutorProxy;
import com.jb.ga0.commerce.util.thread.PriorityThreadPool;
import com.jiubang.commerce.dyload.core.DyIntent;
import com.jiubang.commerce.dyload.core.proxy.activity.DyActivityPlugin;
import com.jiubang.commerce.dyload.core.proxy.activity.StandarProxyActivity;
import com.jiubang.commerce.dyload.manager.DmConstants.ActivityResult;
import com.jiubang.commerce.dyload.manager.DmConstants.DyloadError;
import com.jiubang.commerce.dyload.manager.DmConstants.DyloadStatus;
import com.jiubang.commerce.dyload.update.AbsClientParams;
import com.jiubang.commerce.dyload.update.PluginDownloader;
import com.jiubang.commerce.dyload.update.PluginUpdateInfo;
import com.jiubang.commerce.dyload.update.PluginUpdateMgr;
import com.jiubang.commerce.dyload.update.abtest.AbTestHttpHandler;
import com.jiubang.commerce.dyload.util.DexUtil;
import com.jiubang.commerce.dyload.util.DyConstants;
import com.jiubang.commerce.dyload.util.DyStatistics;
import com.jiubang.commerce.dyload.util.DyloadUtil;
import com.jiubang.commerce.dyload.util.LogTag;
import com.jiubang.commerce.dyload.util.MD5Util;

import java.io.File;
import java.io.FileFilter;
import java.io.IOException;
import java.io.InputStream;
import java.util.Collection;
import java.util.Comparator;
import java.util.HashMap;
import java.util.Iterator;
import java.util.LinkedList;
import java.util.List;
import java.util.Map;
import java.util.PriorityQueue;
import java.util.concurrent.ConcurrentHashMap;
import java.util.concurrent.CopyOnWriteArrayList;

import static com.jiubang.commerce.dyload.util.LogTag.TEST_TAG;

/**
 * 动态加载管理器
 * <br/>异步加载插件，插件接口自行做同步处理，提高性能
 * <br/>支持懒加载
 * <br/>sd卡插件目录下，每种插件只放一个文件。新版本插件更新后，需要将原插件删除
 * <br/>新版本插件加载失败，会尝试使用旧版本插件（实现方式：备用sd卡插件目录，新版本插件更新后，需要将原插件删除移动到备用sd卡插件目录）
 * <br/>插件进程优化（只加载相关插件，如充电锁进程只加载充电锁插件）
 * <br/>优化调用次数：DexUtil.getDexPackageInfo
 * todo 测试{@link DexUtil#getDexPackageInfo(Context, String)}执行耗时，看有无必要
 * <br/>优化：插件名称含包名版本号标识，无需apk包应用信息即可识别包名版本号，用以提高插件加载速度
 * <p>
 * Created by matt on 17/3/27.
 */
public class DyManager implements PluginDownloader.IPluginDownloadListener {
    private static final String TAG = LogTag.TAG;
    /**
     * 是否异步加载插件
     */
    private static final boolean LOAD_PLUGIN_ASYNC = true;

    private static DyManager sInstance;

    private Context mHostContext;
    /**
     * 是否已初始化
     */
    private volatile boolean mIsInited = false;
    /**
     * 插件加载配置map，key为插件包名
     * 相同配置，以序号靠前的为准
     * 以mPluginConfigsHolder是否为null来判断插件配置是否加载完成
     */
    Map<String, PluginConfig> mPluginConfigsHolder;
    byte[] mPluginConfigsWait = new byte[0];
    /**
     * 插件map，key为插件包名
     */
    Map<String, DyPluginInfo> mPluginsHolder = new ConcurrentHashMap<>();
    /**
     * 插件加载状态
     */
    Map<String, Integer> mPluginLoadStatusMap = new ConcurrentHashMap<>();
    /**
     * 插件加载监听器列表
     */
    List<IPluginLoadListener> mInstallListeners = new CopyOnWriteArrayList();
    /**
     * 插件重加载监听器
     */
    Map<String, IPluginReloadListener> mReloadListenerMap = new ConcurrentHashMap<>();
    /**
     * 已经拷贝过的应用内插件，用于防止多次拷贝
     */
    List<String> mCopiedInnerPluginFile = new CopyOnWriteArrayList<>();
    /**
     * 插件更新管理器
     */
    PluginUpdateMgr mPluginUpdateMgr;
    private volatile int mAutoLoadFinishFlag;
    private AbsClientParams mClientParams;

    private DyManager(Context context) {
        mHostContext = context.getApplicationContext();
        mPluginUpdateMgr = new PluginUpdateMgr(context, this);

        DyloadConfig.getInstance().initFromXml(mHostContext);
        initPluginLoader();
    }

    public static DyManager getInstance(Context context) {
        if (sInstance == null) {
            synchronized (DyManager.class) {
                if (sInstance == null) {
                    sInstance = new DyManager(context);
                }
            }
        }
        return sInstance;
    }

    /**
     * 初始化，可重复调用
     */
    public void init() {
        LogUtils.i(TAG, "DyManager#init() called");

        if (mIsInited) {
            return;
        }
        mIsInited = true;

        mPluginUpdateMgr.init();
    }

    /**
     * 设置客户端参数
     * @param params
     */
    public void setClientParams(AbsClientParams params) {
        LogUtils.i(TEST_TAG, "DyManager#setClientParams() called with: params = [" + params + "]");

        AbsClientParams oldParams = mClientParams;
        mClientParams = params;
        mPluginUpdateMgr.setClientParams(params, oldParams);
    }

    public AbsClientParams getClientParams() {
        return mClientParams;
    }

    /**
     * ab服务使用测服
     */
    public static void setUseTestAbServer(boolean useTestAbServer) {
        AbTestHttpHandler.sIS_TEST_SERVER = useTestAbServer;
    }

    /**
     * 检查更新内置插件包
     * <br/> 当新版本插件内置包更新时，插件应该调用此接口来检查是否需要更新
     * <br/> 异步执行
     * <br/>
     *
     * @param pluginPkgName
     * @param isOverridePluginOnline 如果版本号高于已下载的线上包，是否要覆盖
     */
    public void checkInnerPluginUpdate(final String pluginPkgName, final boolean isOverridePluginOnline) {
        LogUtils.i(TAG, "DyManager#checkInnerPluginUpdate() called with: pluginPkgName = [" + pluginPkgName + "], isOverridePluginOnline = [" + isOverridePluginOnline + "]");
        if (pluginPkgName == null) {
            return;
        }

        CustomThreadExecutorProxy.getInstance().execute(new Runnable() {
            @Override
            public void run() {
                checkInnerPluginUpdateSync(pluginPkgName, isOverridePluginOnline);
            }
        });
    }

    /**
     * @param pluginPkgName
     * @param isOverridePluginOnline todo debug 未使用
     */
    private void checkInnerPluginUpdateSync(String pluginPkgName, boolean isOverridePluginOnline) {
        String tempFilePath = DyConstants.getTempDir(mHostContext) + pluginPkgName.hashCode()
                + "_" + System.currentTimeMillis() + ".jar";
        boolean copySucc = copyInnerPluginFile(getPluginConfig(pluginPkgName), tempFilePath);
        if (!copySucc) {
            return;
        }
        PackageInfo newPackageInfo = DexUtil.getDexPackageInfo(mHostContext, tempFilePath);
        //读取完信息即删除临时文件
        FileUtil.delFile(tempFilePath);
        if (newPackageInfo == null) {
            return;
        }
        int newVersionCode = newPackageInfo.versionCode;

        //从内存找
        DyPluginInfo dyPluginInfo = getPluginInfo(pluginPkgName);
        if (dyPluginInfo != null) {
            if (newVersionCode > dyPluginInfo.mVersionCode) {
                FileUtil.delFile(dyPluginInfo.getFileAbsolutePath());
            }
            return;
        }

        //从本地文件找
        String path = DyConstants.getPluginLoadDir(mHostContext);
        File[] files = new File(path).listFiles(new FileFilter() {
            @Override
            public boolean accept(File pathname) {
                return !isPluginFileLoaded(pathname.getAbsolutePath());
            }
        });

        if (files == null) {
            return;
        }

        for (File file : files) {
            PackageInfo pi = DexUtil.getDexPackageInfo(mHostContext, file.getAbsolutePath());
            if (pi != null && pluginPkgName.equals(pi.packageName)) {
                if (newVersionCode > dyPluginInfo.mVersionCode) {
                    FileUtil.delFile(file.getAbsolutePath());
                }
                return;
            }
        }
    }

    /**
     * 初始化插件加载
     */
    private void initPluginLoader() {
        Runnable r = new Runnable() {
            @Override
            public void run() {
                mPluginConfigsHolder = DyConfigParaser.parse(mHostContext, true);
                LogUtils.i(TAG, "[DyManager#initPluginLoader] <wait-notify> notifyAll at mPluginConfigsWait-" + mPluginConfigsWait.hashCode());
                synchronized (mPluginConfigsWait) {
                    mPluginConfigsWait.notifyAll();
                }

                loadPlugins();
            }
        };

        if (LOAD_PLUGIN_ASYNC) {
            new Thread(r).start();
        } else {
            r.run();
        }
    }

    /**
     * 获取插件加载配置信息
     *
     * @param pluginPkgName
     * @return
     */
    public PluginConfig getPluginConfig(String pluginPkgName) {
        if (mPluginConfigsHolder == null || pluginPkgName == null) {
            return null;
        }
        return mPluginConfigsHolder.get(pluginPkgName);
    }

    /**
     * 等待插件配置加载完成
     */
    public void awaitPluginConfigsLoaded() {
        if (mPluginConfigsHolder != null) {
            return;
        }
        synchronized (mPluginConfigsWait) {
            try {
                mPluginConfigsWait.wait();
            } catch (InterruptedException e) {
                e.printStackTrace();
            }
        }
    }

    public Collection<PluginConfig> getAllPluginConfigs() {
        return mPluginConfigsHolder.values();
    }

    @Override
    public void onPluginDownloadSucc(PluginUpdateInfo pluginUpdateInfo) {
        LogUtils.i(TAG, "DyManager#onPluginDownloadSucc() called with: pluginUpdateInfo = [" + pluginUpdateInfo + "]");

        String pluginPkgName = pluginUpdateInfo.mPackageName;
        //删除旧插件
        deleteLocalPluginFile(pluginPkgName);
        //拷贝新插件
        String pluginLoadPath = pluginUpdateInfo.getPluginLoadPath(mHostContext);
        FileUtil.copyFile(pluginUpdateInfo.getDownloadFilePath(mHostContext), pluginLoadPath);

        hotUpdatePlugin(pluginPkgName, pluginUpdateInfo);
    }

    /**
     * 处理没有版本更新信息的插件
     *
     * @param pluginPkgName
     */
    public void handlePluginOnlineCleared(String pluginPkgName) {
        LogUtils.i(TAG, "DyManager#handlePluginOnlineCleared() called with: pluginPkgName = [" + pluginPkgName + "]");

        hotUpdatePlugin(pluginPkgName, null);
    }

    /**
     * 热更新插件
     *
     * @param pluginPkgName
     * @param pluginUpdateInfo
     */
    private void hotUpdatePlugin(final String pluginPkgName, final PluginUpdateInfo pluginUpdateInfo) {
        PluginConfig pluginConfig = getPluginConfig(pluginPkgName);
        if (pluginConfig == null) {
            LogUtils.w(TAG, "[DyManager#hotUpdatePlugin] cannot find PluginConfig with pkgName:"
                    + pluginPkgName);
            return;
        }

        if (pluginConfig.isHotUpdate()) {
            if (isPluglinLoadFinish(pluginPkgName)) {
                hotUpdatePluginSync(pluginPkgName, pluginUpdateInfo);
            } else if (isPluginLoading(pluginPkgName)) {
                //todo 正在加载该插件，延迟5s再加载新插件
                CustomThreadExecutorProxy.getInstance().runOnAsyncThread(new Runnable() {
                    @Override
                    public void run() {
                        hotUpdatePluginSync(pluginPkgName, pluginUpdateInfo);
                    }
                }, 5000);
            }
        } else {
            //不热更新的插件，下载成功即当做升级成功
            int versionCode = 0;
            if (pluginUpdateInfo != null) {
                versionCode = pluginUpdateInfo.mVersionNumber;
            }
            DyStatistics.uploadPluginUpdate(mHostContext, pluginPkgName, versionCode, true);

            LogUtils.i(TEST_TAG, "插件不支持热更新，下次应用启动再运行新版本，包名：" + pluginPkgName);
        }
    }

    private void hotUpdatePluginSync(String pluginPkgName, PluginUpdateInfo pluginUpdateInfo) {
        String pluginFilePath = pluginUpdateInfo != null ? pluginUpdateInfo.getPluginLoadPath(mHostContext) : null;
        int loadResult = loadPlugin(pluginPkgName, pluginFilePath, true);
        boolean updateSucc = loadResult == DmConstants.LoadPluginResult.load_SdPlugin_Success;
        //上传升级统计
        int versionCode = 0;
        if (pluginUpdateInfo != null) {
            versionCode = pluginUpdateInfo.mVersionNumber;
        } else if (updateSucc) {
            DyPluginInfo dyPluginInfo = getPluginInfo(pluginPkgName);
            if (dyPluginInfo != null) {
                versionCode = dyPluginInfo.mVersionCode;
            }
        }
        DyStatistics.uploadPluginUpdate(mHostContext, pluginPkgName, versionCode, updateSucc);
    }

    /**
     * 等待插件加载完，有最长超时限制
     * <br/>配置autoload的插件，自动加载完相应插件即返回
     * <br/>未配置autoload的插件，{@link #loadPlugin(String)}结束即返回
     * <br/>未配置autoload的插件，调用此接口前，必须先调用{@link #loadPlugin(String)}
     *
     * @param pluginPkgName
     */
    public void awaitLoadedLocked(String pluginPkgName) {
//        LogUtils.d(TAG, "DyManager#awaitLoadedLocked() called with: pluginPkgName = [" + pluginPkgName + "]");

        if (getPluginInfo(pluginPkgName) != null) {
            return;
        }

        //等待插件配置文件解析完成
        if (mPluginConfigsHolder == null) {
            LogUtils.i(TAG, "[DyManager#awaitLoadedLocked] <wait-notify> wait on mPluginConfigsWait-" + mPluginConfigsWait.hashCode());
            synchronized (mPluginConfigsWait) {
                try {
                    mPluginConfigsWait.wait();
                } catch (InterruptedException e) {
                    e.printStackTrace();
                }
            }
        }
        PluginConfig config = getPluginConfig(pluginPkgName);
        if (config == null) {
            //不存在插件配置文件，不会加载该插件，直接返回
            return;
        }

        //等待插件加载完成
        if (config.isAutoLoad()) {
            //自动加载插件
            if (isPluglinLoadFinish(pluginPkgName)) {
                //已加载完成
                return;
            } else {
                LogUtils.i(TAG, "[DyManager#awaitLoadedLocked] <wait-notify> wait on 1-PluginConfig-" + config.hashCode());
                synchronized (config) {
                    try {
                        config.wait();
                    } catch (InterruptedException e) {
                        e.printStackTrace();
                    }
                }
            }
        } else {
            if (isPluglinLoadFinish(pluginPkgName)) {
                //已加载完成
                return;
            } else if (isInvokeLoadPlugin(pluginPkgName)) {
                //未加载完成，但已调用了loadPlugin接口
                LogUtils.i(TAG, "[DyManager#awaitLoadedLocked] <wait-notify> wait on 2-PluginConfig-" + config.hashCode());
                synchronized (config) {
                    try {
                        config.wait();
                    } catch (InterruptedException e) {
                        e.printStackTrace();
                    }
                }
            } else {
                //还未调用过loadPlugin接口
                throw new IllegalStateException("invoke DyManager#loadPlugin bebore DyManager#awaitLoadedLocked");
            }
        }
    }

    private int getPluginLoadStatus(String pluginPkgName) {
        Integer status = mPluginLoadStatusMap.get(pluginPkgName);
        if (status == null) {
            return DyloadStatus.PLUGIN_LOAD_STATUS_INIT;
        }
        return status;
    }

    private void setPluginLoadStatus(String pluginPkgName, int status) {
        LogUtils.i(TAG, "DyManager#setPluginLoadStatus() called with: pluginPkgName = [" + pluginPkgName + "], status = [" + status + "]");

        mPluginLoadStatusMap.put(pluginPkgName, status);
    }

    /**
     * 是否正在加载插件
     *
     * @param pluginPkgName
     */
    private boolean isPluginLoading(String pluginPkgName) {
        return getPluginLoadStatus(pluginPkgName) == DyloadStatus.PLUGIN_LOAD_STATUS_START;
    }

    /**
     * 插件是否加载完成（可能成功也可能失败）
     *
     * @param pluginPkgName
     * @return
     */
    private boolean isPluglinLoadFinish(String pluginPkgName) {
        return getPluginLoadStatus(pluginPkgName) == DyloadStatus.PLUGIN_LOAD_STATUS_FINISH;
    }

    /**
     * 是否调用过{@link #loadPlugin(String)}
     *
     * @param pluginPkgName
     * @return
     */
    private boolean isInvokeLoadPlugin(String pluginPkgName) {
        return getPluginLoadStatus(pluginPkgName) == DyloadStatus.PLUGIN_LOAD_STATUS_INVOKE;
    }

    /**
     * 加载插件，用于懒加载插件
     *
     * @param pluginPkgName
     */
    public void loadPlugin(final String pluginPkgName) {
        setPluginLoadStatus(pluginPkgName, DyloadStatus.PLUGIN_LOAD_STATUS_INVOKE);
        if (LOAD_PLUGIN_ASYNC) {
            new Thread() {
                @Override
                public void run() {
                    loadPlugin(pluginPkgName, null, false);
                }
            }.start();
        } else {
            loadPlugin(pluginPkgName, null, false);
        }
    }

    /**
     * 加载所有插件，已经加载进来的不重新加载
     */
    private void loadPlugins() {
        LogUtils.i(TAG, "DyManager#loadPlugins() called");
        notifyAutoLoadStart();

        String path = DyConstants.getPluginLoadDir(mHostContext);

        final Map<String, PluginScanEntry> pluginEntryMap = new HashMap<>();
        File[] files = new File(path).listFiles(new FileFilter() {
            @Override
            public boolean accept(File file) {
                String fileName = file.getName();
                if (!fileName.endsWith(".jar")) {
                    LogUtils.w(TAG, "[DyManager#accept] invalid plugin file name:" + fileName);
                    return false;
                }

                PackageInfo packageInfo = null;
                //不再处理已经加载的插件
                if (!isPluginFileLoaded(file.getAbsolutePath())) {
                    packageInfo = DexUtil.getDexPackageInfo(mHostContext, file.getAbsolutePath());
                }
                if (packageInfo != null) {
                    pluginEntryMap.put(packageInfo.packageName, new PluginScanEntry(file, packageInfo));
                    return true;
                }
                return false;
            }
        });

        Collection<PluginConfig> collection = mPluginConfigsHolder.values();
        PriorityQueue<PluginConfig> queue = new PriorityQueue<PluginConfig>(
                //防止size为0
                Math.max(collection.size(), 1),
                new Comparator<PluginConfig>() {
                    @Override
                    public int compare(PluginConfig lhs, PluginConfig rhs) {
                        return rhs.mAutoLoadPriority - lhs.mAutoLoadPriority;
                    }
                });
        queue.addAll(collection);

//        for (PluginConfig pluginConfig : mPluginConfigsHolder.values()) {
        class AutoLoadRunnable implements Runnable, Comparable<AutoLoadRunnable> {
            PluginConfig mPluginConfig;
            PluginScanEntry mPluginEntry;

            AutoLoadRunnable(PluginConfig pluginConfig, PluginScanEntry pluginEntry) {
                mPluginConfig = pluginConfig;
                mPluginEntry = pluginEntry;
            }

            @Override
            public int compareTo(@NonNull AutoLoadRunnable another) {
                return another.mPluginConfig.mAutoLoadPriority - mPluginConfig.mAutoLoadPriority;
            }

            @Override
            public void run() {
                loadPlugin(mPluginConfig, mPluginEntry);

                mAutoLoadFinishFlag--;
                if (mAutoLoadFinishFlag == 0) {
                    notifyAutoLoadFinish();
                }
            }
        }
        mAutoLoadFinishFlag = queue.size();
        PriorityThreadPool threadPool = new PriorityThreadPool(DyloadConfig.getInstance().getMaxAutoLoadThreadCount());
        while (!queue.isEmpty()) {
            PluginConfig pluginConfig = queue.poll();
//            loadPlugin(pluginConfig, pluginEntryMap.get(pluginConfig.mPluginPkgName));
            threadPool.submit(new AutoLoadRunnable(pluginConfig, pluginEntryMap.get(pluginConfig.mPluginPkgName)));
        }

//        notifyAutoLoadFinish();
    }

    private void loadPlugin(PluginConfig pluginConfig, PluginScanEntry pluginEntry) {
        String pluginPkgName = pluginConfig.mPluginPkgName;
        if (isPluginFileLoaded(pluginPkgName)) {
            //插件已加载
            return;
        }
        if (!pluginConfig.isAutoLoad()) {
            //不自动加载
            return;
        }

        notifyLoadStart(pluginPkgName);

        if (pluginEntry == null) {
            notifyLoadSdFileError(pluginPkgName, 0,
                    DyloadError.ERRORCODE_NO_PLUGINFILE,
                    DyloadError.ERRORMSG_NO_PLUGINFILE);
            loadInnerPlugin(pluginConfig);
            return;
        }
        File pluginFile = pluginEntry.mFile;
        //load dex前再检查一下文件是否存在
        if (pluginFile == null || !pluginFile.exists()) {
            notifyLoadSdFileError(pluginPkgName, 0,
                    DyloadError.ERRORCODE_NO_PLUGINFILE,
                    DyloadError.ERRORMSG_NO_PLUGINFILE);
            loadInnerPlugin(pluginConfig);
            return;
        }
        PackageInfo pluginPkgInfo = pluginEntry.mPackageInfo;

        //检查是否使用内置插件覆盖升级
        if (pluginPkgInfo.versionCode < pluginConfig.mInnerPluginVersion) {
            LogUtils.i(TEST_TAG, "内置插件升级：" + pluginPkgName + ", 内置插件版本号：" + pluginConfig.mInnerPluginVersion
                    + ", sd卡插件版本号：" + pluginPkgInfo.versionCode);
            FileUtil.delFile(pluginFile.getAbsolutePath());
            loadInnerPlugin(pluginConfig);
            return;
        }

        //加载插件
        DyPluginInfo pluginInfo = PluginInfoFactory.create(mHostContext, pluginFile, pluginPkgInfo, pluginConfig);
        pluginInfo.attachContext(mHostContext);
        LogUtils.i(TAG, "[DyManager#loadPlugins] unchecked plugin [" + pluginInfo + "]");
        if (!pluginInfo.isValid()) {
            pluginInfo.destroy(mHostContext);
            FileUtil.delFile(pluginInfo.getFileAbsolutePath());
            notifyLoadSdFileError(pluginPkgName, pluginPkgInfo.versionCode,
                    DyloadError.ERRORCODE_CREATE_DYPLUGIN_FAIL,
                    DyloadError.ERRORMSG_CREATE_DYPLUGIN_FAIL);
            loadInnerPlugin(pluginConfig);
            return;
        }

        //检查框架版本号
        if (!pluginConfig.supportItfVersion(pluginInfo.getInterfaceVersion())) {
            FileUtil.delFile(pluginFile.getAbsolutePath());
            notifyLoadSdFileError(pluginPkgName, pluginPkgInfo.versionCode,
                    DyloadError.ERRORCODE_INTERFACE_VERSION_NOT_MATCH,
                    getErrorMsgInterfaceVersionNotMatch(pluginConfig, pluginInfo));
            loadInnerPlugin(pluginConfig);
            return;
        }

        //保存已加载插件
        mPluginsHolder.put(pluginInfo.getPackageName(), pluginInfo);
        notifyLoadSuccess(pluginPkgName);
    }

    private static String getErrorMsgInterfaceVersionNotMatch(PluginConfig pluginConfig, DyPluginInfo dyPluginInfo) {
        return String.format(DyloadError.ERRORMSG_INTERFACE_VERSION_NOT_MATCH,
                pluginConfig.getMinInterfaceVersion(),
                pluginConfig.getTargetInterfaceVersion(),
                dyPluginInfo.getInterfaceVersion()
        );
    }

    /**
     * 加载插件，如果找不到插件，则加载内置插件
     *
     * @param pluginPkgName
     * @param pluginFilePath
     * @param isUpdate       已经加载的插件，是否更新
     */
    private int loadPlugin(String pluginPkgName, String pluginFilePath, boolean isUpdate) {
        LogUtils.i(TAG, "DyManager#loadPlugin() called with: pluginPkgName = [" + pluginPkgName + "], isUpdate = [" + isUpdate + "]");

        if (pluginPkgName == null) {
            notifyLoadFailed(pluginPkgName, 0, DyloadError.ERRORCODE_NULL_PKGNAME,
                    DyloadError.ERRORMSG_NULL_PKGNAME);
            return DmConstants.LoadPluginResult.load_Plugin_Fail;
        }

        notifyLoadStart(pluginPkgName);

        //检查插件是否已加载
        DyPluginInfo dyPluginInfo = getPluginInfo(pluginPkgName);
        if (dyPluginInfo != null && !isUpdate) {
            notifyLoadFailed(pluginPkgName, 0, DyloadError.ERRORCODE_PLUGIN_LOADED,
                    DyloadError.ERRORMSG_PLUGIN_LOADED);
            return DmConstants.LoadPluginResult.load_Plugin_Fail;
        }
        if (dyPluginInfo != null) {
            //重新加载
            dyPluginInfo.beforeHotUpdate();
            IPluginReloadListener reloadListener = getReloadListener(pluginPkgName);
            if (reloadListener != null) {
                reloadListener.beforeReloadPlugin(pluginPkgName);
            }
            dyPluginInfo.destroy(mHostContext);
            mPluginsHolder.remove(dyPluginInfo);
            dyPluginInfo = null;
        }

        //查找插件配置
        PluginConfig pluginConfig = getPluginConfig(pluginPkgName);
        if (pluginConfig == null) {
            notifyLoadFailed(pluginPkgName, 0, DyloadError.ERRORCODE_NO_CONFIG,
                    DyloadError.ERRORMSG_NO_CONFIG);
            return DmConstants.LoadPluginResult.load_Plugin_Fail;
        }

        File pluginFile = null;
        PackageInfo pluginPkgInfo = null;
        if (pluginFilePath != null) {
            //插件更新时，从外部传入新插件文件路径，无需再查找
            pluginPkgInfo = DexUtil.getDexPackageInfo(mHostContext, pluginFilePath);
            if (pluginPkgInfo != null && pluginPkgName.equals(pluginPkgInfo.packageName)) {
                pluginFile = new File(pluginFilePath);
            } else {
                FileUtil.delFile(pluginFilePath);
            }
        } else {
            //查找插件文件文件
            String path = DyConstants.getPluginLoadDir(mHostContext);
            File[] files = new File(path).listFiles();
            if (files != null && files.length > 0) {
                for (File file : files) {
                    PackageInfo pi = DexUtil.getDexPackageInfo(mHostContext, file.getAbsolutePath());
                    if (pi != null && pluginPkgName.equals(pi.packageName)) {
                        pluginFile = file;
                        pluginPkgInfo = pi;
                    }
                }
            }
        }
        if (pluginFile == null || pluginPkgInfo == null) {
            if (pluginFilePath != null) {
                notifyLoadSdFileError(pluginPkgName, 0,
                        DyloadError.ERRORCODE_GET_PACKAGEINFO,
                        DyloadError.ERRORMSG_GET_PACKAGEINFO);
            } else {
                notifyLoadSdFileError(pluginPkgName, 0,
                        DyloadError.ERRORCODE_NO_PLUGINFILE,
                        DyloadError.ERRORMSG_NO_PLUGINFILE);
            }
            return loadInnerPlugin(pluginConfig) ?
                    DmConstants.LoadPluginResult.load_InnerPlugin_Success :
                    DmConstants.LoadPluginResult.load_Plugin_Fail;
        }

        //检查是否使用内置插件覆盖升级
        if (pluginPkgInfo.versionCode < pluginConfig.mInnerPluginVersion) {
            LogUtils.i(TEST_TAG, "内置插件升级：" + pluginPkgName + ", 内置插件版本号：" + pluginConfig.mInnerPluginVersion
                    + ", sd卡插件版本号：" + pluginPkgInfo.versionCode);
            FileUtil.delFile(pluginFile.getAbsolutePath());
            return loadInnerPlugin(pluginConfig) ?
                    DmConstants.LoadPluginResult.load_InnerPlugin_Success :
                    DmConstants.LoadPluginResult.load_Plugin_Fail;
        }

        //加载插件
        DyPluginInfo pluginInfo = PluginInfoFactory.create(mHostContext, pluginFile, pluginPkgInfo, pluginConfig);
        pluginInfo.attachContext(mHostContext);
        LogUtils.i(TAG, "[DyManager#loadPlugin] unchecked plugin [" + pluginInfo + "]");
        if (!pluginInfo.isValid()) {
            pluginInfo.destroy(mHostContext);
            FileUtil.delFile(pluginInfo.getFileAbsolutePath());
            notifyLoadSdFileError(pluginPkgName, pluginPkgInfo.versionCode,
                    DyloadError.ERRORCODE_CREATE_DYPLUGIN_FAIL,
                    DyloadError.ERRORMSG_CREATE_DYPLUGIN_FAIL);
            return loadInnerPlugin(pluginConfig) ?
                    DmConstants.LoadPluginResult.load_InnerPlugin_Success :
                    DmConstants.LoadPluginResult.load_Plugin_Fail;
        }

        //检查框架版本号
        if (!pluginConfig.supportItfVersion(pluginInfo.getInterfaceVersion())) {
            FileUtil.delFile(pluginFile.getAbsolutePath());
            notifyLoadSdFileError(pluginPkgName, pluginPkgInfo.versionCode,
                    DyloadError.ERRORCODE_INTERFACE_VERSION_NOT_MATCH,
                    getErrorMsgInterfaceVersionNotMatch(pluginConfig, pluginInfo));
            return loadInnerPlugin(pluginConfig) ?
                    DmConstants.LoadPluginResult.load_InnerPlugin_Success :
                    DmConstants.LoadPluginResult.load_Plugin_Fail;
        }

        //保存已加载插件
        mPluginsHolder.put(pluginInfo.getPackageName(), pluginInfo);
        notifyLoadSuccess(pluginPkgName);

        return DmConstants.LoadPluginResult.load_SdPlugin_Success;
    }

    /**
     * 加载内置插件
     *
     * @param pluginConfig todo 考虑内置插件加载失败后，记录下来，不重复加载
     */
    private boolean loadInnerPlugin(PluginConfig pluginConfig) {
        LogUtils.d(TAG, "DyManager#loadInnerPlugin() called with: pluginConfig = [" + pluginConfig + "]");

        if (pluginConfig == null || pluginConfig.mPluginPkgName == null) {
            //这里由上一级调用保证，应不会执行到
            Log.e(TAG, "[DyManager#loadInnerPlugin] parameter is null");
            return false;
        }

        File pluginFile = copyInnerPluginFile(pluginConfig);
        if (pluginFile == null) {
            notifyLoadFailed(pluginConfig.mPluginPkgName, 0, DyloadError.ERRORCODE_NO_INNER_PLUGINFILE,
                    DyloadError.ERRORMSG_NO_INNER_PLUGINFILE);
            return false;
        }

        mCopiedInnerPluginFile.add(pluginConfig.mPluginFileNameInAssets);

        PackageInfo pluginPkgInfo = DexUtil.getDexPackageInfo(mHostContext, pluginFile.getAbsolutePath());
        if (pluginPkgInfo == null || pluginConfig.mPluginPkgName == null
                || !pluginConfig.mPluginPkgName.equals(pluginPkgInfo.packageName)) {
            FileUtil.delFile(pluginFile.getAbsolutePath());
            notifyLoadFailed(pluginConfig.mPluginPkgName, pluginPkgInfo == null ? 0 : pluginPkgInfo.versionCode,
                    DyloadError.ERRORCODE_GET_PACKAGEINFO,
                    DyloadError.ERRORMSG_GET_PACKAGEINFO);
            return false;
        }

        if (pluginConfig.mInnerPluginVersion != 0 && pluginConfig.mInnerPluginVersion != pluginPkgInfo.versionCode) {
            throw new IllegalArgumentException("\n插件配置错误！[" + pluginConfig.mPluginPkgName
                    + "]内置插件版本号为:" + pluginPkgInfo.versionCode
            + ", 而dyloadconfig.xml里inner_plugin_version设置为:" + pluginConfig.mInnerPluginVersion);
        }

        //加载插件
        DyPluginInfo pluginInfo = PluginInfoFactory.create(mHostContext, pluginFile, pluginPkgInfo, pluginConfig);
        pluginInfo.attachContext(mHostContext);
        LogUtils.i(TAG, "[DyManager#loadInnerPlugin] unchecked plugin [" + pluginInfo + "]");
        if (!pluginInfo.isValid()) {
            pluginInfo.destroy(mHostContext);
            FileUtil.delFile(pluginInfo.getFileAbsolutePath());
            notifyLoadFailed(pluginConfig.mPluginPkgName, pluginPkgInfo.versionCode,
                    DyloadError.ERRORCODE_CREATE_DYPLUGIN_FAIL,
                    DyloadError.ERRORMSG_CREATE_DYPLUGIN_FAIL);
            return false;
        }

        //检查框架版本号
        if (!pluginConfig.supportItfVersion(pluginInfo.getInterfaceVersion())) {
            FileUtil.delFile(pluginFile.getAbsolutePath());
            notifyLoadFailed(pluginConfig.mPluginPkgName, pluginPkgInfo.versionCode,
                    DyloadError.ERRORCODE_INTERFACE_VERSION_NOT_MATCH,
                    getErrorMsgInterfaceVersionNotMatch(pluginConfig, pluginInfo));
            return false;
        }

        //保存已加载插件
        mPluginsHolder.put(pluginInfo.getPackageName(), pluginInfo);
        notifyLoadSuccess(pluginConfig.mPluginPkgName);

        return true;
    }

    private File copyInnerPluginFile(PluginConfig pluginConfig) {
        if (pluginConfig == null) {
            return null;
        }
        LogUtils.i(TAG, "DyManager#copyInnerPluginFile() called with: pluginConfig.mPluginFileNameInAssets = [" + pluginConfig.mPluginFileNameInAssets + "]");

        String destFilePath = DyConstants.getPluginLoadDir(mHostContext) + File.separator
                + pluginConfig.mPluginPkgName.hashCode() + ".jar";
        boolean ret = copyInnerPluginFile(pluginConfig, destFilePath);
        return ret ? new File(destFilePath) : null;
    }

    private boolean copyInnerPluginFile(PluginConfig pluginConfig, String destFilePath) {
        if (pluginConfig == null || TextUtils.isEmpty(pluginConfig.mPluginFileNameInAssets)) {
            return false;
        }
        LogUtils.i(TAG, "DyManager#copyInnerPluginFile() called with: pluginConfig.mPluginFileNameInAssets = ["
                + pluginConfig.mPluginFileNameInAssets
                + "], destFilePath = [" + destFilePath
                + "]");

        boolean ret = false;

        try {
            InputStream is = mHostContext.getAssets().open(pluginConfig.mPluginFileNameInAssets);
            ret = FileUtil.saveInputStreamToSDFile(is, destFilePath);
        } catch (Throwable e) {
            LogUtils.e(TAG, "[DyManager#copyInnerPluginFile] ", e);
        }
        return ret;
    }

    /**
     * 加载插件内的类
     *
     * @param pluginPkgName
     * @param className
     * @return
     */
    public Class<?> loadPluginClass(String pluginPkgName, String className) {
        DyPluginInfo dyPluginInfo = getPluginInfo(pluginPkgName);
        if (dyPluginInfo == null) {
            return null;
        }
        return dyPluginInfo.loadClass(className);
    }

    /**
     * 窥探插件框架及插件列表的消息，为crash报告收集数据用
     *
     * @return
     */
    public static Map<String, String> getPluginInfosMsg(Context context) {
        Map<String, String> msg = new HashMap<>();

        msg.put("DyConstants_Version", DyConstants.VERSION_CODE + "");
        if (context != null) {
            msg.put("CurrProcessName", AppUtils.getCurrProcessName(context));
        }
        Map<String, DyPluginInfo> dyPlugs;
        if (sInstance != null && (dyPlugs = sInstance.mPluginsHolder) != null && dyPlugs.size() > 0) {
            Iterator<String> iterator = dyPlugs.keySet().iterator();
            while (iterator.hasNext()) {
                String key = iterator.next();
                if (!TextUtils.isEmpty(key)) {
                    DyPluginInfo info = dyPlugs.get(key);
                    if (info != null) {
                        String dyProjectMsg = info.getDyProjectVersion();
                        dyProjectMsg = TextUtils.isEmpty(dyProjectMsg) ? "" : dyProjectMsg;
                        String keyString = "DyPluginInfo_" + key;
                        String valueString = "packageName:" + info.getPackageName()
                                + ",VersionName:" + info.getVersionName()
                                + ",VersionCode:" + info.getVersionCode()
                                + ",dyProjectMsg:" + dyProjectMsg;
                        //crash后台会以包名的形式过滤日志的堆栈信息，汇总给商业化，
                        // 防止被过滤拦截到，所以包名要把点换成下划线
                        keyString = keyString.replace(".", "_");
                        valueString = valueString.replace(".", "_");
                        msg.put(keyString, valueString);
                    }
                }
            }
        }
        return msg;
    }

    /**
     * 窥探插件框架及插件列表的消息，为crash报告收集数据用
     * instead of {@link #getPluginInfosMsg(Context)}
     *
     * @return
     */
    @Deprecated
    public static Map<String, String> peeepDyMsg(Context context) {
        return getPluginInfosMsg(context);
    }

    private DyPluginInfo getPluginInfo(String pluginPkgName) {
        if (pluginPkgName == null) {
            return null;
        }
        return mPluginsHolder.get(pluginPkgName);
    }

    /**
     * 获取插件
     *
     * @param pluginPkgName
     * @param isWaitLoaded  是否等待插件加载完成再返回，一般用true
     * @return
     */
    public DyPluginInfo getPluginInfo(String pluginPkgName, boolean isWaitLoaded) {
        if (isWaitLoaded) {
            awaitLoadedLocked(pluginPkgName);
        }
        return getPluginInfo(pluginPkgName);
    }

    /**
     * 插件是否已加载
     *
     * @param pluginPkgName 插件包名
     * @return
     */
    public boolean isPluginLoaded(String pluginPkgName) {
        return getPluginInfo(pluginPkgName) != null;
    }

    /**
     * 插件文件是否已加载
     *
     * @param pluginFilePath 插件文件路径
     * @return
     */
    public boolean isPluginFileLoaded(String pluginFilePath) {
        if (pluginFilePath == null) {
            return false;
        }
        for (DyPluginInfo pluginInfo : mPluginsHolder.values()) {
            if (pluginFilePath.equals(pluginInfo.getFileAbsolutePath())) {
                return true;
            }
        }
        return false;
    }

    public void deleteLocalPluginFile(String pluginPkgName) {
        if (pluginPkgName == null) {
            return;
        }
        List<LocalPluginInfo> localPluginInfos = scanLocalPuginInfos();
        for (LocalPluginInfo localPluginInfo : localPluginInfos) {
            if (pluginPkgName.equals(localPluginInfo.getPackageName())) {
                FileUtil.delFile(localPluginInfo.getFileAbsolutePath());
            }
        }
    }

    /**
     * 加载的插件是否是inner的
     *
     * @param pluginConfig
     * @return
     */
    public boolean isInnerPluginLoaded(PluginConfig pluginConfig, boolean isWaitLoaded) {
        DyPluginInfo dyPluginInfo = getPluginInfo(pluginConfig.getPluginPkgName(), isWaitLoaded);
        if (dyPluginInfo == null) {
            return false;
        }
        try {
            InputStream is = mHostContext.getAssets().open(pluginConfig.mPluginFileNameInAssets);
            String md5Inner = MD5Util.md5(is);
            String md5Loaded = MD5Util.md5(new File(dyPluginInfo.mFileAbsolutePath));
            return TextUtils.equals(md5Inner, md5Loaded);
        } catch (IOException e) {
            e.printStackTrace();
        }
        return false;
    }

    /**
     * 扫描本地插件信息
     *
     * @return
     */
    public List<LocalPluginInfo> scanLocalPuginInfos() {
        List<LocalPluginInfo> localPluginInfos = new LinkedList<>();

        //从内存找
        for (DyPluginInfo dyPluginInfo : mPluginsHolder.values()) {
            localPluginInfos.add(new LocalPluginInfo(dyPluginInfo));
        }

        //从本地文件找
        String path = DyConstants.getPluginLoadDir(mHostContext);
        File[] files = new File(path).listFiles(new FileFilter() {
            @Override
            public boolean accept(File pathname) {
                return !isPluginFileLoaded(pathname.getAbsolutePath());
            }
        });


        if (files != null) {
            for (File file : files) {
                PackageInfo pi = DexUtil.getDexPackageInfo(mHostContext, file.getAbsolutePath());
                if (pi != null) {
                    localPluginInfos.add(new LocalPluginInfo(file.getAbsolutePath(), pi));
                }
            }
        }

        return localPluginInfos;
    }

    /**
     * 添加插件加载事件监听器
     *
     * @param listener
     */
    public void addPluginListener(IPluginLoadListener listener) {
        if (listener != null) {
            mInstallListeners.add(listener);
        }
    }

    /**
     * 移除插件加载事件监听器
     *
     * @param listener
     */
    public void removePluginListener(IPluginLoadListener listener) {
        mInstallListeners.remove(listener);
    }

    /**
     * 设置插件重新加载监听器
     *
     * @param pluginPkgName
     * @param listener
     */
    public void setReloadListener(String pluginPkgName, IPluginReloadListener listener) {
        if (pluginPkgName == null) {
            return;
        }
        mReloadListenerMap.put(pluginPkgName, listener);
    }

    /**
     * 移除插件重新加载监听器
     *
     * @param pluginPkgName
     * @param listener
     */
    public void removeReloadListener(String pluginPkgName, IPluginReloadListener listener) {
        if (getReloadListener(pluginPkgName) == listener) {
            mReloadListenerMap.remove(pluginPkgName);
        }
    }

    public IPluginReloadListener getReloadListener(String pluginPkgName) {
        return mReloadListenerMap.get(pluginPkgName);
    }

    void notifyLoadStart(final String pluginPkgName) {
        LogUtils.i(TAG, "DyManager#notifyLoadStart() called with: pluginPkgName = [" + pluginPkgName + "]");

        setPluginLoadStatus(pluginPkgName, DyloadStatus.PLUGIN_LOAD_STATUS_START);
        CustomThreadExecutorProxy.getInstance().runOnAsyncThread(new Runnable() {
            @Override
            public void run() {
                for (IPluginLoadListener listener : mInstallListeners) {
                    listener.onPluginLoadStart(pluginPkgName);
                }
            }
        });
    }

    void notifyLoadSdFileError(final String pluginPkgName, int versionCode, final int errorCode, final String errorMsg) {
        LogUtils.i(TAG, "DyManager#notifyLoadSdFileError() called with: pluginPkgName = [" + pluginPkgName + "], errorCode = [" + errorCode + "], errorMsg = [" + errorMsg + "]");

        if (DyloadUtil.isMainProcess(mHostContext)) {
            if (errorCode != DyloadError.ERRORCODE_NO_PLUGINFILE) {
                DyStatistics.uploadPluginLoadFail(mHostContext, pluginPkgName, versionCode, errorCode);
                LogUtils.i(TEST_TAG, "sd卡插件加载错误，包名：" + pluginPkgName + ", errorcode：" + errorCode + ", 原因：" + errorMsg);
            }
        }

        CustomThreadExecutorProxy.getInstance().runOnAsyncThread(new Runnable() {
            @Override
            public void run() {
                for (IPluginLoadListener listener : mInstallListeners) {
                    listener.onSdcardPluginFileError(pluginPkgName, errorCode, errorMsg);
                }
            }
        });
    }

    void notifyLoadFailed(final String pluginPkgName, int versionCode, final int errorCode, final String errorMsg) {
        LogUtils.i(TAG, "DyManager#notifyLoadFailed() called with: pluginPkgName = [" + pluginPkgName + "], errorCode = [" + errorCode + "], errorMsg = [" + errorMsg + "]");

        setPluginLoadStatus(pluginPkgName, DyloadStatus.PLUGIN_LOAD_STATUS_FINISH);
        PluginConfig pluginConfig = getPluginConfig(pluginPkgName);
        if (pluginConfig != null) {
            synchronized (pluginConfig) {
                pluginConfig.notifyAll();
            }
            LogUtils.i(TAG, "[DyManager#notifyLoadFailed] <wait-notify> notifyAll on PluginConfig-" + pluginConfig.hashCode());
        }

        if (DyloadUtil.isMainProcess(mHostContext)) {
            DyStatistics.uploadPluginLoadFail(mHostContext, pluginPkgName, versionCode, errorCode);
            LogUtils.i(TEST_TAG, "插件加载失败，包名：" + pluginPkgName + ", errorcode：" + errorCode + ", 原因：" + errorMsg);
        }

        CustomThreadExecutorProxy.getInstance().runOnAsyncThread(new Runnable() {
            @Override
            public void run() {
                for (IPluginLoadListener listener : mInstallListeners) {
                    listener.onPluginLoadFailed(pluginPkgName, errorCode, errorMsg);
                }
            }
        });
    }

    void notifyLoadSuccess(final String pluginPkgName) {
        setPluginLoadStatus(pluginPkgName, DyloadStatus.PLUGIN_LOAD_STATUS_FINISH);
        PluginConfig pluginConfig = getPluginConfig(pluginPkgName);
        if (pluginConfig != null) {
            synchronized (pluginConfig) {
                pluginConfig.notifyAll();
            }
            LogUtils.i(TAG, "[DyManager#notifyLoadSuccess] <wait-notify> notifyAll on PluginConfig-" + pluginConfig.hashCode());
        }

        DyPluginInfo dyPluginInfo = getPluginInfo(pluginPkgName);
        if (dyPluginInfo != null && DyloadUtil.isMainProcess(mHostContext)) {
            DyStatistics.uploadPluginLoadSucc(mHostContext, pluginPkgName, dyPluginInfo.mVersionCode);

            if (LogUtils.isShowLog()) {
                LogUtils.i(TAG, "DyManager#notifyLoadSuccess() called with: pluginPkgName = [" + pluginPkgName
                        + "], pluginInfo = [" + getPluginInfo(pluginPkgName) + "]");
                LogUtils.i(TEST_TAG, "插件加载成功，包名：" + pluginPkgName + ", vesioncode：" + dyPluginInfo.getVersionCode());
            }
        }

        CustomThreadExecutorProxy.getInstance().runOnAsyncThread(new Runnable() {
            @Override
            public void run() {
                for (IPluginLoadListener listener : mInstallListeners) {
                    listener.onPluginLoadSuccess(pluginPkgName);
                }
            }
        });
    }

    private long mAutoLoadStartTime;

    void notifyAutoLoadStart() {
        LogUtils.i(TAG, "DyManager#notifyAutoLoadStart() called");

        mAutoLoadStartTime = System.currentTimeMillis();
        CustomThreadExecutorProxy.getInstance().runOnAsyncThread(new Runnable() {
            @Override
            public void run() {
                for (IPluginLoadListener listener : mInstallListeners) {
                    listener.onAutoLoadPluginsStart();
                }
            }
        });
    }

    void notifyAutoLoadFinish() {
        long time = System.currentTimeMillis() - mAutoLoadStartTime;
        LogUtils.i(TAG, "DyManager#notifyAutoLoadFinish() called, 耗时：" + String.format("%.3f", time * 1.0f / 1000));

        CustomThreadExecutorProxy.getInstance().runOnAsyncThread(new Runnable() {
            @Override
            public void run() {
                for (IPluginLoadListener listener : mInstallListeners) {
                    listener.onAutoLoadPluginsFinish();
                }
            }
        });
    }

    /**
     * 获取入口类
     *
     * @param pluginPkgName
     * @return
     */
    public Object getPluginEntrance(String pluginPkgName) {
        DyPluginInfo pluginInfo = getPluginInfo(pluginPkgName);
        if (pluginInfo != null) {
            return pluginInfo.getEntrance();
        }
        return null;
    }

    /**
     * get the proxy activity class, the proxy activity will delegate the plugin
     * activity
     *
     * @param clazz target activity's class
     * @return
     */
    public Class<? extends Activity> getProxyActivityClass(Class<?> clazz) {
        if (!DyActivityPlugin.class.isAssignableFrom(clazz)) {
            return null;
        }
        Class<? extends DyActivityPlugin> activityPlugin = (Class<? extends DyActivityPlugin>) clazz;
        for (PluginConfig config : mPluginConfigsHolder.values()) {
            Class<? extends Activity> activityClass = config.mapActivity(activityPlugin);
            if (activityClass != null) {
                return activityClass;
            }
        }
        return StandarProxyActivity.class;
    }

    public int startPluginActivity(Context context, DyIntent dlIntent) {
        return startPluginActivityForResult(context, dlIntent, -1);
    }

    public int startPluginActivityForResult(Context context, DyIntent dlIntent, int requestCode) {
        String packageName = dlIntent.getPluginPackage();
        if (TextUtils.isEmpty(packageName)) {
            throw new NullPointerException("disallow null packageName.");
        }
        DyPluginInfo pluginInfo = mPluginsHolder.get(packageName);
        if (pluginInfo == null) {
            return ActivityResult.START_RESULT_NO_PKG;
        }

        final String className = getPluginActivityFullPath(dlIntent, pluginInfo);
        Class<?> clazz = pluginInfo.loadClass(className);
        if (clazz == null) {
            return ActivityResult.START_RESULT_NO_CLASS;
        }

        Class<? extends Activity> activityClass = getProxyActivityClass(clazz);
        if (activityClass == null) {
            return ActivityResult.START_RESULT_TYPE_ERROR;
        }

        dlIntent.putExtra(DyConstants.EXTRA_CLASS, className);
        dlIntent.putExtra(DyConstants.EXTRA_PACKAGE, packageName);
        dlIntent.setClass(mHostContext, activityClass);
        performStartActivityForResult(context, dlIntent, requestCode);
        return ActivityResult.START_RESULT_SUCCESS;
    }

    private static String getPluginActivityFullPath(DyIntent dlIntent, DyPluginInfo dyPluginInfo) {
        String className = dlIntent.getPluginClass();
        if (className.startsWith(".")) {
            className = dlIntent.getPluginPackage() + className;
        }
        return className;
    }

    private static void performStartActivityForResult(Context context, DyIntent dyIntent, int requestCode) {
        /**
         * 解决crash， Coolpad A8和Coolpad R108， API 22：
         * Exception android.util.AndroidRuntimeException: Unknown error code -6 when starting Intent { flg=0x10800000 cmp=com.jb.zcamera/com.jiubang.commerce.dyload.pl.chargelocker.cell.ChargeBatteryProxyActivity (has extras) }
         android.app.Instrumentation.checkStartActivityResult (Instrumentation.java:1829)
         android.app.Instrumentation.execStartActivity (Instrumentation.java:1534)
         android.app.ContextImpl.startActivity (ContextImpl.java:1293)
         android.app.ContextImpl.startActivity (ContextImpl.java:1275)
         android.content.ContextWrapper.startActivity (ContextWrapper.java:323)
         com.jiubang.commerce.dyload.manager.DyManager.performStartActivityForResult (ZeroCamera:1183)
         */
        try {
            if (context instanceof Activity) {
                ((Activity) context).startActivityForResult(dyIntent, requestCode);
            } else {
                context.startActivity(dyIntent);
            }
        } catch (AndroidRuntimeException e) {
            e.printStackTrace();
        }
    }

    class PluginScanEntry {
        File mFile;
        PackageInfo mPackageInfo;

        PluginScanEntry(File file, PackageInfo packageInfo) {
            mFile = file;
            mPackageInfo = packageInfo;
        }
    }

    /**
     * 强制删除本地插件，不管当前插件是否已加载
     *
     * @param context
     * @param pluginPkgName 插件包名
     */
    public static void deleteLocalPluginForce(Context context, String pluginPkgName) {
        LogUtils.i(TAG, "DyManager#deleteLocalPlugin() called with: pluginPkgName = [" + pluginPkgName + "]");

        if (pluginPkgName == null) {
            return;
        }
        List<LocalPluginInfo> localPluginInfos = scanLocalPuginInfos(context);

        for (LocalPluginInfo localPluginInfo : localPluginInfos) {
            if (pluginPkgName.equals(localPluginInfo.getPackageName())) {
                FileUtil.delFile(localPluginInfo.getFileAbsolutePath());
            }
        }
    }

    private static List<LocalPluginInfo> scanLocalPuginInfos(Context context) {
        List<LocalPluginInfo> localPluginInfos = new LinkedList<>();

        //从本地文件找
        String path = DyConstants.getPluginLoadDir(context);
        File[] files = new File(path).listFiles(new FileFilter() {
            @Override
            public boolean accept(File pathname) {
                return true;
            }
        });


        if (files != null) {
            for (File file : files) {
                PackageInfo pi = DexUtil.getDexPackageInfo(context, file.getAbsolutePath());
                if (pi != null) {
                    localPluginInfos.add(new LocalPluginInfo(file.getAbsolutePath(), pi));
                }
            }
        }

        return localPluginInfos;
    }
}