package com.jiubang.core.message;

import java.util.ArrayList;
import java.util.Collection;
import java.util.Iterator;
import java.util.List;
import java.util.concurrent.ConcurrentHashMap;

/**
 * 所有消息接收者的管理池
 * 职责：负责管理所有消息的接收者，负责返回特定消息的接收者
 * TODO 防止线程同步问题，比较好的方式是将所有get元素和add，remove，set方法都加入synchronized
 * 
 * @author yuankai
 * @version 1.0
 */
class MessageHandlerPool
{
	/**
	 * 所有消息的接收者，该类接收者在发送任何消息时都被发送到
	 */
	private ConcurrentHashMap<Integer, IMessageHandler> mAllMsgHandlers;
	
	/**
	 * 定点消息接收者，此类接收者只接收某些特定的消息
	 */
	private ConcurrentHashMap<Integer, List<IMessageHandler>> mSpecificMsgHandlers;
	
	MessageHandlerPool()
	{
		mAllMsgHandlers = new ConcurrentHashMap<Integer, IMessageHandler>();
		mSpecificMsgHandlers = new ConcurrentHashMap<Integer, List<IMessageHandler>>();
	}
	
	/**
	 * 根据ID获取消息处理者
	 * @param handlerId 处理者ID
	 * @return 处理者
	 */
	IMessageHandler getHandler(final int msgId, final int handlerId)
	{
		// 先从所有接收者中找，如果找到了就返回，否则找特定消息ID是否有处理者，如果有则返回
		IMessageHandler handler = mAllMsgHandlers.get(new Integer(handlerId));
		if(handler == null)
		{
			return getHandlerInList(handlerId, mSpecificMsgHandlers.get(new Integer(msgId)));
		}
		else
		{
			return handler;
		}
	}
	
	private synchronized static IMessageHandler getHandlerInList(final int handlerId, final List<IMessageHandler> handlers)
	{
		if(handlers == null)
		{
			return null;
		}
		
		final int size = handlers.size();
		for(int i = 0;i < size; i++)
		{
			final IMessageHandler handler = handlers.get(i);
			if(handler != null && handler.getMsgHandlerId() == handlerId)
			{
				return handler;
			}
		}
		
		return null;
	}
	
	/**
	 * 根据消息ID，找出处理此消息的所有接收者
	 * @param msgId 消息ID
	 * @return 接收者列表
	 */
	synchronized List<IMessageHandler> getAllHandlers(final int msgId)
	{
		return mergeHandlers(mAllMsgHandlers.values(), mSpecificMsgHandlers.get(new Integer(msgId)));
	}
	
	/**
	 * 根据消息ID和接收者数组，按照优先级，找出处理此消息的所有接收者
	 * @param msgId 消息ID
	 * @param orderedHandlers 要求被排序到前面的接收者ID数组
	 * @return 接收者列表
	 */
	synchronized List<IMessageHandler> getAllHandlers(final int msgId,
			final int[] orderedHandlers) {
		List<IMessageHandler> handlers = mergeHandlers(mAllMsgHandlers.values(), mSpecificMsgHandlers.get(new Integer(msgId)));
		
		if(orderedHandlers != null && orderedHandlers.length >= 0) {
			// 对列表进行重新排序
			final int orderedSize = orderedHandlers.length;
			// 位置计数索引
			int index = 0;
			for(int i = 0;i < orderedSize; i++) {
				final int orderedHandlerId = orderedHandlers[i];
				// 每次都取最新的长度，保证不会出错
				// TODO 此处可以优化，在保证队列长度不改变的情况下，可以只取一次 
				final int handlerSize = handlers.size();
				for(int j = 0;j < handlerSize; j++) {
					final IMessageHandler handler = handlers.get(j);
					if(handler.getMsgHandlerId() == orderedHandlerId) {
						// 重置位置
						handlers.remove(handler);
						handlers.add(index, handler);
						index ++;
						break;
					}
				}
			}
		}
		return handlers;
	}
	
	/**
	 * 将collection中的消息接受者合并入list的消息接收者之中,剔除重复的接收者，保证只接收一次
	 * @param collections 所有消息接收者
	 * @param lists 特定消息接收者
	 * @return 合并之后的接收者集合
	 */
	private static List<IMessageHandler> mergeHandlers(Collection<IMessageHandler> collections,
			List<IMessageHandler> lists) {
		final List<IMessageHandler> mergeLists = new ArrayList<IMessageHandler>();
		if(lists != null) {
			mergeLists.addAll(lists);
		}
		
		if(collections != null) {
			Iterator<IMessageHandler> iterator = collections.iterator();
			while(iterator.hasNext()) {
				IMessageHandler handler = iterator.next();
				if(!mergeLists.contains(handler)) {
					mergeLists.add(handler);
				}
			}
		}
		
		return mergeLists;
	}
	
	/**
	 * 注册指定消息的接收者 
	 * @param handler 消息接收者
	 * @param msgId 消息ID
	 * @return 是否注册成功
	 */
	synchronized boolean registMsgHandler(final IMessageHandler handler, final int msgId)
	{
		// 作保护
		if(handler == null)
		{
			return false;
		}
		
		boolean ret = false;
		if(msgId <= MessageManager.BASE_MSG_ID)
		{
			// 注册为所有消息接收者
			Integer integer = new Integer(handler.getMsgHandlerId());
			mAllMsgHandlers.put(integer, handler);
		}
		else
		{
			// 注册为特定消息接收者
			Integer integer = new Integer(msgId);
			if(mSpecificMsgHandlers.contains(integer))
			{
				List<IMessageHandler> handlers = mSpecificMsgHandlers.get(integer);
				if(handlers == null)
				{
					handlers = new ArrayList<IMessageHandler>();
				}
				ret = handlers.add(handler);
			}
			else 
			{
				List<IMessageHandler> handlers = new ArrayList<IMessageHandler>();
				handlers.add(handler);
				ret = (mSpecificMsgHandlers.put(integer, handlers) == null ? false : true);
			}
		}
		
		return ret;
	}
	
	/**
	 * 反注册指定消息的接收者
	 * @param handler 消息接收者
	 * @param msgId 消息ID
	 * @return 是否反注册成功
	 */
	synchronized boolean unRegistMsgHandler(final IMessageHandler handler, final int msgId)
	{
		if(handler == null)
		{
			return false;
		}
		
		boolean ret = false;
		if(msgId <= MessageManager.BASE_MSG_ID)
		{
			// 反注册为所有消息接收者
			Integer integer = new Integer(handler.getMsgHandlerId());
			ret = (mAllMsgHandlers.remove(integer) == null ? false : true);
		}
		else
		{
			// 反注册为特定消息接收者
			Integer integer = new Integer(msgId);
			List<IMessageHandler> handlers = mSpecificMsgHandlers.get(integer);
			if(handlers != null)
			{
				ret = handlers.remove(handler);
				
				if(handlers.size() == 0)
				{
					// 则此项也需被删掉
					mSpecificMsgHandlers.remove(integer);
				}
			}
		}
		
		return ret;
	}
	
	/**
	 * 注册所有消息的接收者
	 * @param handler 消息接收者
	 * @return 是否注册成功
	 */
	synchronized boolean registMsgHandler(final IMessageHandler handler)
	{
		// -1代表注册所有消息
		return registMsgHandler(handler, MessageManager.BASE_MSG_ID);
	}
	
	/**
	 * 反注册指定消息的接收者
	 * @param handler 消息接收者
	 * @param msgId 消息ID
	 * @return 是否反注册成功
	 */
	synchronized boolean unRegistMsgHandler(final IMessageHandler handler)
	{
		return unRegistMsgHandler(handler, MessageManager.BASE_MSG_ID);
	}

	public void cleanup()
	{
		mAllMsgHandlers.clear();
		mSpecificMsgHandlers.clear();
		mAllMsgHandlers = null;
		mSpecificMsgHandlers = null;
	}
}
