package com.jiubang.gotoollocker;

import android.Manifest;
import android.content.Context;
import android.content.Intent;
import android.content.IntentFilter;
import android.content.pm.PackageInfo;
import android.content.pm.PackageManager;
import android.os.BatteryManager;
import android.os.Environment;
import android.os.Handler;
import android.os.Looper;
import android.os.storage.StorageManager;
import android.support.annotation.NonNull;
import android.support.v4.app.ActivityCompat;
import android.support.v4.util.Pair;
import android.text.TextUtils;

import com.jb.ga0.commerce.util.LogUtils;
import com.jiubang.gotoollocker.receiver.ChargeLockerEventReceiver;
import com.jiubang.gotoollocker.receiver.PresentReceiver;
import com.jiubang.gotoollocker.statistic.ToolLockerHolderStatistic;

import java.io.File;
import java.io.IOException;
import java.lang.ref.WeakReference;
import java.lang.reflect.Method;
import java.util.ArrayList;
import java.util.List;

/**
 * 展位和显示锁的辅助类,接入方式如下:
 * <p>
 * 1. 在应用启动的时候调用 {@link #applyHolder()}
 * 2. 在需要展示锁屏的时候调用 {@link #prepareLock(ToolLockerCallback)}以规避其他锁屏应用
 * 3. 在锁屏关闭时调用 {@link #releaseLock()} 以接触显示锁
 *
 * @author chenxiqiang
 * @since 2017-06-28
 */
public class DisplayLockHelper {

    public static final String TAG = "DisplayLockHelper_SDK";

    private volatile static DisplayLockHelper instance = null;

    public static DisplayLockHelper getInstance(Context mAppContext) {
        if (instance == null) {
            synchronized (DisplayLockHelper.class) {
                if (instance == null) {
                    instance = new DisplayLockHelper(mAppContext);
                }
            }
        }
        return instance;
    }

    public static String sChargeLockerHolderDir;
    public static String sToolLockerHolderDir;

    private static final String LOCK_FILE = "display.lock";
    private static final String HOLDER_FILE = "hodler.cfg";

    private final SimpleFileLock mToolLockerDisplayLock;
    private final SimpleFileLock mChargeLockerDisplayLock;

    private final Handler mHandler;
    private final Context mAppContext;

    private final ChargeLockerEventReceiver mChargeLockerEventReceiver;
    private final PresentReceiver mPresentReceiver;

    private DisplayLockHelper(Context mAppContext) {
        if (mAppContext == null) {
            throw new IllegalArgumentException();
        }
        this.mAppContext = mAppContext.getApplicationContext();
        mHandler = new Handler(Looper.myLooper());

        if (sChargeLockerHolderDir == null) {
            sChargeLockerHolderDir = getExternalStorageDir(mAppContext)
                    + File.separator + "Android" + File.separator + "data"
                    + File.separator + "com.jiubang.gochargelockor"
                    + File.separator;
        }
        if (sToolLockerHolderDir == null) {
            sToolLockerHolderDir = getExternalStorageDir(mAppContext)
                    + File.separator + "Android" + File.separator + "data"
                    + File.separator + "com.jiubang.gotoollocker"
                    + File.separator;
        }
        mToolLockerDisplayLock = SimpleFileLock.get(sToolLockerHolderDir + LOCK_FILE);
        mChargeLockerDisplayLock = SimpleFileLock.get(sChargeLockerHolderDir + LOCK_FILE);

        mChargeLockerEventReceiver = new ChargeLockerEventReceiver() {
            @Override
            public void onChargeLockerStateChanged(String pkg, boolean shown) {
                super.onChargeLockerStateChanged(pkg, shown);
                DisplayLockHelper.this.onChargeLockerStateChanged(pkg, shown);
            }
        };
        mChargeLockerEventReceiver.register(this.mAppContext);
        mPresentReceiver = new PresentReceiver().register(this.mAppContext);

        LogUtils.d(TAG, "初始化规避接口");
    }

    /*ChargeLockerListener*/

    public interface OnChargeLockerListener {
        /**
         * 充电锁状态变更的广播。
         * <p>
         * 返回值会影响统计
         *
         * @param clPkg 充电锁应用包名
         * @param shown 充电锁切换为展示还是关闭
         * @return 是否主动关闭工具锁屏
         */
        boolean onChargeLockerStateChanged(String clPkg, boolean shown);
    }

    private final List<OnChargeLockerListener> mOnChargeLockerListeners = new ArrayList<>();

    public void addOnChargeLockerListener(OnChargeLockerListener onChargeLockerListener) {
        if (onChargeLockerListener != null) {
            synchronized (mOnChargeLockerListeners) {
                mOnChargeLockerListeners.add(onChargeLockerListener);
            }
        }
    }

    public boolean removeOnChargeLockerListener(OnChargeLockerListener onChargeLockerListener) {
        if (onChargeLockerListener != null) {
            synchronized (mOnChargeLockerListeners) {
                return mOnChargeLockerListeners.remove(onChargeLockerListener);
            }
        }
        return false;
    }

    private void onChargeLockerStateChanged(String clPkg, boolean shown) {
        if (!hasPermission()) {//无文件读写权限则不对工具锁逻辑进行干预
            LogUtils.d(TAG, "onChargeLockerStateChanged-> 无文件读写权限，不对逻辑进行干预");
            return;
        }

        PrepareTask prepareTask = mLastPrepareTask;
        if (prepareTask != null) {
            LogUtils.i(TAG, "CANCEL BY CHARGE LOCKER STATE CHANGE");
            prepareTask.cancel(ToolLockerCallback.CANCEL_BY_CHARGE_LOCKER, clPkg);
        }

        boolean isFinish = false;
        synchronized (mOnChargeLockerListeners) {
            for (OnChargeLockerListener listener : mOnChargeLockerListeners) {
                if (listener.onChargeLockerStateChanged(clPkg, shown)) {
                    isFinish = true;
                }
            }
        }

        LogUtils.d(TAG, "onChargeLockerStateChanged-> 规避充电锁而关闭界面：" + isFinish);

        if (isFinish) {//添加统计
            ToolLockerHolderStatistic.uploadToolLockBreak(mAppContext, mAppContext.getPackageName(), clPkg);
        }
    }

    /*ExternalStorage*/

    private static String sExternalStorageDir;

    /**
     * 获取内置存储目录。
     * <p>
     * 方法实现来自 DyloadSdk
     *
     * @param context
     * @return
     */
    public static String getExternalStorageDir(Context context) {
        if (sExternalStorageDir == null) {

            String externalPath;
            try {
                externalPath = Environment.getExternalStorageDirectory().getAbsolutePath();
            } catch (Exception e) {
                e.printStackTrace();
                externalPath = "/sdcard";
            }
            List<String> sdList = getAllSDPaths(context);
            int index = sdList.indexOf(externalPath);

            String commonFilePath;

            if (index >= 0) {
                commonFilePath = externalPath;
            } else {
                if (sdList.size() > 0) {
                    commonFilePath = sdList.get(0);
                } else {
                    commonFilePath = externalPath;
                }
            }
            sExternalStorageDir = commonFilePath;
        }
        return sExternalStorageDir;
    }

    private static List<String> getAllSDPaths(Context context) {
        List<String> files = new ArrayList<>();
        StorageManager sm = (StorageManager) context.getSystemService(Context.STORAGE_SERVICE);
        try {
            Class[] paramClasses = new Class[0];
            Method getVolumePathsMethod = StorageManager.class.getMethod("getVolumePaths", paramClasses);
            getVolumePathsMethod.setAccessible(true);
            Object[] params = new Object[0];
            String[] paths = (String[]) getVolumePathsMethod.invoke(sm, params);

            for (int i = 0; i < paths.length; i++) {
                String status = (String) sm.getClass().getMethod("getVolumeState", new Class[]{String.class}).invoke(sm, paths[i]);

                if (status.equals("mounted"))
                    files.add(paths[i]);
            }
        } catch (Exception e) {
            e.printStackTrace();
            return files;
        }
        return files;
    }

    /*Hold*/

    public boolean applyHolder() {
        boolean result;
        if (hasPermission()) {
            result = new LockerHolder(mAppContext, sToolLockerHolderDir + HOLDER_FILE).applyHolder();
            LogUtils.d(TAG, "applyHolder:拥有文件读写权限，执行applyHolder结果：" + result);
        } else {
            result = false;
            LogUtils.d(TAG, "applyHolder：没有文件读写权限，applyHolder默认结果为 false");
        }
        return result;
    }

    @NonNull
    private Pair<Boolean, String> applyHolderResult() {
        LockerHolder lockerHolder = new LockerHolder(mAppContext, sToolLockerHolderDir + HOLDER_FILE);
        boolean result = lockerHolder.applyHolder();
        String pkgName = "";
        if (!result) {//仅当占位失败时才去读取占位包名
            pkgName = lockerHolder.getInfo().getHolderPackageName();
            if (pkgName == null) {
                pkgName = "";
            }
        }
        LogUtils.d(TAG, "applyHolder: " + result + " pkgName:" + pkgName);
        return Pair.create(result, pkgName);
    }

    public void recordShown() {
        if (hasPermission()) {
            new LockerHolder(mAppContext, sToolLockerHolderDir + HOLDER_FILE).recordShowed();
            LogUtils.d(TAG, "recordShown->写入占位文件及展示状态为true");
        } else {
            LogUtils.d(TAG, "recordShown-> 没有文件读写权限，不处理");
        }
    }
    
    /*Lock*/

    public boolean isChargeLockerFileExists() {
        return mChargeLockerDisplayLock.isFileExists();
    }

    public boolean isChargeLockerExists() {
        if (!mChargeLockerDisplayLock.isFileExists()) {
            LogUtils.d(TAG, "isChargeLockerExists: " + false + ",锁文件不存在");
            return false;
        } else {
            String pkg = null;
            try {//由于占位文件不可靠，仅通过锁文件的内容来判断充电锁是否存在
                pkg = mChargeLockerDisplayLock.read();
            } catch (IOException e) {
                e.printStackTrace();
            }

            boolean result = !TextUtils.isEmpty(pkg) && isAppInstalled(pkg);
            LogUtils.d(TAG, "isChargeLockerExists: " + result);
            return result;
        }
    }

    public boolean isChargeLockerShowing() {
        boolean result = isChargeLockerExists() && !mChargeLockerDisplayLock.testLock();
        LogUtils.d(TAG, "isChargeLockerShowing: " + result);
        return result;
    }

    private boolean isCharging() {
        IntentFilter filter = new IntentFilter(Intent.ACTION_BATTERY_CHANGED);
        Intent batteryStatus = mAppContext.registerReceiver(null, filter);
        Integer status = batteryStatus != null ? batteryStatus.getIntExtra(BatteryManager.EXTRA_STATUS, -1) : null;
        boolean result = status != null && (status == BatteryManager.BATTERY_STATUS_CHARGING ||
                status == BatteryManager.BATTERY_STATUS_FULL);
        LogUtils.d(TAG, "isCharging: " + result);
        return result;
    }

    public interface ToolLockerCallback {

        // 占位失败导致失败
        int FAIL_BY_HOLDER = 1;
        // 无法对显示锁文件上锁导致失败
        int FAIL_BY_LOCK_FILE = 2;
        // 充电锁的显示锁文件已经被上锁，说明充电锁已经启动，导致失败
        int FAIL_BY_CHARGE_LOCKER_DISPLAY_LOCK = 3;
        // 充电锁短时间内有动作，避让
        int FAIL_BY_CHARGE_LOCKER_RECENT_STATE = 4;
        // 3秒内发起重复请求
        int FAIL_BY_TOO_OFTEN_CALL = 5;

        // 充电锁存在时延迟等待结果，此时主动取消
        int CANCEL = 6;
        // 充电锁存在时延迟等待结果，此时用户快速解锁
        int CANCEL_BY_UNLOCK = 7;
        // 充电锁存在时延迟等待结果，此时再次发起新的请求
        int CANCEL_BY_NEW_CALL = 8;
        // 充电锁存在时延迟等待结果，此时充电锁状态变化
        int CANCEL_BY_CHARGE_LOCKER = 9;

        int SUCCESS = 0;

        void onResult(int result);

    }

    public class PrepareTask implements Runnable {

        WeakReference<ToolLockerCallback> mReference;

        public PrepareTask(ToolLockerCallback reference) {
            mReference = new WeakReference<>(reference);
        }

        @Override
        public void run() {
            synchronized (PrepareTask.class) {
                WeakReference<ToolLockerCallback> ref = mReference;
                ToolLockerCallback callback = ref != null ? ref.get() : null;
                if (callback != null) {
                    performLock(callback, true);
                }

                mReference = null;
                DisplayLockHelper.this.mLastPrepareTask = null;
            }
        }

        public boolean isCanceled() {
            return mReference == null;
        }

        public void cancel() {
            cancel(ToolLockerCallback.CANCEL, null);
        }

        private void cancel(int result, String conflictPkg) {
            synchronized (PrepareTask.class) {
                LogUtils.i(TAG, "cancel:" + result);
                mHandler.removeCallbacks(this);

                WeakReference<ToolLockerCallback> ref = mReference;
                ToolLockerCallback callback = ref != null ? ref.get() : null;
                if (callback != null) {
                    call(callback, result, conflictPkg);
                }
                if (ref != null) {
                    ref.clear();
                }

                mReference = null;
                DisplayLockHelper.this.mLastPrepareTask = null;
            }
        }

    }

    //当充电锁存在,延迟若干时间让其优先处理
    private static final long DELAY_IF_CHARGE_LOCKER = 3000;
    private long mLastCallTime;
    private PrepareTask mLastPrepareTask;

    private boolean hasPermission() {
        return ActivityCompat.checkSelfPermission(mAppContext, Manifest.permission.WRITE_EXTERNAL_STORAGE) == PackageManager.PERMISSION_GRANTED;
    }

    public PrepareTask prepareLock(final ToolLockerCallback callback) {
        LogUtils.d(TAG, "调用：prepareLock");

        //检查权限
        if (!hasPermission()) {
            LogUtils.d(TAG, "prepareLock-> 应用没有文件读写权限，默认判定成功");
            call(callback, ToolLockerCallback.SUCCESS, null);//没有文件读写权限默认判定为成功
            return null;
        }

        //检查两次启动间隔
        long nowTime = System.currentTimeMillis();
        long toolLockerDeltaTime = nowTime - mLastCallTime;
        if (toolLockerDeltaTime <= DELAY_IF_CHARGE_LOCKER) {
            LogUtils.d(TAG, "prepareLock: FAIL_BY_TOO_OFTEN_CALL,deltaTime:" + toolLockerDeltaTime);
            call(callback, ToolLockerCallback.FAIL_BY_TOO_OFTEN_CALL, null);
            return null;
        }
        mLastCallTime = nowTime;

        //清除上次任务
        PrepareTask prepareTask = mLastPrepareTask;
        if (prepareTask != null) {
            prepareTask.cancel(ToolLockerCallback.CANCEL_BY_NEW_CALL, null);
        }

        //检查充电锁启动间隔
        long chargeLockerDeltaTime = System.currentTimeMillis() - ChargeLockerEventReceiver.getLastChangedTime();
        String pkg = ChargeLockerEventReceiver.getLastPackage();
        if (pkg == null) {
            pkg = "";
        }

        LogUtils.d(TAG, "prepareLock: 距充电锁状态变化时间:" + chargeLockerDeltaTime);
        if (chargeLockerDeltaTime <= DELAY_IF_CHARGE_LOCKER) {
            LogUtils.d(TAG, "prepareLock: FAIL_BY_CHARGE_LOCKER_RECENT_STATE_充电锁短时间内状态变更过，避让");
            call(callback, ToolLockerCallback.FAIL_BY_CHARGE_LOCKER_RECENT_STATE, pkg);
            return null;
        }

        Pair<Boolean, String> applyResult = applyHolderResult();
        if (Boolean.TRUE.equals(applyResult.first)) {
            if (!mChargeLockerDisplayLock.isFileExists()) {//充电锁文件不存在，准备展示
                performLock(callback, false);

            } else if (isChargeLockerExists() && isCharging()) {//充电锁存在且充电中，延迟让充电锁优先处理
                synchronized (PrepareTask.class) {
                    mLastPrepareTask = new PrepareTask(callback);
                }
                mHandler.postDelayed(mLastPrepareTask, DELAY_IF_CHARGE_LOCKER);
                return mLastPrepareTask;

            } else {//充电锁文件存在，但不在充电
                performLock(callback, false);

            }

        } else {
            String tlPkg = applyResult.second;
            call(callback, ToolLockerCallback.FAIL_BY_HOLDER, tlPkg);
            LogUtils.i(TAG, "performLock: FAIL_BY_HOLDER");
        }
        return null;
    }

    private void performLock(ToolLockerCallback callback, boolean fromTask) {
        if (fromTask && mPresentReceiver.isUserPresent()) {//用户快速解锁导致失败
            call(callback, ToolLockerCallback.CANCEL_BY_UNLOCK, null);
            LogUtils.i(TAG, "performLock: CANCEL_BY_UNLOCK");
            return;
        }

        if (mChargeLockerDisplayLock.isFileExists() && !mChargeLockerDisplayLock.testLock()) {//充电锁文件存在，且展示中
            String clPkg = mChargeLockerDisplayLock.readSafe();
            call(callback, ToolLockerCallback.FAIL_BY_CHARGE_LOCKER_DISPLAY_LOCK, clPkg);
            LogUtils.i(TAG, "performLock: FAIL_BY_CHARGE_LOCKER_DISPLAY_LOCK");

        } else if (mToolLockerDisplayLock.tryLock()) {
            //向锁文件中写入包名
            mToolLockerDisplayLock.writeSafe(mAppContext.getPackageName());
            call(callback, ToolLockerCallback.SUCCESS, null);
            LogUtils.i(TAG, "performLock: SUCCESS");

        } else {
            String tlPkg = mToolLockerDisplayLock.readSafe();
            call(callback, ToolLockerCallback.FAIL_BY_LOCK_FILE, tlPkg);
            LogUtils.i(TAG, "performLock: FAIL_BY_LOCK_FILE");
        }
    }

    private void call(ToolLockerCallback callback, int result, String conflictPkg) {
        if (LogUtils.isShowLog()) {
            String logMsg = null;
            switch (result) {
                case ToolLockerCallback.FAIL_BY_HOLDER:
                    logMsg = "占位失败导致失败";
                    break;
                case ToolLockerCallback.FAIL_BY_LOCK_FILE:
                    logMsg = "无法对显示锁文件上锁导致失败";
                    break;
                case ToolLockerCallback.FAIL_BY_CHARGE_LOCKER_DISPLAY_LOCK:
                    logMsg = "充电锁的显示锁文件已经被上锁，说明充电锁已经启动，导致失败";
                    break;
                case ToolLockerCallback.FAIL_BY_CHARGE_LOCKER_RECENT_STATE:
                    logMsg = "充电锁短时间内有动作，避让";
                    break;
                case ToolLockerCallback.FAIL_BY_TOO_OFTEN_CALL:
                    logMsg = "3秒内发起重复请求";
                    break;
                case ToolLockerCallback.CANCEL:
                    logMsg = "充电锁存在时延迟等待结果，此时主动取消";
                    break;
                case ToolLockerCallback.CANCEL_BY_UNLOCK:
                    logMsg = "充电锁存在时延迟等待结果，此时用户快速解锁";
                    break;
                case ToolLockerCallback.CANCEL_BY_NEW_CALL:
                    logMsg = "充电锁存在时延迟等待结果，此时再次发起新的请求";
                    break;
                case ToolLockerCallback.CANCEL_BY_CHARGE_LOCKER:
                    logMsg = "充电锁存在时延迟等待结果，此时充电锁状态变化";
                    break;

                case ToolLockerCallback.SUCCESS:
                    logMsg = "获取展示权成功";
                    break;
            }
            if (logMsg != null && !TextUtils.isEmpty(conflictPkg)) {
                logMsg += ", 冲突包名：" + conflictPkg;
            }
            LogUtils.d(TAG, "规避结果：" + result + "  ,msg:" + logMsg);
        }

        if (callback != null) {
            callback.onResult(result);
        }
        boolean success = result == ToolLockerCallback.SUCCESS;
        ToolLockerHolderStatistic.uploadToolLockCanShow(
                mAppContext,
                mAppContext.getPackageName(),
                success,
                success ? 0 : result,
                conflictPkg
        );
    }

    public void releaseLock() {
        mToolLockerDisplayLock.release();
        LogUtils.i(TAG, "releaseLock: ");
    }

    private boolean isAppInstalled(String pkgName) {
        PackageManager pkgMgr = mAppContext.getPackageManager();
        PackageInfo info = null;
        try {
            info = pkgMgr.getPackageInfo(pkgName, PackageManager.GET_CONFIGURATIONS);
        } catch (PackageManager.NameNotFoundException e) {
            e.printStackTrace();
        }

        LogUtils.i(TAG, pkgName + " installed:" + info);
        return info != null;
    }

    /*Config*/

    public static class PackageConfig {

        private int mVersionCode;
        private String mCid;

        public int getVersionCode() {
            return mVersionCode;
        }

        public void setVersionCode(int versionCode) {
            mVersionCode = versionCode;
        }

        public String getCid() {
            return mCid;
        }

        public void setCid(String cid) {
            mCid = cid;
        }

    }

    private PackageConfig mConfig;

    public PackageConfig getPackageConfig() {
        if (mConfig == null) {
            throw new IllegalStateException("请先设置PackageConfig");
        }
        return mConfig;
    }

    public DisplayLockHelper setPackageConfig(PackageConfig config) {
        mConfig = config;
        return this;
    }
}
